package proxy

import (
	"crypto/ecdsa"
	"crypto/rand"
	"crypto/sha512"
	"crypto/x509"
	"encoding/base64"
	"encoding/json"
	"encoding/pem"
	"errors"
	"fmt"
	"net/http"
	"net/url"
	"strconv"
	"strings"
	"time"

	uuid "github.com/google/uuid"
	log "github.com/sirupsen/logrus"
	ed "golang.org/x/crypto/ed25519"
)

const (
	contentTypeHeader   = "content-type"
	applicationProtobuf = "application/x-protobuf"
	signPattern         = "%s\n%s\n%s\n%s\n%s"
	layout              = "Mon, 02 Jan 2006 15:04:05 GMT"
)

type metastr struct {
	AccessKey  string `json:"access_key_id"`
	AuthMethod string `json:"auth_method"`
}

type tokenstr struct {
	AccessKeyId string `json:"access_key_id"`
	ExpiredAt   string `json:"expiredAt"`
	SignedStr   string `json:"signedStr"`
}

func RequestSign(config *ReverseProxyConfig, credentials *CdpCredential, tokenHolder *TokenHolder, req *http.Request, baseAPIPath string) error {
	date := formatDate()
	auth, msgToVerify, err := authHeader(credentials.AccessKeyId, credentials.PrivateKey, credentials.SignAlgorithm, credentials.SignAlgorithmAuthHeader, req.Method, resourcePath(baseAPIPath, req.URL.Path, req.URL.Query().Encode()), date)
	if err != nil {
		return err
	}
	if config.Sign.UseTokenHeader {
		token, err := tokenHeader(credentials.PrivateKey, credentials.SignAlgorithm, tokenHolder)
		if err != nil {
			return err
		}
		req.Header.Set(config.Sign.TokenHeaderName, token)
	}
	if len(config.Sign.ResourceCrnHeader) > 0 && len(config.Sign.ResourceCrnHeaderName) > 0 {
		req.Header.Set(config.Sign.ResourceCrnHeaderName, config.Sign.ResourceCrnHeader)
	}
	if len(config.Sign.InstanceHeader) > 0 && len(config.Sign.InstanceHeaderName) > 0 {
		req.Header.Set(config.Sign.InstanceHeaderName, config.Sign.InstanceHeader)
	}
	req.Header.Set(config.Sign.AuthHeaderName, auth)
	req.Header.Set(contentTypeHeader, applicationProtobuf)
	req.Header.Set(config.Sign.DateHeaderName, date)
	if config.Sign.UseMsgHeader {
		req.Header.Set(config.Sign.MsgToVerifyName, urlSafeBase64Encode([]byte(msgToVerify)))
	}
	if config.Sign.UseRequestIdHeader {
		requestId := uuid.New().String()
		req.Header.Set(config.Sign.RequestIdHeaderName, requestId)
		log.Debugf("Request generated with request id: %s", requestId)
	}
	return nil
}

func newMetastr(accessKeyID string, algoHeader string) *metastr {
	return &metastr{accessKeyID, algoHeader}
}

func newTokenstr(token *Token, privateKey string, algo string) (*tokenstr, error) {
	strTimestamp := strconv.FormatInt(token.ExpireAt.Unix(), 10)
	signature, err := signWithAlgo(privateKey, algo, strTimestamp)
	if err != nil {
		return nil, err
	}
	return &tokenstr{AccessKeyId: token.AccessKeyId, ExpiredAt: strTimestamp, SignedStr: urlSafeBase64Encode(signature)}, nil
}

var prefixTrim = []string{"http://", "https://"}

func CutAndTrimAddress(address string) (string, string) {
	for _, v := range prefixTrim {
		address = strings.TrimPrefix(address, v)
	}
	address = strings.TrimRight(address, "/ ")
	basePath := ""
	slashIndex := strings.Index(address, "/")
	if slashIndex != -1 {
		basePath = address[slashIndex:]
		address = address[0:slashIndex]
	}
	return address, basePath
}

func resourcePath(baseAPIPath, path, query string) string {
	base := escapePath(strings.ReplaceAll(baseAPIPath+path, "//", "/"))
	if len(query) > 0 {
		return fmt.Sprintf("%s?%s", base, query)
	}
	return base
}

func escapePath(path string) string {
	spl := strings.Split(path, "/")
	encoded := []string{}
	for _, e := range spl {
		encoded = append(encoded, url.PathEscape(e))
	}
	return strings.Join(encoded, "/")
}

func authHeader(accessKeyID, privateKey, algo, algoAuthHeader, method, path, date string) (string, string, error) {
	meta, err := urlSafeMeta(accessKeyID, algoAuthHeader)
	if err != nil {
		return "", "", err
	}
	sig, msgToVerify, err := urlSafeSignature(privateKey, algo, algoAuthHeader, method, path, date)
	if err != nil {
		return "", "", err
	}
	return fmt.Sprintf("%s.%s", meta, sig), msgToVerify, nil
}

func urlSafeSignature(privateKey, algo, algoAuthHeader, method, path, date string) (string, string, error) {
	message := fmt.Sprintf(signPattern, method, applicationProtobuf, date, path, algoAuthHeader)
	log.Debugf("Message to sign: \n%s\n", message)
	signature, err := signWithAlgo(privateKey, algo, message)
	if err != nil {
		return "", "", err
	}
	return urlSafeBase64Encode(signature), message, nil
}

func signWithAlgo(privateKey string, algo string, message string) ([]byte, error) {
	var signature []byte
	if algo == ECDSA_ALGORITHM {
		block, _ := pem.Decode([]byte(privateKey))
		edcsaKey, err := x509.ParsePKCS8PrivateKey(block.Bytes)
		if err != nil {
			return nil, err
		}
		key, ok := edcsaKey.(*ecdsa.PrivateKey)
		if !ok {
			return nil, errors.New("Cannot cast key to ECDSA private key.")
		}
		digest := sha512.Sum512([]byte(message))
		signature, _ = ecdsa.SignASN1(rand.Reader, key, digest[:])
	} else {
		seed, err := base64.StdEncoding.DecodeString(privateKey)
		if err != nil {
			return nil, err
		}
		k := ed.NewKeyFromSeed(seed)
		signature = ed.Sign(k, []byte(message))
	}
	return signature, nil
}

func urlSafeMeta(accessKeyID string, algo string) (string, error) {
	b, err := json.Marshal(newMetastr(accessKeyID, algo))
	if err != nil {
		return "", err
	}
	return urlSafeBase64Encode(b), nil
}

func urlSafeBase64Encode(data []byte) string {
	return base64.URLEncoding.EncodeToString(data)
}

func formatDate() string {
	return time.Now().UTC().Format(layout)
}

func tokenHeader(privateKey string, algo string, tokenHolder *TokenHolder) (string, error) {
	token := tokenHolder.GetToken()
	tokenJsonObj, err := newTokenstr(token, privateKey, algo)
	if err != nil {
		return "", err
	}
	b, err := json.Marshal(tokenJsonObj)
	if err != nil {
		return "", err
	}
	return urlSafeBase64Encode(b), nil
}
