package proxy

import (
	"fmt"
	"os"
	"strings"

	log "github.com/sirupsen/logrus"
	"gopkg.in/ini.v1"
)

type CdpCredential struct {
	AccessKeyId             string
	PrivateKey              string
	SignAlgorithm           string
	SignAlgorithmAuthHeader string
}

const (
	ECDSA_ALGORITHM   = "ecdsa"
	ECDSA_HEADER      = "ecdsav1"
	ED25519_ALGORITHM = "ed25519"
	ED25519_HEADER    = "ed25519v1"
)

func GetCdpCredential(path string, profile string, cdpAccessKeyIdPropertyKey string, cdpPrivateKeyPropertyKey string, cdpSigningAlgorithmPropertyKey string) (*CdpCredential, error) {
	cdpAccessKeyIdFromEnv := os.Getenv("CDP_ACCESS_KEY_ID")
	cdpPrivateKeyFromEnv := os.Getenv("CDP_PRIVATE_KEY")
	cdpSignAlgorithm := os.Getenv("CDP_SIGN_ALGORITHM")
	if len(cdpAccessKeyIdFromEnv) > 0 && len(cdpPrivateKeyFromEnv) > 0 {
		log.Infof("CDP_ACCESS_KEY_ID and CDP_PRIVATE_KEY environment variables are set. Using those for cdp credentials. (CDP_SIGN_ALGORITHM is: %s)", cdpSignAlgorithm)
		signAlgo := getSignAlgorithm(cdpSignAlgorithm)
		return &CdpCredential{AccessKeyId: cdpAccessKeyIdFromEnv, PrivateKey: formatPrivateKey(cdpPrivateKeyFromEnv, signAlgo), SignAlgorithm: signAlgo, SignAlgorithmAuthHeader: getSignAlgorithmAuthHeader(cdpSignAlgorithm)}, nil
	}
	profileMap, err := RawParseConfigFile(path)
	if err != nil {
		return nil, err
	}

	profileData, ok := profileMap[profile]
	if !ok || strings.TrimSpace(profile) == "" {
		return nil, fmt.Errorf("cannot find profile %s in the credentials file", profile)
	}

	accessKeyId, ok := profileData[cdpAccessKeyIdPropertyKey]
	if !ok || strings.TrimSpace(accessKeyId) == "" {
		return nil, fmt.Errorf("cannot find %s in profile %s in the credentials file", cdpAccessKeyIdPropertyKey, profile)
	}

	privateKey, ok := profileData[cdpPrivateKeyPropertyKey]
	if !ok || strings.TrimSpace(privateKey) == "" {
		return nil, fmt.Errorf("cannot find %s in profile %s in the credentials file", cdpPrivateKeyPropertyKey, profile)
	}

	signAlgo, ok := profileData[cdpSigningAlgorithmPropertyKey]
	if !ok || strings.TrimSpace(signAlgo) == "" {
		log.Info("Signing algorithm is not set in the profile. The default will be used.")
	}
	signAlgo = getSignAlgorithm(signAlgo)
	return &CdpCredential{AccessKeyId: accessKeyId, PrivateKey: formatPrivateKey(privateKey, signAlgo), SignAlgorithm: signAlgo, SignAlgorithmAuthHeader: getSignAlgorithmAuthHeader(signAlgo)}, nil
}

func RawParseConfigFile(path string) (map[string]map[string]string, error) {
	properties := make(map[string]map[string]string)
	cfg, err := ini.InsensitiveLoad(path)
	if err != nil {
		return nil, err
	}
	for _, section := range cfg.Sections() {
		profile := make(map[string]string)
		properties[section.Name()] = profile
		for _, key := range section.Keys() {
			profile[key.Name()] = key.Value()
		}
	}
	return properties, nil
}

func getSignAlgorithm(algo string) string {
	if len(algo) > 0 && strings.ToLower(algo) == ECDSA_ALGORITHM {
		log.Infof("Picked signing algorithm: %v", ECDSA_ALGORITHM)
		return ECDSA_ALGORITHM
	}
	log.Infof("Picked signing algorithm: %v", ED25519_ALGORITHM)
	return ED25519_ALGORITHM
}

func getSignAlgorithmAuthHeader(algo string) string {
	if len(algo) > 0 && strings.ToLower(algo) == ECDSA_ALGORITHM {
		return ECDSA_HEADER
	}
	return ED25519_HEADER
}

func formatPrivateKey(privateKey, algo string) string {
	if algo == ECDSA_ALGORITHM && len(privateKey) > 0 {
		return strings.ReplaceAll(privateKey, "\\n", "\n")
	}
	return privateKey
}
