package proxy

import (
	"io"
	"io/ioutil"
	"os"
	"path/filepath"
	"time"

	"github.com/sirupsen/logrus"
	log "github.com/sirupsen/logrus"
	"gopkg.in/natefinch/lumberjack.v2"
)

func InitLogger(config *ReverseProxyConfig) {
	logLevel, err := log.ParseLevel(config.LoggerConfig.Level)
	if err != nil {
		panic(err)
	}
	log.SetLevel(logLevel)
	if config.LoggerConfig.File {
		fileName := "cdp-request-signer.log"
		logDir := "/var/log/cdp-request-signer"
		logFileName := filepath.Join(logDir, fileName)
		if _, err := os.Stat(logDir); os.IsNotExist(err) {
			logFileName = fileName
		}
		rotateFileHook, err := NewRotateFileHook(RotateFileConfig{
			Filename:   logFileName,
			MaxSize:    20,
			MaxBackups: 5,
			MaxAge:     28,
			Level:      logLevel,
			Formatter: &log.JSONFormatter{
				TimestampFormat: time.RFC3339,
				FieldMap: log.FieldMap{
					log.FieldKeyMsg: "message",
				},
			},
		})
		if err != nil {
			log.Fatalf("Failed to initialize file rotate hook: %v", err)
		}
		log.AddHook(rotateFileHook)
		log.SetOutput(ioutil.Discard)
	} else {
		log.SetFormatter(&log.JSONFormatter{
			TimestampFormat: time.RFC822,
			FieldMap: log.FieldMap{
				log.FieldKeyMsg: "message",
			},
		})
	}
}

type RotateFileConfig struct {
	Filename   string
	MaxSize    int
	MaxBackups int
	MaxAge     int
	Compress   bool
	Level      logrus.Level
	Formatter  logrus.Formatter
}

type RotateFileHook struct {
	Config    RotateFileConfig
	logWriter io.Writer
}

func NewRotateFileHook(config RotateFileConfig) (logrus.Hook, error) {

	hook := RotateFileHook{
		Config: config,
	}
	hook.logWriter = &lumberjack.Logger{
		Filename:   config.Filename,
		MaxSize:    config.MaxSize,
		MaxBackups: config.MaxBackups,
		MaxAge:     config.MaxAge,
		Compress:   config.Compress,
	}

	return &hook, nil
}

func (hook *RotateFileHook) Levels() []logrus.Level {
	return logrus.AllLevels[:hook.Config.Level+1]
}

func (hook *RotateFileHook) Fire(entry *logrus.Entry) (err error) {
	b, err := hook.Config.Formatter.Format(entry)
	if err != nil {
		return err
	}
	hook.logWriter.Write(b)
	return nil
}
