package proxy

import (
	"path/filepath"
	"strings"

	"github.com/spf13/viper"
)

type ReverseProxyConfig struct {
	Endpoint     string
	Port         int32
	BasePath     string
	Credential   *Credential
	Sign         *Sign
	Proxy        *Proxy
	TLS          *TLS
	Auth         *Auth
	Client       *Client
	LoggerConfig *LoggerConfig
}

type Client struct {
	Dump    bool
	Timeout int64
}

type Auth struct {
	Username string
	Password string
}

type Credential struct {
	File              string
	Profile           string
	AccessKeyName     string
	SecretKeyName     string
	SignAlgorithmName string
}

type Sign struct {
	AuthHeaderName        string
	DateHeaderName        string
	MsgToVerifyName       string
	TokenHeaderName       string
	RequestIdHeaderName   string
	ResourceCrnHeader     string
	ResourceCrnHeaderName string
	InstanceHeader        string
	InstanceHeaderName    string
	TokenValidityMin      int64
	TokenDateFile         string
	UseTokenHeader        bool
	UseMsgHeader          bool
	UseRequestIdHeader    bool
}

type Proxy struct {
	HttpsProxy string
	HttpProxy  string
	NoProxy    string
}

type TLS struct {
	Key  string
	Cert string
}

type LoggerConfig struct {
	Level string
	File  bool
}

func CreateReverseProxyConfig(configLocation string) (*ReverseProxyConfig, error) {
	v := viper.GetViper()
	file := filepath.Base(configLocation)
	parentDir := filepath.Dir(configLocation)
	v.SetConfigName(fileNameWithoutExtTrimSuffix(file))
	v.SetConfigType("yaml")
	v.AddConfigPath(parentDir)
	v.AutomaticEnv()
	v.SetEnvKeyReplacer(strings.NewReplacer(`.`, `_`))
	err := v.ReadInConfig()
	if err != nil {
		return nil, err
	}
	config := ReverseProxyConfig{}
	config.Endpoint = v.GetString("endpoint")
	v.SetDefault("port", 61095)
	config.Port = v.GetInt32("port")
	config.BasePath = v.GetString("base_path")
	v.SetDefault("credential.profile", "default")
	v.SetDefault("credential.access_key_name", "cdp_access_key_id")
	v.SetDefault("credential.secret_key_name", "cdp_private_key")
	v.SetDefault("credential.sign_algorithm_name", "cdp_sign_algorithm")
	credential := Credential{
		File:              v.GetString("credential.file"),
		Profile:           v.GetString("credential.profile"),
		AccessKeyName:     v.GetString("credential.access_key_name"),
		SecretKeyName:     v.GetString("credential.secret_key_name"),
		SignAlgorithmName: v.GetString("credential.sign_algorithm_name"),
	}
	config.Credential = &credential
	proxy := Proxy{
		HttpsProxy: v.GetString("proxy.https_proxy"),
		HttpProxy:  v.GetString("proxy.http_proxy"),
		NoProxy:    v.GetString("proxy.no_proxy"),
	}
	config.Proxy = &proxy
	tls := TLS{
		Key:  v.GetString("tls.key"),
		Cert: v.GetString("tls.cert"),
	}
	config.TLS = &tls
	auth := Auth{
		Username: v.GetString("auth.username"),
		Password: v.GetString("auth.password"),
	}
	config.Auth = &auth
	v.SetDefault("client.timeout", 60)
	client := Client{
		Dump:    v.GetBool("client.dump"),
		Timeout: v.GetInt64("client.timeout"),
	}
	config.Client = &client
	v.SetDefault("logger.level", "info")
	loggerConfig := LoggerConfig{
		File:  v.GetBool("logger.file"),
		Level: v.GetString("logger.level"),
	}
	config.LoggerConfig = &loggerConfig
	v.SetDefault("sign.auth_header_name", "x-altus-auth")
	v.SetDefault("sign.date_header_name", "x-altus-date")
	v.SetDefault("sign.msg_header_name", "x-altus-msg")
	v.SetDefault("sign.token_header_name", "x-altus-token")
	v.SetDefault("sign.request_id_header_name", "x-cdp-request-id")
	v.SetDefault("sign.resource_crn_header_name", "x-cdp-resource-crn")
	v.SetDefault("sign.instance_header_name", "x-cdp-instance")
	v.SetDefault("sign.token_validity_min", 60)
	v.SetDefault("sign.token_date_file", "token_exp")
	v.SetDefault("sign.use_token_header", true)
	v.SetDefault("sign.use_msg_header", true)
	v.SetDefault("sign.use_request_id_header", true)
	sign := Sign{
		AuthHeaderName:        v.GetString("sign.auth_header_name"),
		DateHeaderName:        v.GetString("sign.date_header_name"),
		MsgToVerifyName:       v.GetString("sign.msg_header_name"),
		TokenHeaderName:       v.GetString("sign.token_header_name"),
		RequestIdHeaderName:   v.GetString("sign.request_id_header_name"),
		ResourceCrnHeaderName: v.GetString("sign.resource_crn_header_name"),
		ResourceCrnHeader:     v.GetString("sign.resource_crn_header"),
		InstanceHeaderName:    v.GetString("sign.instance_header_name"),
		InstanceHeader:        v.GetString("sign.instance_header"),
		TokenValidityMin:      MaxValue(v.GetInt64("sign.token_validity_min"), 60),
		TokenDateFile:         v.GetString("sign.token_date_file"),
		UseTokenHeader:        v.GetBool("sign.use_token_header"),
		UseMsgHeader:          v.GetBool("sign.use_msg_header"),
		UseRequestIdHeader:    v.GetBool("sign.use_request_id_header"),
	}
	config.Sign = &sign
	return &config, nil
}

func MaxValue(value int64, maxValue int64) int64 {
	if value > maxValue {
		return maxValue
	}
	return value
}

func fileNameWithoutExtTrimSuffix(fileName string) string {
	return strings.TrimSuffix(fileName, filepath.Ext(fileName))
}
