package proxy

import (
	"fmt"
	"io"
	"net/http"
	"net/http/httputil"
	"net/url"
	"strings"

	log "github.com/sirupsen/logrus"
)

// ProxyRequestHandler handles the http request using proxy
func ProxyRequestHandler(client *http.Client, config *ReverseProxyConfig, cdpCredential *CdpCredential, url *url.URL, tokenHolder *TokenHolder) func(http.ResponseWriter, *http.Request) {
	return func(w http.ResponseWriter, r *http.Request) {
		if len(config.Auth.Username) > 0 {
			u, p, ok := r.BasicAuth()
			if !ok {
				log.Error("Basic auth is missing")
				w.WriteHeader(401)
				return
			}
			if u != config.Auth.Username {
				log.Error(fmt.Sprintf("Username provided is not correct: %s\n", u))
				w.WriteHeader(401)
				return
			}
			if p != config.Auth.Password {
				log.Error(fmt.Sprintf("Password provided is not correct for %s\n", u))
				w.WriteHeader(401)
				return
			}
		}
		path := r.URL.Path
		urlToCall := strings.TrimSuffix(fmt.Sprintf("%s", singleJoiningSlash(url.String(), path)), "/")
		newReq, err := http.NewRequest(r.Method, urlToCall, r.Body)
		if err != nil {
			log.Error(fmt.Sprintf("Error during request processing: %v", err))
			ResponseCounter.WithLabelValues(fmt.Sprint(400)).Inc()
			w.WriteHeader(400)
			return
		}
		requestSignError := RequestSign(config, cdpCredential, tokenHolder, newReq, config.BasePath)
		if requestSignError != nil {
			log.Error(fmt.Sprintf("Error during request signing: %v", requestSignError))
			ResponseCounter.WithLabelValues(fmt.Sprint(400)).Inc()
			w.Write([]byte("Error during request signing."))
			w.WriteHeader(400)
			return
		}
		if config.Client.Dump {
			reqDump, err := httputil.DumpRequest(newReq, false)
			if err != nil {
				log.Error(err)
			} else {
				log.Info(fmt.Sprintf("DUMP REQUEST:\n%s", string(reqDump)))
			}
		}
		for k, vv := range r.Header {
			for _, v := range vv {
				if len(config.Auth.Username) > 0 && k == "Authorization" {
					continue
				}
				if k != "Host" && k != "X-Forwarded-For" && k != "Content-Length" {
					newReq.Header.Set(k, v)
				}
			}
		}
		resp, respErr := client.Do(newReq)
		if respErr != nil {
			log.Error(fmt.Sprintf("Response error: %v", respErr))
			ResponseCounter.WithLabelValues(fmt.Sprint(444)).Inc()
			w.WriteHeader(444)
			return
		}
		if resp == nil {
			ResponseCounter.WithLabelValues(fmt.Sprint(444)).Inc()
			w.WriteHeader(444)
			return
		}
		ResponseCounter.WithLabelValues(fmt.Sprint(resp.StatusCode)).Inc()
		copyHeader(w.Header(), resp.Header)
		w.WriteHeader(resp.StatusCode)
		io.Copy(w, resp.Body)
		if config.Client.Dump || (resp.StatusCode > 399 && resp.StatusCode < 600) {
			resDump, err := httputil.DumpResponse(resp, false)
			if err != nil {
				log.Error(err)
			} else {
				log.Info(fmt.Sprintf("DUMP RESPONSE:\n%s", string(resDump)))
			}
		}
		resp.Body.Close()
	}
}

// copyHeader and singleJoiningSlash are copy from "/net/http/httputil/reverseproxy.go"
func copyHeader(dst, src http.Header) {
	for k, vv := range src {
		for _, v := range vv {
			dst.Add(k, v)
		}
	}
}

func singleJoiningSlash(a, b string) string {
	aslash := strings.HasSuffix(a, "/")
	bslash := strings.HasPrefix(b, "/")
	switch {
	case aslash && bslash:
		return a + b[1:]
	case !aslash && !bslash:
		if b == "" {
			return a
		}
		return a + "/" + b
	}
	return a + b
}
