package proxy

import (
	"crypto/tls"
	"encoding/json"
	"fmt"
	"net"
	"net/http"
	"net/url"
	"os"
	"time"

	log "github.com/sirupsen/logrus"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promauto"
	"github.com/prometheus/client_golang/prometheus/promhttp"
)

var (
	ResponseCounter = promauto.NewCounterVec(prometheus.CounterOpts{
		Name: "request_signer_remote_write_http_requests_total",
		Help: "HTTP remote response of the request signer",
	}, []string{"status_code"})
)

func Serve(configLocation string) {
	requestSignerProxyConf, err := CreateReverseProxyConfig(configLocation)
	if err != nil {
		fmt.Println(err)
		os.Exit(1)
	}
	InitLogger(requestSignerProxyConf)
	confJson, jsonWriteErr := json.Marshal(requestSignerProxyConf)
	if jsonWriteErr != nil {
		fmt.Println(jsonWriteErr)
		os.Exit(1)
	}
	log.Printf("Application config: %v", string(confJson))
	c := requestSignerProxyConf.Credential
	credential, cerr := GetCdpCredential(c.File, c.Profile, c.AccessKeyName, c.SecretKeyName, c.SignAlgorithmName)
	if cerr != nil {
		panic(cerr)
	}
	url, err := url.Parse(requestSignerProxyConf.Endpoint)
	if err != nil {
		panic(err)
	}
	hasHttpsProxy := len(requestSignerProxyConf.Proxy.HttpsProxy) > 0
	hasHttpProxy := len(requestSignerProxyConf.Proxy.HttpProxy) > 0
	hasNoProxy := len(requestSignerProxyConf.Proxy.NoProxy) > 0
	if hasHttpsProxy || hasHttpProxy {
		log.Info("Proxy settings found in the configuration. Initialize proxy settings ...")
		if hasHttpsProxy {
			os.Setenv("HTTPS_PROXY", requestSignerProxyConf.Proxy.HttpsProxy)
			os.Setenv("https_proxy", requestSignerProxyConf.Proxy.HttpsProxy)
		}
		if hasHttpProxy {
			os.Setenv("HTTP_PROXY", requestSignerProxyConf.Proxy.HttpProxy)
			os.Setenv("http_proxy", requestSignerProxyConf.Proxy.HttpProxy)
		}
		if hasNoProxy {
			os.Setenv("NO_PROXY", requestSignerProxyConf.Proxy.NoProxy)
			os.Setenv("no_proxy", requestSignerProxyConf.Proxy.NoProxy)
		}
	}
	timeout := time.Duration(requestSignerProxyConf.Client.Timeout) * time.Second
	var transport http.RoundTripper = &http.Transport{
		Proxy:           http.ProxyFromEnvironment,
		TLSClientConfig: &tls.Config{InsecureSkipVerify: true},
		DialContext: (&net.Dialer{
			Timeout: timeout,
		}).DialContext,
	}
	tokenHolder := InitTokenHolder(requestSignerProxyConf, credential)
	client := &http.Client{Transport: transport}
	http.Handle("/metrics", promhttp.Handler())
	http.HandleFunc("/", ProxyRequestHandler(client, requestSignerProxyConf, credential, url, tokenHolder))
	if len(requestSignerProxyConf.TLS.Key) > 0 && len(requestSignerProxyConf.TLS.Cert) > 0 {
		log.Infof("Start application (https) on port: %d", requestSignerProxyConf.Port)
		log.Fatal(http.ListenAndServeTLS(fmt.Sprintf(":%d", requestSignerProxyConf.Port), requestSignerProxyConf.TLS.Cert, requestSignerProxyConf.TLS.Key, nil))
	} else {
		log.Infof("Start application (http) on port: %d", requestSignerProxyConf.Port)
		log.Fatal(http.ListenAndServe(fmt.Sprintf(":%d", requestSignerProxyConf.Port), nil))
	}
}
