package proxy

import (
	"errors"
	"os"
	"strconv"
	"time"

	log "github.com/sirupsen/logrus"
)

type TokenHolder struct {
	Token  *Token
	Config *Sign
}

type Token struct {
	AccessKeyId string
	ExpireAt    *time.Time
}

func InitTokenHolder(config *ReverseProxyConfig, cdpCredential *CdpCredential) *TokenHolder {
	if !config.Sign.UseTokenHeader {
		return nil
	}
	token := Token{AccessKeyId: cdpCredential.AccessKeyId}
	if _, err := os.Stat(config.Sign.TokenDateFile); err == nil {
		content, err := os.ReadFile(config.Sign.TokenDateFile)
		if err != nil {
			panic(err)
		}
		unixTimestamp, err := strconv.ParseInt(string(content), 10, 64)
		if err != nil {
			panic(err)
		}
		if IsTokenStillValid(unixTimestamp, config.Sign.TokenValidityMin) {
			log.Infof("Token validity timestamp (%s) is in the right range (+%d min), keep using it.", string(content), config.Sign.TokenValidityMin)
			expireAt := time.Unix(unixTimestamp, 0)
			token.ExpireAt = &expireAt
		} else {
			log.Infof("Token validity timestamp (%s) is not in the right range (+%d min), regenerating it...", string(content), config.Sign.TokenValidityMin)
			newT, newTimestamp, err := createNewTokenDateFile(config.Sign)
			if err != nil {
				panic(err)
			}
			token.ExpireAt = newT
			log.Infof("Generated new token validity file with timestamp: %s (validity: %d min)", newTimestamp, config.Sign.TokenValidityMin)
		}
	} else if errors.Is(err, os.ErrNotExist) {
		log.Infof("Token validity file does not exist. Creating it ...")
		newT, newTimestamp, err := createNewTokenDateFile(config.Sign)
		if err != nil {
			panic(err)
		}
		token.ExpireAt = newT
		log.Infof("Generated new token validity file with timestamp: %s (validity: %d min)", newTimestamp, config.Sign.TokenValidityMin)
	} else {
		panic(err)
	}
	return &TokenHolder{Token: &token, Config: config.Sign}
}

func createNewTokenDateFile(signConfig *Sign) (*time.Time, string, error) {
	f, err := os.Create(signConfig.TokenDateFile)
	tt := time.Now().UTC()
	timeWithoutSeconds := time.Date(tt.Year(), tt.Month(), tt.Day(), tt.Hour(), tt.Minute(), 0, 0, tt.Location())
	timePassed := tt.Minute() % int(signConfig.TokenValidityMin)
	minToAdd := signConfig.TokenValidityMin - int64(timePassed)
	timein := timeWithoutSeconds.Add(time.Minute * time.Duration(minToAdd)).UTC()
	if err != nil {
		return nil, "", err
	}
	strTimestamp := strconv.FormatInt(timein.Unix(), 10)
	_, err = f.WriteString(strTimestamp)
	if err != nil {
		return nil, "", err
	}
	defer f.Close()
	return &timein, strTimestamp, nil
}

func IsTokenStillValid(unixTimestamp int64, tokenValidityMin int64) bool {
	timeNow := time.Now().UTC()
	timePlus := timeNow.Add(time.Minute * time.Duration(tokenValidityMin))
	return unixTimestamp > timeNow.Unix() && unixTimestamp < timePlus.Unix()
}

func (h *TokenHolder) GetToken() *Token {
	token := h.Token
	unixTimestamp := token.ExpireAt.Unix()
	if IsTokenStillValid(unixTimestamp, h.Config.TokenValidityMin) {
		return token
	} else {
		newToken := Token{AccessKeyId: token.AccessKeyId}
		newT, newTimestamp, err := createNewTokenDateFile(h.Config)
		if err != nil {
			panic(err)
		}
		log.Infof("Generated new token validity file with timestamp: %s (validity: %d min)", newTimestamp, h.Config.TokenValidityMin)
		newToken.ExpireAt = newT
		h.Token = &newToken
		return &newToken
	}
}
