#!/usr/bin/env python

import click
import sys
import os
from fluent import event

from metricscollector import metrics_collector
from filecollector import filecollector as collector
from nodestatus import monitor as nodestatus_monitor, salt_report as nodestatus_salt_report
from databus import databus as dbus
from csh import csh as cloudera_support_hub_client
from storage import s3 as s3_client, azure_blob, gcs as gcs_client
from doctor import ccm as doctor_ccm
from doctor import commands as doctor_commands
from doctor import fluentd as doctor_fluentd
from doctor import network as doctor_network
from doctor import service as doctor_service
from doctor import recipe as doctor_recipe
from doctor import system as doctor_system
from doctor import salt as doctor_salt
from doctor import parcel as doctor_parcel
from doctor import scm as doctor_scm
from doctor import ipa as doctor_ipa
from doctor import metering as doctor_metering
from doctor import monitoring as doctor_monitoring
from doctor import server as doctor_server
from utils import cleaner, connection, timeout
from utils import version as cdp_telemetry_version

# setup global logging here
from utils import telemetry_logger
telemetry_logger.setup_logger()

__version__=cdp_telemetry_version.get_binary_version()

class NotRequiredIf(click.Option):
    def __init__(self, *args, **kwargs):
        self.not_required_if = kwargs.pop('not_required_if')
        assert self.not_required_if, "'not_required_if' parameter required"
        kwargs['help'] = (kwargs.get('help', '') +
            ' NOTE: This argument is mutually exclusive with %s' %
            self.not_required_if
        ).strip()
        super(NotRequiredIf, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        we_are_present = self.name in opts
        other_present = self.not_required_if in opts
        if other_present:
            if we_are_present:
                raise click.UsageError(
                    "Illegal usage: `%s` is mutually exclusive with `%s`" % (
                        self.name, self.not_required_if))
            else:
                self.prompt = None
        if not other_present and not we_are_present:
            raise click.UsageError(
                    "Illegal usage: at least on of these options are required: `%s`,`%s`" % (
                        self.name, self.not_required_if))

        return super(NotRequiredIf, self).handle_parse_result(
            ctx, opts, args)

@click.group()
@click.version_option(__version__, message='%(version)s')
def main():
    pass

@main.group()
def filecollector():
    """File collector related commands"""

@filecollector.command('collect')
@click.option('--config','-c', help='Path to filecollector configuration', type=click.Path(exists=True), required=True)
@click.option('--label','-l', help='Label for filtering file type for collection (multiple)', multiple=True)
@click.option('--start-time','-s', help='Start (last modified) datestamp (epoh unix format) for the monitored logs', type=float)
@click.option('--end-time','-e', help='End (creation) datestamp (epoh unix format) for the monitored logs (epoh unix format)', type=float)
def filecollector_collect(config, label, start_time, end_time):
    """Collect files based on a configuration"""
    labels=','.join(label) if label else None
    collector.run(config, labels, start_time, end_time)

@main.group()
def metricscollector():
    """Metrics collector related commands"""

@metricscollector.command('start')
@click.option('--config','-c', help='Path to metrics collector configuration', type=click.Path(exists=True), required=True)
def metrics_collector_start(config):
    """Start metrics collection"""
    metrics_collector.run(config)

@main.group()
def nodestatus():
    """Node status monitor for CDP VM nodes"""

@nodestatus.command("server")
@click.option('--config','-c', help='Path to node status monitor configuration', type=click.Path(exists=True), required=True, default="/opt/cdp-telemetry/conf/nodestatus-monitor.yaml")
def nodestatus_monitor_start(config):
    """Start CDP node status monitor server"""
    nodestatus_monitor.start(config)

@nodestatus.command("update")
@click.option('--config','-c', help='Path to node status monitor configuration', type=click.Path(exists=True), required=True, default="/opt/cdp-telemetry/conf/nodestatus-monitor.yaml")
def nodestatus_update(config):
    """Executes CDP node status commands manually"""
    nodestatus_monitor.update(config)

@nodestatus.command("collect")
@click.option('--directory', '-d', help='Base folder where the salt minion results are collected.', type=click.Path(exists=True), required=True, default="/var/lib/cdp-nodestatus")
@click.option('--filter','-f', help='Filter to run check commands only on specific resources. (multiple)', type=click.Choice(['services', 'metering', 'network', "system-metrics"]), multiple=True)
@click.option('--databus-url', '-u', help='DataBus URL for network checks. (default: https://dbusapi.us-west-1.altus.cloudera.com)', default='https://dbusapi.us-west-1.altus.cloudera.com')
@click.option('--databus-s3-url', help='DataBus S3 URL for network checks.')
@click.option('--cloud-provider', '-c', type=click.Choice(['AWS', 'AZURE', 'GCP']), help='Cloud provider for the running instances (if any).')
@click.option('--pretty-print', '-p', help='Pretty print (only for JSON format)', is_flag=True)
def nodestatus_collect(directory, filter, databus_url, databus_s3_url, cloud_provider, pretty_print):
    """Collect CDP doctor outputs through salt (metering/services/network data)"""
    nodestatus_salt_report.generate_salt_healthcheck(directory, filter, databus_url, databus_s3_url, cloud_provider, pretty=pretty_print)

@nodestatus.group()
def salt_ping():
    """Checking salt services for nodestatus usage"""

@salt_ping.command("local-master")
@click.option('--directory', '-d', help='Base folder where the salt minion results are collected.', type=click.Path(exists=True), required=True, default="/var/lib/cdp-nodestatus")
@click.option('--pretty-print', '-p', help='Pretty print (only for JSON format)', is_flag=True)
def nodestatus_local_master(directory, pretty_print):
    """Check of local salt-master/api/boostrap services"""
    nodestatus_salt_report.generate_local_salt_master_healthcheck(directory, pretty=pretty_print)

@salt_ping.command("minions")
@click.option('--directory', '-d', help='Base folder where the salt minion results are collected.', type=click.Path(exists=True), required=True, default="/var/lib/cdp-nodestatus")
@click.option('--pretty-print', '-p', help='Pretty print (only for JSON format)', is_flag=True)
def nodestatus_salt_minions(directory, pretty_print):
    """Test salt minions"""
    nodestatus_salt_report.generate_salt_minions_healthcheck(directory, pretty=pretty_print)

@salt_ping.command("all")
@click.option('--directory', '-d', help='Base folder where the salt minion results are collected.', type=click.Path(exists=True), required=True, default="/var/lib/cdp-nodestatus")
@click.option('--pretty-print', '-p', help='Pretty print (only for JSON format)', is_flag=True)
def nodestatus_salt_ping_all(directory, pretty_print):
    """Check both local salt master services and minions."""
    nodestatus_salt_report.generate_salt_service_ping_healthcheck(directory, pretty=pretty_print)

@main.group()
def storage():
    """Storage related commands"""

@storage.group('s3')
def s3():
    """S3 related commands"""

@s3.command('upload')
@click.option('--bucket','-b', help='S3 bucket', required=True)
@click.option('--file','-f', help='Path to the file that will be uploaded to s3', type=click.Path(exists=True), cls=NotRequiredIf, not_required_if='paths')
@click.option('--paths','-p', help='Wildcard path(s) to the file(s) that will be uploaded to s3', cls=NotRequiredIf, not_required_if='file')
@click.option('--location','-l', help='S3 base location for the upload', required=True)
@click.option('--region','-r', help='AWS region for S3 endpoint')
@click.option('--encryption', '-e', help='Use encryption (tries: AES256, KMS)', is_flag=True)
def s3_upload(bucket, file, paths, location, region, encryption):
    """Upload file to S3"""
    s3_client.upload(bucket, file, paths, location, region, encryption)

@s3.command('pull')
@click.option('--bucket','-b', help='S3 bucket', required=True)
@click.option('--location', '-l', help='Location folder to download the files', type=click.Path(exists=True))
@click.option('--prefix','-p', help='S3 folder prefix.', required=True)
@click.option('--region','-r', help='AWS region for S3 endpoint')
def s3_pull(bucket, location, prefix, region):
    """Pull file(s) from S3"""
    s3_client.pull(bucket, prefix, location, region)

@s3.command('enc')
@click.option('--bucket','-b', help='S3 bucket', required=True)
@click.option('--region','-r', help='AWS region for S3 endpoint')
def check_ecnryption(bucket, region):
    """Check bucket encryption"""
    s3_client.check_encryption(bucket, region)

@s3.command('get-lifecycle', hidden=True)
@click.option('--bucket','-b', help='S3 bucket', required=True)
@click.option('--region','-r', help='AWS region for S3 endpoint')
def get_lifecycle(bucket, region):
    """Get bucket lifecycle configuration"""
    s3_client.get_lifecycle(bucket, region)

@storage.group()    
def abfs():
    """AdlsV2 related commands"""

@abfs.command('upload')
@click.option('--account','-a', help='AdlsV2 storage account name', required=True)
@click.option('--container','-c', help='Container (filesystem) for AdlsV2 storage account', required=True)
@click.option('--file','-f', help='Path to the file that will be uploaded to AdlsV2', type=click.Path(exists=True), cls=NotRequiredIf, not_required_if='paths')
@click.option('--paths','-p', help='Wildcard path(s) to the file(s) that will be uploaded to AdlsV2', cls=NotRequiredIf, not_required_if='file')
@click.option('--location','-l', help='AdlsV2 base location for the upload', required=True)
@click.option('--use-msi', '-m', help='Force to use Azure managed identity for authentication', is_flag=True)
@click.option('--resource', '-r', help='Storage resource url', default='https://storage.azure.com/', show_default=True)
@click.option('--verbose', '-v', help='Verbose output', is_flag=True)
def abfs_upload(account, container, file, paths, location, use_msi, resource, verbose):
    """Upload file to AdlsV2"""
    azure_blob.upload(account, container, file, paths, location, use_msi, verbose, resource)

@storage.group()    
def gcs():
    """Google Cloud Storage related commands"""

@gcs.command('upload')
@click.option('--bucket','-b', help='Google Cloud Storage bucket', required=True)
@click.option('--file','-f', help='Path to the file that will be uploaded to Google Cloud Storage', type=click.Path(exists=True), cls=NotRequiredIf, not_required_if='paths')
@click.option('--paths','-p', help='Wildcard path(s) to the file(s) that will be uploaded to Google Cloud Storage', cls=NotRequiredIf, not_required_if='file')
@click.option('--location','-l', help='Google Cloud Storage base location for the upload', required=True)
@click.option('--credential', '-c', help='Google Cloud JSON credential file', default=lambda: os.environ.get('GOOGLE_APPLICATION_CREDENTIALS'), type=click.Path(exists=True))
def gcs_upload(bucket, file, paths, location, credential):
    """Upload file to Google Cloud Storage"""
    gcs_client.upload(bucket, file, paths, location, credential)

@main.group()
def databus():
    """Databus related commands"""

@databus.command('upload')
@click.option('--credential-file','-c', help='Credential file that continas databus credentials', type=click.Path(exists=True), required=True)
@click.option('--file','-f', help='Path to the file that will be uploaded to DataBus API', type=click.Path(exists=True), cls=NotRequiredIf, not_required_if='paths')
@click.option('--paths','-p', help='Wildcard path(s) to the file(s) that will be uploaded to DataBus API', cls=NotRequiredIf, not_required_if='file')
@click.option('--url', '-u', help='Databus endpoint url', required=True)
@click.option('--header', help='Header (key-value pairs in <k>:<v> format)for Databus record', multiple=True)
@click.option('--stream', '-s', help='Stream name for Databus record', required=True)
@click.option('--partition-key', '-k', help='Partition key for Databus record (hostname is used if not set)')
@click.option('--extra-headers-file','-e', help='Yaml formatted config file that contains key-value pairs for additional headers', type=click.Path(exists=True))
def databus_upload(credential_file, file, paths, url, header, stream, partition_key, extra_headers_file):
    """Upload file(s) through Databus API"""
    dbus.upload(credential_file, url, file, paths, header, stream, partition_key, extra_headers_file=extra_headers_file)

@databus.command('put-record')
@click.option('--credential-file','-c', help='Credential file that continas databus credentials', type=click.Path(exists=True), required=True)
@click.option('--value', '-v', help='Databus record content', cls=NotRequiredIf, not_required_if='file')
@click.option('--file', '-f', help='File that contains the raw data for Databus record', type=click.Path(exists=True), cls=NotRequiredIf, not_required_if='value')
@click.option('--url', '-u', help='Databus endpoint url', required=True)
@click.option('--header', help='Header (key-value pairs in <k>:<v> format)for Databus record', multiple=True)
@click.option('--stream', '-s', help='Stream name for Databus record', required=True)
@click.option('--partition-key', '-k', help='Partition key for Databus record (hostname is used if not set)')
def databus_put_record(credential_file, value, file, url, header, stream, partition_key):
    """Upload a record through Databus API"""
    dbus.put_record(credential_file, url, value, file, header, stream, partition_key)

@databus.command('test-put-record')
@click.option('--credential-file','-c', help='Credential file that continas databus credentials', type=click.Path(exists=True), required=True, multiple=True)
@click.option('--url', '-u', help='Databus endpoint url', required=True)
def databus_test_put_record(credential_file, url):
    """Test put record call through Databus API"""
    dbus.test_put_record(credential_file, url)

@main.group("csh", hidden=True)
def csh():
    """Cloudera Support Hub operations"""

@csh.command("upload")
@click.option('--file','-f', help='Path to the file that will be uploaded to CSH', type=click.Path(exists=True), cls=NotRequiredIf, not_required_if='paths')
@click.option('--paths','-p', help='Wildcard path(s) to the file(s) that will be uploaded to CSH', cls=NotRequiredIf, not_required_if='file')
def upload(file, paths):
    """Upload support bundle file(s) to CSH"""
    cloudera_support_hub_client.upload(file, paths)

@main.group("doctor")
def doctor():
    """System diagnostics related operations (CDP specific)"""

@doctor.group("ccm")
def ccm():
    """Troubleshooting CCM related resources"""

@ccm.command("report")
@click.option('--output', '-o', help='Output report filename (destination) without compression extenison (zip)', default='ccm_debugging_report')
@click.option('--skip-tcpdump', '-s', help='Skip collecting tcp dump.', is_flag=True)
def ccm_report(output, skip_tcpdump):
    """Generating a report about CCM troubleshooting results"""
    doctor_ccm.report(output, skip_tcpdump)

@ccm.command("status")
def ccm_status():
    """Check CCM network and service status"""
    doctor_ccm.status()

@doctor.group("network")
def network():
    """Analyze network on CDP host"""

@network.command("status")
@click.option('--format', type=click.Choice(['table', 'json']), help='Command output format. (default: table)', default='table')
@click.option('--skip-scan-neighbours', '-s', help='Skip scanning the neighbours. DEPRECATED', is_flag=True)
@click.option('--scan-neighbours', '-s', help='Force ARP scanning of the neighbours.', is_flag=False)
@click.option('--include-neighbours', '-i', help='Include neighbours in the results (for JSON format).', is_flag=True)
@click.option('--databus-url', '-d', help='DataBus URL for network checks. (default: https://dbusapi.us-west-1.altus.cloudera.com)', default='https://dbusapi.us-west-1.altus.cloudera.com')
@click.option('--databus-s3-url', help='DataBus S3 URL for network checks.')
@click.option('--cloud-provider', '-c', type=click.Choice(['AWS', 'AZURE', 'GCP']), help='Cloud provider for the running (local) instance.')
@click.option('--pretty-print', '-p', help='Pretty print (only for JSON format)', is_flag=True)
def network_analyze(format, scan_neighbours, skip_scan_neighbours, include_neighbours, databus_url, databus_s3_url, cloud_provider, pretty_print):
    """Summary of available network interfaces, Dns resolves, CCM info"""
    timeout.timeout(doctor_network.main, format, scan_neighbours, skip_scan_neighbours, include_neighbours, databus_url, databus_s3_url, cloud_provider=cloud_provider, pretty=pretty_print, context="Network status check", exit_on_timeout=True)

@network.command("addresses")
@click.option('--format', type=click.Choice(['table', 'json']), help='Command output format. (default: table)', default='table')
@click.option('--databus-url', '-d', help='DataBus URL for network checks. (default: https://dbusapi.us-west-1.altus.cloudera.com)', default='https://dbusapi.us-west-1.altus.cloudera.com')
@click.option('--databus-s3-url', help='DataBus S3 URL for network checks.')
@click.option('--cloud-provider', '-c', type=click.Choice(['AWS', 'AZURE', 'GCP']), help='Cloud provider for the running (local) instance.')
@click.option('--pretty-print', '-p', help='Pretty print (only for JSON format)', is_flag=True)
def network_adresses(format, databus_url, databus_s3_url, cloud_provider, pretty_print):
    "Get the addresses that needs to be checked"
    doctor_network.get_addresses(format, databus_url, databus_s3_url, cloud_provider, pretty_print)

@doctor.group("recipe")
def recipe():
    """Recipe related diagnostics"""

@recipe.command("results")
def check_recipes():
    """Fetch the result of all executed recipes on this node"""
    doctor_recipe.fetch_results()

@recipe.command("describe")
@click.option('--name','-n', help='Check detailed execution log of recipe. eg. "post-cloudera-manager-start/example-recipe.log"', required=True)
def describe_recipe(name):
    """Fetch STDOUT+STDERR of recipe execution"""
    doctor_recipe.describe(name)

@doctor.group("ipa")
def ipa():
    """IPA related operations"""

@ipa.command("status")
@click.option('--format', type=click.Choice(['table', 'json']), help='Command output format. (default: table)', default='table')
def ipa_status(format):
    """IPA related host/DNS details"""
    doctor_ipa.status(format)

@doctor.group("metering")
def metering():
    """Metering related operations"""

@doctor.group("monitoring")
def monitoring():
    """Monitoring related operations"""

@doctor.group("service")
def service():
    """Service related diagnostics"""

@doctor.group("system")
def system():
    """System related metrics (cpu,disk,processes)"""

@doctor.group("parcel")
def parcel():
    """Parcel relateted diagnostic operations"""

@parcel.command('status')
@click.option('--image-id', '-i', help='Image UUID for checking parcels')
def parcel_status_get(image_id):
    """Get status for parcels"""
    doctor_parcel.run(image_id)

@doctor.group("salt")
def salt():
    """Salt related diagnostic operations"""

@salt.command("ping")
def salt_ping():
    """Ping salt agents (from salt master node)"""
    doctor_salt.ping()

@doctor.group("scm")
def scm():
    """Cloudera Manger service related commands"""

@doctor.group("fluentd")
def fluentd():
    """Logging agent (fluentd) diagnostics"""

@monitoring.command('status')
@click.option('--config','-c', help='Prometheus config (yaml) location. Default: /opt/cdp-prometheus/prometheus.yml', type=click.Path(exists=True), required=True, default="/opt/cdp-prometheus/prometheus.yml")
@click.option('--request-signer-config','-r', help='Custom request signer configuration. Default: /opt/cdp-request-signer/conf/cdp-request-signer.yaml', required=False, default="/opt/cdp-request-signer/conf/cdp-request-signer.yaml")
@click.option('--metrics','-m', help='List of metircs to show (comma separated)', required=False)
@click.option('--remote-write-url','-u', help='Custom remote write url', required=False)
@click.option('--format', type=click.Choice(['table', 'json']), help='Command output format. (default: table)', default='table')
def monitoring_status(config, request_signer_config, metrics, remote_write_url, format):
    """Analyze monitoring services"""
    doctor_monitoring.status(metrics, format, config, request_signer_config, remote_write_url)

@fluentd.command('status')
@click.option('--metering', '-m', help='Force to use metering', is_flag=True)
@click.option('--format', type=click.Choice(['table', 'json']), help='Command output format. (default: table)', default='table')
@click.option('--databus-url', '-u', help='DataBus URL for network checks. (default: https://dbusapi.us-west-1.altus.cloudera.com)', default='https://dbusapi.us-west-1.altus.cloudera.com')
@click.option('--databus-s3-url', help='DataBus S3 URL for network checks.')
def fluentd_status(metering, format, databus_url, databus_s3_url):
    """Analyze fluentd processes"""
    doctor_fluentd.status(format, metering, databus_url=databus_url, databus_s3_url=databus_s3_url)

@fluentd.command('doctor')
@click.option('--metering', '-m', help='Force to use metering', is_flag=True)
@click.option('--metering-buffer-limit', '-l', help='Limit of the metering buffers in MB. (trigger restart on limit)', default=10)
@click.option('--databus-url', '-u', help='DataBus URL for network checks. (default: https://dbusapi.us-west-1.altus.cloudera.com)', default='https://dbusapi.us-west-1.altus.cloudera.com')
@click.option('--databus-s3-url', help='DataBus S3 URL for network checks.')
def fluentd_doctor(metering, metering_buffer_limit, databus_url, databus_s3_url):
    """Trroubleshoot and fix fluentd agents (by salt)"""
    doctor_fluentd.doctor(metering, metering_buffer_limit, databus_url, databus_s3_url)

@system.command('metrics')
@click.option('--format', type=click.Choice(['table', 'json']), help='Command output format. (default: table)', default='table')
@click.option('--verbose', '-v', help='Verbose output', is_flag=True)
@click.option('--pretty-print', '-p', help='Pretty print (only for JSON format)', is_flag=True)
def system_metrics(format, verbose, pretty_print):
    """Collect cpu/memory/disk/network metrics"""
    doctor_system.collect_system_metrics(format, verbose, pretty=pretty_print)

@system.command('process')
@click.option('--format', type=click.Choice(['table', 'json']), help='Command output format. (default: table)', default='table')
@click.option('--number', '-n', help='Top N memory/cpu consumer processes.', default=10)
@click.option('--pretty-print', '-p', help='Pretty print (only for JSON format)', is_flag=True)
def system_process(format, number, pretty_print):
    """Collect process details (top cpu/memory consumer processes)"""
    doctor_system.collect_process_details(number, format, pretty=pretty_print)

@system.command('connections')
@click.option('--pid', '-p', help="Filter against specific PID")
@click.option('--raddr-ip', '-ri', help="Filter against specific remote address IP")
@click.option('--raddr-port', '-rp', help="Filter against specific remote address port")
@click.option('--laddr-ip', '-li', help="Filter against specific local address IP")
@click.option('--laddr-port', '-lp', help="Filter against specific local address port")
@click.option('--status', '-s', help="Filter against specific connection status")
@click.option('--format', type=click.Choice(['table', 'json']), help='Command output format. (default: table)', default='table')
@click.option('--verbose', '-v', help='Verbose output', is_flag=True)
@click.option('--pretty-print', '-pp', help='Pretty print (only for JSON format)', is_flag=True)
def system_connections(pid, raddr_ip, raddr_port, laddr_ip, laddr_port, status, format, verbose, pretty_print):
    """Collect network connection details"""
    doctor_system.collect_connections_details(pid, raddr_ip, raddr_port, laddr_ip, laddr_port, status, format, verbose, pretty=pretty_print)

@doctor.command("server", hidden=True)
def doctor_server_command():
    """Start cdp-doctor server"""
    doctor_server.start()

@metering.command("status")
@click.option('--format', type=click.Choice(['table', 'json']), help='Command output format. (default: table)', default='table')
@click.option('--pretty-print', '-p', help='Pretty print JSON format', is_flag=True)
@click.option('--metrics-folder', help="Node exporter textfiles folder location (if installed)", default="/var/lib/node_exporter/files")
def metering_status(format, pretty_print, metrics_folder):
    """Metering related status checks"""
    timeout.timeout(doctor_metering.status, json_format=True if format.lower() == "json" else False, metrics_folder=metrics_folder, pretty=pretty_print, context="Metering status check", exit_on_timeout=True, format=format)

@metering.command("salt-healthcheck")
@click.option('--directory', '-d', help='Directory that is used to dump the healthecheck results.', type=click.Path(exists=True), required=True)
def salt_healthcheck(directory):
    """Run cdp-doctor metering healthcheck on all salt minions"""
    doctor_metering.generate_salt_healthcheck(directory)

@metering.command("process-all")
def metering_process_all():
    """Send all heartbeat JSON content to databus"""
    doctor_metering.process_heartbeat_files()

@metering.command("reprocess")
def metering_reprocess():
    """Reset fluent based metering event processing (salt)"""
    doctor_metering.restart_event_processing()

@scm.command("list-commands")
@click.option('--name','-n', help='Filter based on command name', required=False)
def scm_list_commands(name):
    """Listing commands of Cloudera Manager agents (local)"""
    doctor_scm.list_commands(name)

@scm.command("describe-command")
@click.option('--id','-i', help='Filter based on command name', cls=NotRequiredIf, not_required_if='name')
@click.option('--name','-n', help='Filter based on command name', cls=NotRequiredIf, not_required_if='id')
@click.option('--include-logs', '-l', help='Include stdout/stderr logs in the describe output', is_flag=True)
def scm_describe_command(id, name, include_logs):
    """Describe a Cloudera Manager agent command (local)"""
    doctor_scm.describe_command(id, name, include_logs)

@service.command('status')
@click.option('--names','-n', help='Filter for desired services', required=False)
@click.option('--format', type=click.Choice(['table', 'json']), help='Command output format. (default: table)', default='table')
@click.option('--pretty-print', '-p', help='Pretty print JSON format', is_flag=True)
def service_status_get(names, format, pretty_print):
    """Get status for services"""
    doctorService = doctor_service.DoctorService(format, pretty=pretty_print)
    doctorService.run(names)

# in progress
@service.command('systemd', hidden=True)
@click.option('--names','-n', help='Filter for desireds systemd services', required=False)
@click.option('--metrics-dir', '-m', help='Metrics directory, output will be written to this folder in OpenMetrics format.', type=click.Path(exists=True))
@click.option('--format', type=click.Choice(['table', 'json']), help='Command output format. (default: table)', default='table')
@click.option('--pretty-print', '-p', help='Pretty print JSON format', is_flag=True)
def service_systemd_report(names, metrics_dir, format, pretty_print):
    """Get resource usage report of systemd services"""
    doctorService = doctor_service.DoctorService(format, pretty=pretty_print)
    doctorService.systemd_report(names, metrics_dir)

@doctor.group("commands")
def commands():
    """Create report from custom commands"""

@commands.command("create")
@click.option('--config','-c', help='Yaml configuration that contains command and file output pairs.', type=click.Path(exists=True), required=True)
def create_report(config):
    """Write different command outputs to files"""
    doctor_commands.command_report(config)

@main.group()
def utils():
    """Helper commands"""

@utils.command('clean')
@click.option('--dir', '-d', help='Path to the file that should be uploaded to Databus.', type=click.Path(exists=True), required=True)
@click.option('--pattern','-p', help='Wildcard inside the directory that will be deleted')
def utils_clean(dir, pattern):
    """Helper command to clean files from directory"""
    cleaner.clean_directory(dir, pattern)

@utils.command('check-connection')
@click.option('--url', '-u', help='URL ', required=True)
def utils_connection_check(url):
    """Helper command to check internet connection by an URL"""
    connection.check_connection(url)

@utils.command('check-version')
@click.option('--version', '-v', help='Version that will be check aginst actual version', required=True)
@click.option('--operator','-o', help='Operator against a version', type=click.Choice(['eq', 'ge', 'gt', 'le', 'lt', 'ne']), required=True)
def utils_check_version(version, operator):
    """Helper command to check version is proper for cdp-telemetry cli"""
    cdp_telemetry_version.check_version(__version__, version, operator)

if __name__ == "__main__":
    main()
