#!/usr/bin/env python
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements. See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership. The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.
#

import base64
import dbus_client
import configparser
import os
import socket
import sys
import requests
import logging
import glob
import yaml
from dbus_client.rest import ApiException
from dbus_client.models.record import Record
from dbus_client.models.header import Header
from utils import altus_auth, proxy_settings, connection
from requests.packages.urllib3.exceptions import InsecureRequestWarning

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

logger=logging.getLogger('databus.put_record')

INLINE_RECORD_THRESHOLD = 600 * 1024
DEFAULT_DBUS_S3_PUT_TIMEOUT = 120

def upload(credential_file, url, file, paths, headers, stream, partition_key, profile='default', access_key_field='cdp_access_key_id', private_key_field='cdp_private_key', extra_headers_file=None):
    api_client=__prepare_api_client(credential_file, url, "POST", "/dbus/putRecord", profile, access_key_field, private_key_field)
    api_instance=dbus_client.DefaultApi(api_client=api_client)
    static_headers={}
    if extra_headers_file and os.path.exists(extra_headers_file):
        with open(extra_headers_file) as header_file:
            static_headers = yaml.load(header_file, yaml.SafeLoader)
    if paths:
        logger.debug("Use matcher to search for files: %s" % (paths))
        files = glob.glob(paths)
        num_files=len(files)
        extra_headers_file_context={}
        actual_part=0
        for file_to_upload in files:
            actual_part=actual_part+1
            file_name=os.path.basename(file_to_upload)
            extra_headers_file_context["filename"]=file_name
            extra_headers_file_context["fileTotalNum"]=num_files
            extra_headers_file_context["filePartNum"]=actual_part

            headers=resolve_headers_additional_headers(headers, static_headers, extra_headers_file_context)
            logger.debug("File to upload: %s" % (file_to_upload))
            __put_record(api_instance, credential_file, url, None, file_to_upload, headers, stream, partition_key)
    else:
        logger.debug("File to upload: %s" % (file))
        __put_record(api_instance, credential_file, url, None, file, headers, stream, partition_key)

def test_put_record(credential_file, url, profile='default', access_key_field='cdp_access_key_id', private_key_field='cdp_private_key', proxySettings=None, silenced=False):
    status=None
    api_client=__prepare_api_client(credential_file, url, "POST", "/dbus/testPutRecord", profile, access_key_field, private_key_field, proxySettings)
    api_instance=dbus_client.DefaultApi(api_client=api_client)
    input = dbus_client.TestPutRecordRequest()
    try:
        api_response = api_instance.test_put_record(input)
        logger.debug("Test put record response: %s\n" % api_response)
        status = "OK"
    except ApiException as e:
        if not silenced:
            logger.error("Exception when calling Databus API -> test_put_record: %s\n" % e)
        status=str(e.status)
    return status

def put_record(credential_file, url, value, file, headers, stream, partition_key, proxySettings=None, apiInstance=None):
    api_instance=None
    if apiInstance:
        api_instance=apiInstance
    else:
        api_client=__prepare_api_client(credential_file, url, "POST", "/dbus/putRecord", proxySettings=proxySettings)
        api_instance=dbus_client.DefaultApi(api_client=api_client)
    __put_record(api_instance, credential_file, url, value, file, headers, stream, partition_key, proxySettings)

def __put_record(api_instance, credential_file, url, value, file, headers, stream, partition_key, proxySettings=None):
    payload_size=0
    payload=None
    hostname=socket.gethostname()
    
    payload = None
    raw_data = None
    if value:
        raw_data = value
        payload_size=len(raw_data)
    else:
        payload_size=get_size(file)
    if not partition_key:
        partition_key = hostname
    if payload_size <= INLINE_RECORD_THRESHOLD and value:
        payload = base64.b64encode(value.encode('ascii')).decode('ascii')
    dbusHeaders=list()
    defaultHostHeader=Header('host', hostname)
    dbusHeaders.append(defaultHostHeader)
    if headers:
        for header in headers:
            splitted=header.split(":", 1)
            if len(splitted) == 2:
                dbusHeaders.append(Header(splitted[0], splitted[1]))
    logger.debug("Using DBUS headers for put-record: %s" % str(dbusHeaders))
    record=Record(stream, partition_key, payload_size, dbusHeaders, payload)
    input = dbus_client.PutRecordRequest(record)
    try:
        put_record_response = api_instance.put_record(input)
        logger.debug("Put record response: %s\n" % str(put_record_response))
        record_status = put_record_response.record.status
        if record_status == 'PENDING':
            resp=None
            if not os.getenv("REQUESTS_CA_BUNDLE"):
                cert=connection.get_ssl_cert_file()
                if cert:
                    os.environ["REQUESTS_CA_BUNDLE"]=cert
            if file:
                with open(file, 'rb') as file_data:
                    resp=requests.put(put_record_response.record.upload_url, file_data, timeout=DEFAULT_DBUS_S3_PUT_TIMEOUT, verify=False)
            else:
                resp=requests.put(put_record_response.record.upload_url, raw_data, timeout=DEFAULT_DBUS_S3_PUT_TIMEOUT, verify=False)
            if resp.status_code >= 200 and resp.status_code < 300:
                logger.info("Raw data for databus (from S3 reply) was successfully uploaded.")
            else:
                logger.error("Status code (from S3 reply, then upload): %s" % str(resp.status_code))
                if resp.status_code >= 400 and resp.status_code < 500:
                    logger.error(str(resp))
                    sys.exit(1)
                else:
                    # todo: retry on these errors
                    pass
        elif record_status == 'SENT':
            logger.info("Record sent to DBUS successfully. Record id: %s" % put_record_response.record.record_id)
    except ApiException as e:
        logger.error("Exception when calling Databus API -> put_record: %s\n" % e)
        sys.exit(1)

def create_api_instance(credential_file, url, proxySettings=None):
    api_client=__prepare_api_client(credential_file, url, "POST", "/dbus/putRecord", proxySettings=proxySettings)
    api_instance=dbus_client.DefaultApi(api_client=api_client)
    return api_instance

def __prepare_api_client(credential_file, url, request_method, request_url, profile='default', access_key_field='cdp_access_key_id', private_key_field='cdp_private_key', proxySettings=None):
    dbus_cred_config=configparser.ConfigParser()
    dbus_cred_config.read(credential_file)
    default_section=None
    if profile in dbus_cred_config:
        default_section=dbus_cred_config[profile]
    elif "dbus" in dbus_cred_config:
        default_section=dbus_cred_config["dbus"]
    dbus_access_key_id=None
    dbus_private_key=None
    if access_key_field in default_section:
        dbus_access_key_id=default_section[access_key_field]
    elif 'databus_access_key_id' in default_section:
        dbus_access_key_id=default_section['databus_access_key_id']
    elif 'cdp_access_key_id' in default_section:
        dbus_access_key_id=default_section['cdp_access_key_id']
    elif 'altus_access_key_id' in default_section:
        dbus_access_key_id=default_section['altus_access_key_id']
    
    if private_key_field in default_section:
        dbus_private_key=default_section[private_key_field]
    elif 'databus_access_secret_key' in default_section:
        dbus_private_key=default_section['databus_access_secret_key']
    elif 'cdp_private_key' in default_section:
        dbus_private_key=default_section['cdp_private_key']
    elif 'altus_private_key' in default_section:
        dbus_private_key=default_section['altus_private_key']
    
    sign_algo='Ed25519'
    if 'databus_access_secret_key_algo' in default_section:
        sign_algo=default_section['databus_access_secret_key_algo']
    elif 'cdp_access_secret_key_algo' in default_section:
        sign_algo=default_section['cdp_access_secret_key_algo']
    elif 'altus_access_secret_key_algo' in default_section:
        sign_algo=default_section['altus_access_secret_key_algo']
    api_configuration=dbus_client.Configuration()
    api_configuration.host=url
    if proxySettings is None:
        proxySettings=proxy_settings.ProxySettings()
    connection.init_no_proxy()
    if not requests.utils.should_bypass_proxies(url, None):
        logger.debug("Will bypass proxies for %s" % url)
        api_configuration.proxy=proxySettings.get_url()
        api_configuration.proxy_username=proxySettings.get_user()
        api_configuration.proxy_password=proxySettings.get_password()
    api_configuration.ssl_ca_cert=connection.get_ssl_cert_file()
    
    api_client=dbus_client.ApiClient(configuration=api_configuration)
    signer=None
    if "ecdsa" in sign_algo.lower():
        signer=altus_auth.ECDSAv1Auth(dbus_access_key_id, dbus_private_key, logger)
    else:
        signer=altus_auth.Ed25519v1Auth(dbus_access_key_id, dbus_private_key, logger)
    headers=signer.add_auth(request_method, request_url)
    __add_auth_headers(api_client, headers)
    return api_client

def __add_auth_headers(api_client, headers):
    for k,v in headers.items():
        api_client.set_default_header(k, v)

def resolve_headers_additional_headers(headers, static_headers, extra_headers_file_context):
    if static_headers:
        resolved_headers={}
        hostname = socket.gethostname()
        list_headers=list(headers)
        for sHeaderKey, sHeaderValue in static_headers.items():
            resolvedValue=None
            if sHeaderValue is not None:
                resolvedFilename=str(sHeaderValue).replace("$filename", extra_headers_file_context["filename"])
                resolvedHostname=resolvedFilename.replace("$hostname", hostname)
                resolvedParts=resolvedHostname.replace("$filePartNum", str(extra_headers_file_context["filePartNum"]))
                resolvedValue=resolvedParts.replace("$fileTotalNum", str(extra_headers_file_context["fileTotalNum"]))
            resolved_headers[sHeaderKey]=resolvedValue
        for rHeaderKey, rHeaderValue in resolved_headers.items():
            list_headers.append(rHeaderKey + ":" + rHeaderValue)
        headers=tuple(list_headers)
    return headers

def get_size(filename):
    st = os.stat(filename)
    return st.st_size