import os, signal
import time
import logging
import sys
from shutil import which, copyfile
from utils import response, telemetry_logger, ccm_config_holder
from .service import DoctorService
from .network import getCCMConfig, getResult, get_ccm_answers, is_ccm_accessible, http_test
from filecollector.archiver import Archiver

CCM_DEBUGGING_REPORT_FILENAME = 'ccm_debugging_report'

CCM_DEBUGGING_REPORT_TEMP_FOLDER = os.path.join('/tmp', CCM_DEBUGGING_REPORT_FILENAME)

TCPDUMP_OUTPUT_FILENAME = os.path.join(CCM_DEBUGGING_REPORT_TEMP_FOLDER, 'tcpdump_output.pcap')

CONNECTION_CHECK_OUTPUT = os.path.join(CCM_DEBUGGING_REPORT_TEMP_FOLDER, "connectivity_check.txt")

SERVICE_CHECK_OUTPUT = os.path.join(CCM_DEBUGGING_REPORT_TEMP_FOLDER, "autossh_service_check.txt")

PROXY_CHECK_OUTPUT = os.path.join(CCM_DEBUGGING_REPORT_TEMP_FOLDER, "proxy_check.txt")

FILES_TO_COLLECT = [
    '/var/log/messages',
    '/var/log/autossh-GATEWAY.log',
    '/var/log/autossh-KNOX.log',
    '/var/log/socket_wait_cleanup.log',
    '/cdp/bin/reverse-tunnel-values-GATEWAY.sh',
    '/cdp/bin/reverse-tunnel-values-KNOX.sh',
    '/var/log/ccmv2-inverting-proxy-agent.log',
    '/var/log/jumpgate/access.log',
    '/var/log/jumpgate/out.log'
]

logger=logging.getLogger(__name__)

def check_ccm_values(ccm_host, ccm_port):
    if not ccm_host and not ccm_port:
        logger.error(telemetry_logger.fail_message("FAILED"))
        sys.exit(1)

def status(output_format="table"):
    ccmConfigHolder=ccm_config_holder.CcmConfigHolder()
    ccm_host, ccm_port = getCCMConfig(ccmConfigHolder)
    check_ccm_values(ccm_host, ccm_port)
    logger.info("Connectivity checks:")
    report_ccm(ccm_host, ccm_port, ccmConfigHolder, False)
    if not ccmConfigHolder.is_v2():
        logger.info("Service checks:")
        autossh_service_check(False)
    check_proxy(ccmConfigHolder, False)

def report(output_file=None, skipTcpDump=False):
    if not os.path.exists(CCM_DEBUGGING_REPORT_TEMP_FOLDER):
        os.makedirs(CCM_DEBUGGING_REPORT_TEMP_FOLDER)
    ccmConfigHolder=ccm_config_holder.CcmConfigHolder()
    ccm_host, ccm_port = getCCMConfig(ccmConfigHolder)
    check_ccm_values(ccm_host, ccm_port)
    logger.info("Running connectivity checks...")
    report_ccm(ccm_host, ccm_port, ccmConfigHolder)
    if not ccmConfigHolder.is_v2():
        logger.info("Running service checks...")
        autossh_service_check()
    logger.info("Gathering proxy settings...")
    check_proxy(ccmConfigHolder)
    if not skipTcpDump:
        host_to_check=ccm_host
        port_to_check=ccm_port
        if ccmConfigHolder.is_v2():
            ccmV2Proxy=ccmConfigHolder.get_ccm_v2_proxy_settings()
            if ccmV2Proxy:
                ccm_v2_proxy_host, ccm_v2_proxy_port=ccmV2Proxy.get_host_and_port()
                if ccm_v2_proxy_host and ccm_v2_proxy_port:
                    host_to_check=ccm_v2_proxy_host
                    port_to_check=ccm_v2_proxy_port
        else:
            proxy_host, proxy_port, proxy_username, proxy_password=ccmConfigHolder.get_ccm_v1_proxy_settings()
            if proxy_host and proxy_port:
                host_to_check=proxy_host
                port_to_check=proxy_port

        collect_tcpdump_output(host_to_check, port_to_check)
    collect_files_and_create_archive(output_file)
    logger.info(telemetry_logger.success_message("DONE"))

def report_ccm(ccm_host, ccm_port, ccmConfigHolder, writeToFile=True):
    output_format = "json" if writeToFile else None
    ccm_check_result=ccm_check(ccm_host, ccm_port, ccmConfigHolder, output_format)
    output=telemetry_logger.format_table(ccm_check_result, headers=[])
    if writeToFile:
        with open(CONNECTION_CHECK_OUTPUT, 'w') as f:
            print("Connectivity checks:\n" + output, file=f)
    else:
        logger.info(output)

def ccm_check(ccm_host, ccm_port, ccmConfigHolder, output_format="table", output_file=None):
    results=[]
    cmd_output_format='json' if output_file else output_format
    ccm_proxy_host, ccm_proxy_port, ccm_proxy_user, ccm_proxy_password = None, None, None, None
    ccm_proxy_url=None
    ccm_v2=True
    ccmV2Url=None
    ccmV2ProxySettings=None
    ccm_v2=ccmConfigHolder.is_v2()
    if ccm_v2:
        ccmV2Url=ccmConfigHolder.get_ccm_v2_url()
        ccmV2ProxySettings = ccmConfigHolder.get_ccm_v2_proxy_settings()
        if ccmV2ProxySettings:
            ccm_proxy_host, ccm_proxy_port = ccmV2ProxySettings.get_host_and_port()
            ccm_proxy_user = ccmV2ProxySettings.get_user()
            ccm_proxy_password = ccmV2ProxySettings.get_password()
            ccm_proxy_url = ccmV2ProxySettings.get_url()
        else:
            ccm_proxy_host, ccm_proxy_port, ccm_proxy_user, ccm_proxy_password = ccmConfigHolder.get_ccm_v1_proxy_settings()
    if ccm_v2:
        cacert, cert, key = ccmConfigHolder.get_ccm_v2_certs()
        ccmAccessible=getResult(http_test(ccmV2Url, ccm_proxy_url, ccm_proxy_user, ccm_proxy_password, cmd_output_format, cacert=cacert, cert=cert, key=key), cmd_output_format)
        ccmConfigHolder.cleanup_jumpgate_tmp_certs()
    else:
        ccmAccessible=getResult(is_ccm_accessible(ccm_host, ccm_port, proxy_host=ccm_proxy_host, proxy_port=ccm_proxy_port, proxy_user=ccm_proxy_user, proxy_password=ccm_proxy_password), cmd_output_format)
    results.append(["CCM address", ccm_host + ":" + ccm_port])
    results.append(["DNS Answers", get_ccm_answers(ccm_host, cmd_output_format)])
    results.append(["CCM is accessible", ccmAccessible])
    return results

def collect_files_and_create_archive(output_file, compression="zip"):
    logger.info("Start copying file into %s ..." % CCM_DEBUGGING_REPORT_TEMP_FOLDER)
    for file_to_collect in FILES_TO_COLLECT:
        if os.path.exists(file_to_collect):
            logger.info("Collecting file: %s ..." % file_to_collect)
            base_filename=os.path.basename(file_to_collect)
            copyfile(file_to_collect, os.path.join(CCM_DEBUGGING_REPORT_TEMP_FOLDER, base_filename))
        else:
            logger.info("File '%s' does not exist. Skipping from collection" % file_to_collect)
    destination=CCM_DEBUGGING_REPORT_FILENAME
    if output_file:
        if os.path.exists(output_file):
            if os.path.isdir(output_file):
                destination=os.path.join(output_file, destination)
                logger.info("Path '%s' is an already existing directory. Use '%s' as destination." % (output_file, __add_comp_ext(destination, compression))) 
        destination=output_file
    if destination.endswith('.' + compression):
        destination = destination[:-4]
    dir_path=os.path.dirname(os.path.abspath(destination))
    if not os.path.exists(dir_path):
        logger.info("Path %s do not exist. Creating it ..." % dir_path)
        os.makedirs(dir_path)
    logger.info("Writing report into %s ..." % os.path.abspath(__add_comp_ext(destination, compression)))
    archiver=Archiver(CCM_DEBUGGING_REPORT_TEMP_FOLDER, destination, compression, True)
    archiver.make_archive()

def autossh_service_check(writeToFile=True):
    doctorService=DoctorService('pretty')
    service_state_list=[]
    service_check_value=doctorService.get_service_report_by_name("autossh")
    service_state_list.append(("autossh", service_check_value))
    table_output=response.ResponseHolder(tuple_content=service_state_list).to_table()
    if writeToFile:
        output="Services: \n%s" % table_output
        with open(SERVICE_CHECK_OUTPUT, 'w') as f:
            print(output, file=f)
    else:
        logger.info(table_output)

def check_proxy(ccmConfigHolder, writeToFile=True):
    proxy_host, proxy_port = '', ''
    proxy_user=None
    if ccmConfigHolder.is_v2():
        proxyConf=ccmConfigHolder.get_ccm_v2_proxy_settings()
        if proxyConf:
            proxy_host, proxy_port = proxyConf.get_host_and_port()
            proxy_user = proxyConf.get_user()
    else:
        proxy_host, proxy_port, proxy_user, proxy_pass = ccmConfigHolder.get_ccm_v1_proxy_settings()
    if proxy_host:
        if not writeToFile:
            logger.info("Proxy settings:")
        proxy_resp_list=[]
        has_proxy_auth = "YES" if proxy_user else "NO"
        proxy_resp_list.append(("host", proxy_host))
        proxy_resp_list.append(("port", proxy_port))
        proxy_resp_list.append(("auth", has_proxy_auth))
        table_output=response.ResponseHolder(tuple_content=proxy_resp_list).to_table()
        if writeToFile:
            output="Proxy settings: \n%s" % table_output
            with open(PROXY_CHECK_OUTPUT, 'w') as f:
                print(output, file=f)
        else:
            logger.info(table_output)

def collect_tcpdump_output(host, port):
    if which('tcpdump') is not None:
        cmd = 'tcpdump -i any host ' + host + ' and port ' + port + ' -w ' + TCPDUMP_OUTPUT_FILENAME + ' &'
        os.system(cmd)
        logger.info('Capturing tcpdump (host: %s, port: %s) ... Please wait for 60 seconds ...' % (str(host), str(port)))
        number_of_sleeps=0
        max_number_of_sleeps=20
        while number_of_sleeps < max_number_of_sleeps:
            time.sleep(3)
            sys.stdout.write(".")
            number_of_sleeps+=1
            if number_of_sleeps == max_number_of_sleeps:
                sys.stdout.write("\n")
            sys.stdout.flush()
        for line in os.popen("ps ax | grep tcpdump | grep -v grep"):
            fields = line.split()
            pid = fields[0]
            os.kill(int(pid), signal.SIGKILL)
        logger.info('tcpdump captured successfully!')
    else:
        logger.info("tcpdump is not installed on this machine. please install it")

def __add_comp_ext(name, compression):
    return name + "." + compression if not name.endswith("." + compression) else name
