#!/usr/bin/env python

import os
import logging
import sys
import psutil
from pathlib import Path
from datetime import datetime
from utils import response, command_helper, telemetry_logger, logging_agent_config_holder
from .service import DoctorService
from .network import network_status
from google.protobuf.json_format import ParseDict, MessageToDict
from nodestatus.generated import nodestatus_pb2 as NodeStatus

logger=logging.getLogger(__name__)

def status(format, check_metering=False, silenced=False, databus_url=None, databus_s3_url=None):
    nowTs=int(datetime.now().strftime('%s'))
    use_logging_agent=False
    use_td_agent=False
    binaryCmd=None
    jsonResponse={}
    loggingServiceName="cdp-logging-agent"
    loggingAgentConfig="/etc/cdp-logging-agent/cdp-logging-agent.conf"
    metering_conf = "/etc/metering/generate_heartbeats.ini"
    loggingAgentLogDir="/var/log/cdp-logging-agent"
    if os.path.exists(metering_conf):
        check_metering=True
    if os.path.exists("/etc/cdp-logging-agent"):
        use_logging_agent=True
        binaryCmd="/opt/cdp-logging-agent/bin/ruby"
    elif os.path.exists("/etc/td-agent"):
        use_td_agent=True
        loggingServiceName="td-agent"
        loggingAgentConfig="/etc/td-agent/td-agent.conf"
        binaryCmd="/opt/td-agent/bin/ruby"
        loggingAgentLogDir="/var/log/td-agent"
    jsonResponse["timestamp"]=nowTs
    if not use_td_agent and not use_logging_agent:
        if format != "json":
            logging.info("Logging agent does not exist.")
        return jsonResponse
    jsonResponse["loggingServiceName"]=loggingServiceName
    isConfiguredBySalt=False
    with open(loggingAgentConfig) as f:
        if "CONFIGURED BY SALT" in f.read():
            isConfiguredBySalt=True
    saltResponse="OK"
    if format == "json":
        saltResponse = "OK" if isConfiguredBySalt else "NOK"
    else:
        saltResponse = telemetry_logger.success_message("[OK]") if isConfiguredBySalt else telemetry_logger.fail_message("[NOK]")
    jsonResponse["loggingServiceStates"]=[]
    jsonResponse["saltConfigured"]=saltResponse
    service_state_list=__service_checks(loggingServiceName, check_metering, format)
    if format == "json":
        for service_state in service_state_list:
            service_state_json=MessageToDict(service_state)
            jsonResponse["loggingServiceStates"].append(service_state_json)
    else:
        service_state_list.append(("salt configured", saltResponse))
        servicesResponse=response.ResponseHolder(tuple_content=service_state_list)
        logger.info("Logging agent service details:")
        logger.info(servicesResponse.to_table())
    loggingAgentConfHolder=logging_agent_config_holder.LoggingAgentConfigHolder()
    cloudProvider=None
    logging_output_conf=loggingAgentConfHolder.get_output_configs()
    if logging_output_conf:
        if "s3_bucket" in logging_output_conf:
            cloudProvider="AWS"
        elif "azure_storage_account" in logging_output_conf:
            cloudProvider="AZURE"
        elif "bucket" in logging_output_conf:
            cloudProvider="GCP"
    logging_metering_output_conf=loggingAgentConfHolder.get_metering_output_configs()
    logging_databus_output_conf=loggingAgentConfHolder.get_databus_output_configs()
    checkDatabus=True if logging_metering_output_conf or logging_databus_output_conf else False
    jsonResponse["networkDetails"]={}
    network_response={}
    network_checks_failed=False
    try:
        network_response=network_status(output_format='json', scan_neighbours=False, skip_scan_neighbours=True, silenced=True, skip_ccm=True, databus_address=databus_url, databus_s3_address=databus_s3_url, cloud_provider=cloudProvider)
        network_obj=ParseDict(network_response, NodeStatus.NetworkDetails(), ignore_unknown_fields=True)
    except Exception as e:
        logger.debug(str(e))
        network_checks_failed=True
    network_settings=[]
    if "s3Accessible" in network_response:
        s3Resp=getFormattedResult(network_response["s3Accessible"], format)
        network_settings.append(("S3 Accessible", s3Resp))
        jsonResponse["networkDetails"]["s3Accessible"]=s3Resp
    if "stsAccessible" in network_response:
        stsResp=getFormattedResult(network_response["stsAccessible"], format)
        network_settings.append(("STS Accessible", stsResp))
        jsonResponse["networkDetails"]["stsAccessible"]=stsResp
    if "abfsAccessible" in network_response:
        abfsResp=getFormattedResult(network_response["abfsAccessible"], format)
        network_settings.append(("ABFS Accessible", abfsResp))
        jsonResponse["networkDetails"]["abfsAccessible"]=abfsResp
    if "gcsAccessible" in network_response:
        gcsResp=getFormattedResult(network_response["gcsAccessible"], format)
        network_settings.append(("GCS Accessible", gcsResp))
        jsonResponse["networkDetails"]["gcsAccessible"]=gcsResp
    if "databusAccessible" in network_response:
        databusApiResp=getFormattedResult(network_response["databusAccessible"], format)
        network_settings.append(("Databus API Accessible", databusApiResp))
        jsonResponse["networkDetails"]["databusAccessible"]=databusApiResp
    if "databusS3Accessible" in network_response:
        databusS3Resp=getFormattedResult(network_response["databusS3Accessible"], format)
        network_settings.append(("Databus S3 API Accessible", databusS3Resp))
        jsonResponse["networkDetails"]["databusS3Accessible"]=databusS3Resp
    if format != "json":
        if network_settings:
            network_settings_resp=response.ResponseHolder(tuple_content=network_settings)
            logger.info("Network checks:")
            logger.info(network_settings_resp.to_table())
        elif network_checks_failed:
            logger.info("Network checks failed.")
        else:
            logger.info("Network checks skipped.")
    metering_log_details=[]
    if use_logging_agent or use_td_agent:
        logging_processes=[]
        agent_number=0
        for proc in psutil.process_iter():
            proc_map={}
            try:
                if len(proc.cmdline()) > 0 and proc.cmdline()[0] == binaryCmd:
                    if proc.pid:
                        proc_map["pid"]=proc.pid
                    if sys.platform == "linux" and proc.pid:
                        ps_memory_cpu_cmd='ps -eo %%mem,%%cpu --no-headers -f %s | sed "s/^ *//" | sed "s/[[:space:]]\{1,\}/,/g"'
                        out,err,code=command_helper.run_command(ps_memory_cpu_cmd % str(proc.pid), True)
                        if (str(code).rstrip()) == "0":
                            stroutput=str(out.rstrip().decode())
                            for line in stroutput.splitlines():
                                splitted=line.split(",")
                                if len(splitted) == 2:
                                    proc_map['memory_percent']=str(splitted[0])
                                    proc_map['cpu_percent']=str(splitted[1])
                    else:
                        proc_map['memory_percent']=proc.memory_percent()
                    proc_map['vms']=proc.memory_info().vms
                    proc_map['rss']=proc.memory_info().rss
                    open_files=proc.open_files()
                    proc_map['number_of_open_files']=len(open_files)
                    is_supervisor=False if "--under-supervisor" in proc.cmdline() else True
                    proc_map['supervisor']=is_supervisor
                    if is_supervisor:
                        proc_map["name"]="Supervisor"
                    else:
                        agent_number=agent_number+1
                        proc_map["name"]="Agent #%s" % str(agent_number)
                    logging_processes.append(proc_map)
            except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
                pass
        jsonResponse["processes"]=logging_processes
        
        if format != "json":
            if logging_processes:
                for logging_process in logging_processes:
                    logging_proc_data=list()
                    if "pid" in logging_process:
                        logging_proc_data.append(("PID", logging_process["pid"]))
                    if "cpu_percent" in logging_process:
                       logging_proc_data.append(("CPU", logging_process["cpu_percent"])) 
                    if "memory_percent" in logging_process:
                        logging_proc_data.append(("Memory", logging_process["memory_percent"]))
                    logging_proc_data.append(("VMS", __sizeof_fmt(logging_process["vms"])))
                    logging_proc_data.append(("RSS", __sizeof_fmt(logging_process["rss"])))
                    logging_proc_data.append(("Open files", logging_process["number_of_open_files"]))
                    logging_process_resp=response.ResponseHolder(tuple_content=logging_proc_data)
                    logger.info("Logging process '%s':" % logging_process["name"])
                    logger.info(logging_process_resp.to_table())
        buffers_size_map={}
        if os.path.exists(loggingAgentLogDir):
            files_in_var_log=os.listdir(loggingAgentLogDir)
            for fp in files_in_var_log:
                path_of_buffer=os.path.join(loggingAgentLogDir, fp)
                if os.path.isdir(path_of_buffer) and fp != "pos":
                    buffer_directory = Path(path_of_buffer)
                    buffer_size=sum(f.stat().st_size for f in buffer_directory.glob('**/*') if f.is_file())
                    buffers_size_map[fp]=buffer_size
        jsonResponse["buffers"]=buffers_size_map
        if format == "json":
            resp_json=response.ResponseHolder(map_content=jsonResponse).to_json(False, False)
            if not silenced:
                logger.info(resp_json)
        else:
            if buffers_size_map:
                formatted_buffer_resp={}
                for k, v in buffers_size_map.items():
                    formatted_buffer_resp[k]=__sizeof_fmt(v)
                buffer_info=response.ResponseHolder(map_content=formatted_buffer_resp)
                logger.info("Size of buffers:")
                logger.info(buffer_info.to_table())
        return jsonResponse

def doctor(check_metering=False, max_metering_file_size=10, databus_url=None, databus_s3_url=None):
    exit_code, message=troubleshoot(check_metering, max_metering_file_size, databus_url, databus_s3_url)
    if message:
        logger.error(message)
    sys.exit(exit_code)

def troubleshoot(check_metering=False, max_metering_file_size=10, databus_url=None, databus_s3_url=None):
    logger.info("Running logging agent checks ...")
    statusResp=status('json', check_metering, True, databus_url, databus_s3_url)
    okMessage=telemetry_logger.success_message("OK")
    notOkMessage=telemetry_logger.fail_message("NOK")
    hasMetering=False
    loggingServiceName="cdp-logging-agent"
    if "loggingServiceName" in statusResp:
        logger.info("Logging agent installed ... %s" % okMessage)
        logger.info("Logging agent found: %s" % statusResp["loggingServiceName"])
        loggingServiceName=statusResp["loggingServiceName"]
    else:
        logger.info("Logging agent installed ... %s" % notOkMessage)
        return 0, "Logging agent is not installed, skipping status checks"
    if "saltConfigured" in statusResp and "OK" == statusResp["saltConfigured"]:
        logger.info("Logging agent configured by Salt ... %s" % okMessage)
    else:
        logger.info("Logging agent configured by Salt ... %s" % notOkMessage)
        return 0, ""
    if "loggingServiceStates" in statusResp:
        services_list=statusResp["loggingServiceStates"]
        meteringRunning=False
        loggingRunning=False
        for service in services_list:
            if service["name"] == "metering-heartbeat":
                hasMetering=True
                meteringRunning=True if service["status"] == "OK" else False
            elif service["name"] == statusResp["loggingServiceName"]:
                loggingRunning=True if service["status"] == "OK" else False
        if hasMetering and not meteringRunning:
            logger.info("Metering agent service status ... %s" % notOkMessage)
            return 1, "Please configure and start metering heartbeat process properly. Exiting ..."
        if loggingRunning:
            logger.info("Logging agent service status ... %s" % okMessage)
        else:
            logger.info("Logging agent service status ... %s" % notOkMessage)
            __start_logging_agent(loggingServiceName)
            return 0, ""
    cloudStorageType=None
    cloudStorageAccessible=True
    databusApiAccessible=True
    if "networkDetails" in statusResp:
        networkDetails=statusResp["networkDetails"]
        if "s3Accessible" in networkDetails:
            if networkDetails["s3Accessible"] == "OK":
                logger.info("S3 endpoint accessibility ... %s" % okMessage)
            elif networkDetails["s3Accessible"] == "NOK":
                logger.info("S3 endpoint accessibility ... %s" % notOkMessage)
                cloudStorageType="S3"
                cloudStorageAccessible=False
        if "abfsAccessible" in networkDetails:
            if networkDetails["abfsAccessible"] == "OK":
                logger.info("ABFS endpoint accessibility ... %s" % okMessage)
            elif networkDetails["abfsAccessible"] == "NOK":
                logger.info("ABFS endpoint accessibility ... %s" % notOkMessage)
                cloudStorageType="ABFS"
                cloudStorageAccessible=False
        if "gcsAccessible" in networkDetails:
            if networkDetails["gcsAccessible"] == "OK":
                logger.info("GCS endpoint accessibility ... %s" % okMessage)
            elif networkDetails["gcsAccessible"] == "NOK":
                logger.info("GCS endpoint accessibility ... %s" % notOkMessage)
                cloudStorageType="GCS"
                cloudStorageAccessible=False
        if "databusAccessible" in networkDetails:
            if networkDetails["databusAccessible"] == "OK":
                logger.info("Databus API endpoint accessibility ... %s" % okMessage)
            elif networkDetails["databusAccessible"] == "NOK":
                logger.info("Databus API endpoint accessibility ... %s" % notOkMessage)
                databusApiAccessible=False
    meteringBufferTooLarge=False
    if "buffers" in statusResp:
        buffers=statusResp["buffers"]
        if buffers and "metering_databus" in buffers:
            metering_buffer_size=buffers["metering_databus"]
            if max_metering_file_size * 1024 ** 2 <= metering_buffer_size:
                logger.info("Metering buffer size check ... %s" % notOkMessage)
                logger.info("Metering buffer size is too large (%s, limit: %s MB)" % (__sizeof_fmt(metering_buffer_size), str(max_metering_file_size)))
                meteringBufferTooLarge=True
            else:
                logger.info("Metering buffer size check (limit: %s MB)... %s" % (str(max_metering_file_size), okMessage))
    if hasMetering:
        if meteringBufferTooLarge:
            if databusApiAccessible:
                __restart_logging_agent(loggingServiceName)
                return 0, ""
            else:
                logger.info("Metering buffer size is too large, but Databus API is also not accessible, fix that one first.")
    return 0, ""
    
def __restart_logging_agent(loggingServiceName):
    logger.info("Restarting %s ..." % loggingServiceName)
    os.system("systemctl restart %s" % loggingServiceName)

def __start_logging_agent(loggingServiceName):
    logger.info("Starting %s ..." % loggingServiceName)
    os.system("systemctl start %s" % loggingServiceName)

def __service_checks(loggingServiceName, metering=False, format="table"):
    service_state_list=list()
    output_map={}
    doctorService=DoctorService(format)
    heartbeatServiceName="metering-heartbeat"
    serviceStatuses=[]
    serviceResults=[]
    
    serviceArr = []
    serviceArr.append(loggingServiceName)
    if metering:
        serviceArr.append("metering-heartbeat")
    for serviceName in serviceArr:
        result_pair = []
        result_pair.append(serviceName)
        result_pair.append(doctorService.get_service_report_by_name(serviceName))
        serviceResults.append(result_pair)
    if format == "json":
        for service_tuple in serviceResults:
            serviceStatus=NodeStatus.ServiceStatus()
            serviceStatus.name=service_tuple[0]
            serviceStatus.status=NodeStatus.HealthStatus.Value(service_tuple[1])
            serviceStatuses.append(serviceStatus)
        return serviceStatuses
    else:
        return serviceResults

def getFormattedResult(result, output_format='table'):
    if output_format != "json":
        if result == "OK":
            return  telemetry_logger.success_message("[%s]" % str(result))
        else:
            return telemetry_logger.fail_message("[%s]" % str(result))
    else:
        return result

def __sizeof_fmt(num, suffix='B'):
    for unit in ['','K','M','G','T','P','E','Z']:
        if abs(num) < 1024.0:
            return "%3.1f %s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f %s%s" % (num, 'Y', suffix)