import datetime
from databus import databus
from utils import proxy_settings, connection, response, telemetry_logger, command_helper, version, doctor_output
import sys
import glob
import os
import shutil
import json
import platform
import configparser
import logging
import traceback
from .service import DoctorService
from .salt import run_salt_command
from pathlib import Path
from google.protobuf.json_format import MessageToDict
from nodestatus.generated import nodestatus_pb2 as NodeStatus

logger=logging.getLogger(__name__)

def restart_event_processing(heartbeat_json_file=None, logging_agent_conf=None):
    use_td_agent=False
    if os.path.exists("/etc/td-agent"):
        use_td_agent=True
    run_and_log_salt_command('state.apply fluent.agent-stop')
    
    if use_td_agent:
        run_and_log_salt_command("cmd.run 'rm -r /var/log/td-agent/pos/metering.json.pos && rm -rf /var/log/td-agent/metering_databus/**'", True)
    else:
        run_and_log_salt_command("cmd.run 'rm -r /var/log/cdp-logging-agent/pos/metering.json.pos && rm -rf /var/log/cdp-logging-agent/metering_databus/**'", True)
    run_and_log_salt_command('state.apply fluent.init')

def generate_salt_healthcheck(metering_healthcheck_dir):
    output_file=os.path.join(metering_healthcheck_dir, "salt_metering_healthcheck_result.json")
    metering_output_file=os.path.join(metering_healthcheck_dir, "metering_health_status.json")
    run_and_log_salt_command("cmd.run 'cdp-doctor metering status --format json' --out=json --out-file=%s" % output_file, True)
    if os.path.exists(output_file):
        with open('salt_metering_healthcheck_result.json') as json_file:
            try:
                data=json.load(json_file)
                meteringJsonData=[]
                for instanceKey in data:
                    jsonStr=data[instanceKey]
                    instanceData={}
                    instanceData["host"]=instanceKey
                    try:
                        metering_check_details=json.loads(jsonStr)
                        instanceData["meteringDetails"]=metering_check_details
                        instanceData["responseStatus"]="OK"
                    except Exception as e:
                        logger.error("Error occurred during read json string: %s" % str(traceback.format_exc()))
                        instanceData["meteringDetails"]={}
                        instanceData["responseStatus"]="NOK"
                    meteringJsonData.append(instanceData)
                with open(metering_output_file, "w") as met_out_file:
                    nowTs=int(datetime.datetime.now().strftime('%s'))
                    meteringReport={}
                    meteringReport["nodes"]=meteringJsonData
                    meteringReport["timestamp"]=nowTs
                    json.dump(meteringReport, met_out_file, indent=2, sort_keys=True)
            except Exception as e:
                logger.error("Error occurred during load json file '%s': %s" % (output_file, str(traceback.format_exc())))
                sys.exit(1)

def process_heartbeat_files(heartbeat_json_file=None, logging_agent_conf=None, dbus_credential_file=None):
    #heartbeat_json_file="examples/metering/heartbeats.json"
    #heartbeat_conf="examples/metering/generate_heartbeats.ini"
    #logging_agent_conf="examples/metering/databus_metering.conf"
    #dbus_credential_file="/Users/oliverszabo/.cdp/credentials"
    if heartbeat_json_file is None:
        heartbeat_json_file = "/var/log/metering/heartbeats.json"
    meteringDatabusConf=MeteringDatabusSettings(logging_agent_conf, dbus_credential_file)
    if meteringDatabusConf.databusEndpoint is None:
        logger.error("Cannot find metering databus configuration. Exiting ...")
        sys.exit(1)
    proxySettings=None
    if meteringDatabusConf.proxyUrl:
        proxySettings=proxy_settings.ProxySettings(proxy_url=meteringDatabusConf.proxyUrl, proxy_user=meteringDatabusConf.proxyUser, proxy_password=meteringDatabusConf.proxyPassword)
    heartbeat_json_pattern="%s*" % heartbeat_json_file
    record_threshold = 599 * 1024
    defaultHeaders=meteringDatabusConf.databusHeaders.split(",")
    databusClientInstance=databus.create_api_instance(meteringDatabusConf.databusCredentilFile, meteringDatabusConf.databusEndpoint, proxySettings=proxySettings)

    for heartbeat_file in sorted(glob.glob(heartbeat_json_pattern, recursive=True), key=os.path.getmtime):
        logger.info("Processing heartbeats from file: %s" % heartbeat_file)
        buffer_size=0
        buffer=""
        with open(heartbeat_file) as heartbeat_file_obj:
            for line in heartbeat_file_obj:
                buffer_size=buffer_size+sys.getsizeof(line)
                if buffer_size > record_threshold:
                    databus.put_record(meteringDatabusConf.databusCredentilFile, meteringDatabusConf.databusEndpoint, buffer, None, defaultHeaders, meteringDatabusConf.databusStream, None, proxySettings=proxySettings, apiInstance=databusClientInstance)
                    buffer_size=sys.getsizeof(line)
                    buffer=line
                else:
                    buffer=buffer+line
        if buffer and buffer_size > 0:
            databus.put_record(meteringDatabusConf.databusCredentilFile, meteringDatabusConf.databusEndpoint, buffer, None, defaultHeaders, meteringDatabusConf.databusStream, None, proxySettings=proxySettings, apiInstance=databusClientInstance)

def run_and_log_salt_command(command, doNotSplit=False):
    logger.info("Running salt command: %s" % command)
    out, err, exit_code = run_salt_command(command, doNotSplit=doNotSplit)
    if (str(exit_code).rstrip() == "0"):
        logger.info(telemetry_logger.success_message("[OK]"))
    else:
        logger.error(telemetry_logger.fail_message("[FAILED]"))
        logger.info("StdOut: \n%s" % out)
        logger.info("StdErr: \n%s" % err)
        logger.info("Exit code: \n%s" % str(exit_code))

def status(metrics_folder="/var/lib/node_exporter/files", heartbeat_json_file=None, heartbeat_conf=None, logging_agent_conf=None, dbus_credential_file=None, json_format=False, pretty=False):
    #heartbeat_json_file="examples/metering/heartbeats.json"
    #heartbeat_conf="examples/metering/generate_heartbeats.ini"
    #logging_agent_conf="examples/metering/databus_metering.conf"
    #dbus_credential_file="/Users/oliverszabo/.cdp/credentials"
    use_td_agent=False
    binaryVersion=version.get_binary_version()
    meteringDetails=NodeStatus.MeteringDetails(cdpTelemetryVersion=binaryVersion)
    logging_agent_log_dir = "/var/log/cdp-logging-agent"
    if heartbeat_json_file is None:
        heartbeat_json_file = "/var/log/metering/heartbeats.json"
    if heartbeat_conf is None:
        heartbeat_conf = "/etc/metering/generate_heartbeats.ini"
    if logging_agent_conf is None:
        if os.path.exists("/etc/cdp-logging-agent"):
            logging_agent_conf = "/etc/cdp-logging-agent/databus_metering.conf"
        elif os.path.exists("/etc/td-agent"):
            use_td_agent=True
            logging_agent_conf = "/etc/td-agent/databus_metering.conf"
            logging_agent_log_dir = "/var/log/td-agent"
    
    logging_agent_exists=True if logging_agent_conf and os.path.exists(logging_agent_conf) else False
    metering_service_exists=True if os.path.exists(heartbeat_conf) else False
    if logging_agent_exists:
        if json_format:
            meteringDetails.loggingAgentConfig=NodeStatus.HealthStatus.OK
    else:
        if json_format:
            meteringDetails.loggingAgentConfig=NodeStatus.HealthStatus.NOK
        else:
            logger.info(telemetry_logger.fail_message("Not found any logging agent configuration file! Heartbeat shipping will not work."))
    if metering_service_exists:
        if json_format:
            meteringDetails.heartbeatConfig=NodeStatus.HealthStatus.OK
    else:
        if json_format:
            meteringDetails.heartbeatConfig=NodeStatus.HealthStatus.NOK
        else:
            logger.info(telemetry_logger.fail_message("Not found any heartbeat metering configuration file! Without this heartbeats will not be generated."))
    meteringDatabusConf=MeteringDatabusSettings(logging_agent_conf, dbus_credential_file)

    proxySettings=None
    if meteringDatabusConf.proxyUrl:
        proxySettings=proxy_settings.ProxySettings(proxy_url=meteringDatabusConf.proxyUrl, proxy_user=meteringDatabusConf.proxyUser, proxy_password=meteringDatabusConf.proxyPassword)
    meteringDatabusConf.databusEndpoint
    
    doctorService=DoctorService('pretty')
    service_state_list=list()
    if use_td_agent:
        logging_service_key="td-agent"
        logging_service_value=doctorService.get_service_report_by_name(logging_service_key)
    else:
        logging_service_key="cdp-logging-agent"
        logging_service_value=doctorService.get_service_report_by_name(logging_service_key)
    heartbeat_service_key="metering-heartbeat"
    heartbeat_service_value=doctorService.get_service_report_by_name(heartbeat_service_key)
    service_state_list.append((heartbeat_service_key, heartbeat_service_value))
    service_state_list.append((logging_service_key, logging_service_value))
    servicesResponse=response.ResponseHolder(tuple_content=service_state_list)
    servicesTable=servicesResponse.to_table(headers=['Service name', 'Service state'])
    if json_format:
        meteringDetails.heartbeatAgentRunning=NodeStatus.HealthStatus.OK if "[OK]" in heartbeat_service_value else NodeStatus.HealthStatus.NOK
        meteringDetails.loggingServiceRunning=NodeStatus.HealthStatus.OK if "[OK]" in logging_service_value else NodeStatus.HealthStatus.NOK
        if platform.system() == "Linux":
            bootTs, err, retCode = command_helper.run_command("stat -c %Z /proc/")
            if retCode != 0:
                logger.debug("systemBootTimestamp will be missing from metering status json response because of: ret code: %s, stdErr: %s" % (str(retCode), str(err)))
            else:
                # todo check format
                meteringDetails.systemBootTimestamp=int(bootTs.rstrip().decode())
    else:
        logger.info("Required services:")
        logger.info(servicesTable)

    if logging_agent_exists and metering_service_exists:
        if not json_format:
            logger.info("Required configs:")
        configs_state_list=list()
        heartbeat_conf_state=telemetry_logger.success_message("[OK]") if os.path.exists(heartbeat_conf) else telemetry_logger.fail_message("[NOK]")
        configs_state_list.append((heartbeat_conf, heartbeat_conf_state))
        logging_agent_conf_state=telemetry_logger.success_message("[OK]") if os.path.exists(logging_agent_conf) else telemetry_logger.fail_message("[NOK]")
        configs_state_list.append((logging_agent_conf, logging_agent_conf_state))
        configsResponse=response.ResponseHolder(tuple_content=configs_state_list)
        configsTable=configsResponse.to_table(headers=['Config location', 'Exists'])
        if not json_format:
            logger.info(configsTable)
    if logging_agent_exists and logging_agent_log_dir:
        metering_buffer_size=0
        buffers_size_map={}
        if os.path.exists(logging_agent_log_dir):
            files_in_var_log=os.listdir(logging_agent_log_dir)
            for fp in files_in_var_log:
                path_of_buffer=os.path.join(logging_agent_log_dir, fp)
                if os.path.isdir(path_of_buffer) and fp != "pos":
                    buffer_directory = Path(path_of_buffer)
                    buffer_size=sum(f.stat().st_size for f in buffer_directory.glob('**/*') if f.is_file())
                    buffers_size_map[fp]=buffer_size
        if buffers_size_map and "metering_databus" in buffers_size_map:
            metering_buffer_size=buffers_size_map["metering_databus"]
        if json_format:
            meteringDetails.loggingAgentBufferSize=metering_buffer_size
        else:
            buffer_details_map={}
            buffer_details_map["Location"]=os.path.join(logging_agent_log_dir, "metering_databus")
            buffer_details_map["Size"]=metering_buffer_size
            bufferDetailsResponse=response.ResponseHolder(map_content=buffer_details_map)
            bufferDetailsTable=bufferDetailsResponse.to_table()
            logger.info("Buffer details:")
            logger.info(bufferDetailsTable)
    
    if metering_service_exists:
        events=list()
        if os.path.exists(heartbeat_conf):
            heartbeat_config_obj=configparser.ConfigParser()
            heartbeat_config_obj.read(heartbeat_conf)
            events=heartbeat_config_obj.items("Event")
        eventResponse=response.ResponseHolder(tuple_content=events)
        eventTable=eventResponse.to_table(headers=['Event field name', 'Event field value'])
        if json_format and events:
            for k, v in events:
                meteringDetails.eventDetails[k]=v
            #meteringDetails.eventDetails.extend(dict((x, y) for x, y in events))
        else:
            logger.info("Heartbeat events:")
            logger.info(eventTable)

    if logging_agent_exists and meteringDatabusConf.databusEndpoint:
        datbus_info=list()
        databus_reachable=connection.connected(meteringDatabusConf.databusEndpoint, proxySettings=proxySettings)
        databus_test_record_resp=telemetry_logger.bold_message("[UNKNOWN]")
        if json_format:
            meteringDetails.databusTestResponse=NodeStatus.HealthStatus.UNKNOWN
        databus_reachable_val=telemetry_logger.fail_message("[NOK]")
        if json_format:
            meteringDetails.databusReachable=NodeStatus.HealthStatus.OK if databus_reachable else NodeStatus.HealthStatus.NOK
        if databus_reachable:
            databus_test_record_resp=databus.test_put_record(meteringDatabusConf.databusCredentilFile, meteringDatabusConf.databusEndpoint, meteringDatabusConf.databusProfile, meteringDatabusConf.databusAccessKeyField, meteringDatabusConf.databusPrivateKeyField, proxySettings=proxySettings, silenced=json_format)
            databus_reachable_val=telemetry_logger.success_message("[OK]")
            if json_format:
                grpc_databus_test_record_resp_enum=databus_test_record_resp if databus_test_record_resp == "OK" else "NOK"
                meteringDetails.databusTestResponse=NodeStatus.HealthStatus.Value(grpc_databus_test_record_resp_enum)
            databus_test_record_resp=telemetry_logger.success_message("[OK]") if databus_test_record_resp == "OK" else telemetry_logger.fail_message("[" + databus_test_record_resp + "]")
        datbus_info.append(('DataBus endpoint', meteringDatabusConf.databusEndpoint))
        datbus_info.append(('Proxy URL', meteringDatabusConf.proxyUrl if meteringDatabusConf.proxyUrl else "NO PROXY"))
        datbus_info.append(('Reachable', databus_reachable_val))
        datbus_info.append(('DataBus stream', meteringDatabusConf.databusStream))
        datbus_info.append(('Test put record response', databus_test_record_resp))
        dbusInfoResponse=response.ResponseHolder(tuple_content=datbus_info)
        dbusInfoTable=dbusInfoResponse.to_table()
        if json_format:
            meteringDetails.databusDetails["endpoint"]=meteringDatabusConf.databusEndpoint
            meteringDetails.databusDetails["proxyUrl"]=meteringDatabusConf.proxyUrl if meteringDatabusConf.proxyUrl else "NO PROXY"
            meteringDetails.databusDetails["stream"]=meteringDatabusConf.databusStream
        else:
            logger.info("Databus API info:")
            logger.info(dbusInfoTable)
    
        dbusHeadersMap={}
        if meteringDatabusConf.databusHeaders:
            databusHeadersArr=meteringDatabusConf.databusHeaders.split(",")
            for dbusHeaderKeyValue in databusHeadersArr:
                dbusHeaderArr=dbusHeaderKeyValue.split(":", 1)
                dbusHeadersMap[dbusHeaderArr[0]]=dbusHeaderArr[1]
        dbusHeadersResponse=response.ResponseHolder(map_content=dbusHeadersMap)
        dbusHeadersTable=dbusHeadersResponse.to_table()
        if not json_format:
            logger.info("Databus metering headers:")
            logger.info(dbusHeadersTable)
    elif logging_agent_exists:
        if not json_format:
            logger.info(telemetry_logger.fail_message("Cannot find databus endpoint in databus metering configuration. Probably this cluster type does not support metering?"))

    heartbeat_json_pattern="%s*" % heartbeat_json_file
    heartbeat_json_info=list()
    if os.path.exists(heartbeat_json_file):
        heartbeat_json_pattern="%s*" % heartbeat_json_file
        heartbeat_json_info=list()
        for heartbeat_file in glob.glob(heartbeat_json_pattern, recursive=True):
            rawcount=rawgencount(heartbeat_file)
            timeStr="NONE"
            if rawcount > 0:
                fline=open(heartbeat_file).readline().rstrip()
                jsonEvent=json.loads(fline)
                if "timestamp" in jsonEvent:
                    tsStr=jsonEvent["timestamp"]
                    timestamp = int(int(tsStr) / 1000)
                    value = datetime.datetime.fromtimestamp(timestamp)
                    timeStr=f"{value:%Y-%m-%d %H:%M:%S}"
                    if json_format:
                        meteringDetails.firstHeartbeatEventTimestamp=timestamp
            heartbeat_json_info.append((heartbeat_file, rawcount, timeStr))
            if json_format:
                meteringDetails.heartbeatEventCount=rawcount

        heartbeatJsonResponse=response.ResponseHolder(tuple_content=heartbeat_json_info)
        heartbeatJsonTable=heartbeatJsonResponse.to_table(headers=['Filename', 'Row count', 'First event date'])
        if not json_format:
            logger.info("Heartbeat JSON file(s) info:")
            logger.info(heartbeatJsonTable)
    else:
        if not json_format:
            logger.info(telemetry_logger.fail_message("Cannot parse %s - could mean problems with the metering service. Please check syslog of the service." % heartbeat_json_file))
    if json_format:
        nowTs=int(datetime.datetime.now().strftime('%s'))
        meteringDetails.timestamp=nowTs
        meteringDetailsProtoDict = MessageToDict(meteringDetails, including_default_value_fields=True)
        meteringJsonOutput=response.ResponseHolder(map_content=meteringDetailsProtoDict).to_json(pretty=pretty, sortKeys=True)
        logger.info(meteringJsonOutput)
        doctor_output.save_doctor_status(meteringJsonOutput, "metering")
        save_openmetrics_format(meteringDetailsProtoDict, metrics_folder)

def save_openmetrics_format(protoDict, metrics_dir):
    metrics_dir="dist"
    prefix="cdp_doctor_metering_"
    result={}
    if not os.path.exists(metrics_dir):
        return
    if "databusDetails" in protoDict and "endpoint" in protoDict["databusDetails"] and "proxyUrl" in protoDict["databusDetails"]:
        proxy_info="No" if "NO PROXY" == protoDict["databusDetails"]["proxyUrl"] else "Yes"
        result[prefix + "databus_details{databus_endpoint=\"%s\", proxy=\"%s\"}" % (protoDict["databusDetails"]["endpoint"], proxy_info)]=0
    if "heartbeatEventCount" in protoDict:
        result[prefix + "heartbeat_event_count"]=protoDict["heartbeatEventCount"]
    add_health_info(protoDict, result, "databusReachable", "databus_reachable", prefix)
    add_health_info(protoDict, result, "databusTestResponse", "databus_test_response", prefix)
    add_health_info(protoDict, result, "loggingAgentConfig", "logging_agent_config", prefix)
    add_health_info(protoDict, result, "loggingServiceRunning", "logging_service_running", prefix)
    add_health_info(protoDict, result, "heartbeatConfig", "heartbeat_config", prefix)
    add_health_info(protoDict, result, "heartbeatAgentRunning", "heartbeat_agent_running", prefix)
    if result:
        result_str=""
        for k in result:
            result_str+="%s %d\n" % (k, result[k])
        metrics_str=result_str.rstrip()
        pid = os.getpid()
        final_file=os.path.join(metrics_dir, "doctor_metering.prom")
        temp_file="%s.%s" % (final_file, pid)
        for file in glob.glob("%s.*" % final_file, recursive=False):
            os.remove(file)
        with open(temp_file, 'w') as fw:
            fw.write(metrics_str)
        if os.path.exists(temp_file):
            shutil.move(temp_file, final_file)
        
def add_health_info(protoDict, result, key, metrics_key, prefix):
    if key in protoDict:
        if protoDict[key] == "OK":
            result[prefix + metrics_key + "{doctor_status=\"OK\"}"]=0
        elif protoDict[key] == "NOK":
            result[prefix + metrics_key + "{doctor_status=\"NOK\"}"]=1
        else:
            result[prefix + metrics_key +"{doctor_status=\"UNKNOWN\"}"]=2

def _fill_missing_fields_as_unknown(map, fields):
    for field in fields:
        if field not in map:
            map[field]="UNKNOWN"
    return map

def _fill_missing_fields_as_empty_map(map, fields):
    for field in fields:
        if field not in map:
            map[field]={}
    return map

def _make_gen(reader):
    b = reader(1024 * 1024)
    while b:
        yield b
        b = reader(1024*1024)

def rawgencount(filename):
    f = open(filename, 'rb')
    f_gen = _make_gen(f.raw.read)
    return sum( buf.count(b'\n') for buf in f_gen )

class MeteringDatabusSettings:

    def __init__(self, logging_agent_conf, dbus_credential_file):
        use_td_agent=False
        if logging_agent_conf is None:
            if os.path.exists("/etc/cdp-logging-agent"):
                logging_agent_conf = "/etc/cdp-logging-agent/databus_metering.conf"
            elif os.path.exists("/etc/td-agent"):
                use_td_agent=True
                logging_agent_conf = "/etc/td-agent/databus_metering.conf"
        self.databusEndpoint=None
        self.proxyUrl=None
        self.proxyUser=None
        self.proxyPassword=None
        self.databusCredentilFile=None
        self.databusStream=None
        self.databusHeaders=None
        self.databusProfile=None
        self.databusAccessKeyField='cdp_access_key_id'
        self.databusPrivateKeyField='cdp_private_key'
        if os.path.exists(logging_agent_conf):
            with open(logging_agent_conf) as f:
                for line in f:
                    lineStripped=line.strip()
                    splitted=lineStripped.split()
                    if len(splitted) > 1:
                        if splitted[0] == "endpoint":
                            self.databusEndpoint=splitted[1].replace('"', '')
                        if splitted[0] == "proxy_url":
                            self.proxyUrl=splitted[1].replace('"', '')
                        if splitted[0] == "proxy_username":
                            self.proxyUser=splitted[1].replace('"', '')
                        if splitted[0] == "proxy_password":
                            self.proxyPassword=splitted[1].replace('"', '')
                        if splitted[0] == "headers":
                            self.databusHeaders=splitted[1]
                        if splitted[0] == "stream_name":
                            self.databusStream=splitted[1]
                        if splitted[0] == "credential_file":
                            self.databusCredentilFile=splitted[1]
                            self.databusProfile="dbus"
        if dbus_credential_file:
            self.databusCredentilFile=dbus_credential_file
            self.databusProfile='default'
        else:
            self.databusAccessKeyField='databus_access_key_id'
            self.databusPrivateKeyField='databus_access_secret_key'
    
    def databusEndpoint(self):
        return self.databusEndpoint

    def proxyUrl(self):
        return self.proxyUrl

    def proxyUser(self):
        return self.proxyUser

    def proxyPassword(self):
        return self.proxyPassword

    def databusHeaders(self):
        return self.databusHeaders

    def databusStream(self):
        return self.databusStream

    def databusCredentilFile(self):
        return self.databusCredentilFile

    def databusProfile(self):
        return self.databusProfile
    
    def databusAccessKeyField(self):
        return self.databusAccessKeyField

    def databusPrivateKeyField(self):
        return self.databusPrivateKeyField