#!/usr/bin/env python

import datetime
import os
import shutil
import subprocess
import socket
import logging
import json

from google.protobuf.json_format import MessageToDict
from nodestatus.generated import nodestatus_pb2 as NodeStatus
from utils import telemetry_logger, response, version, doctor_output, command_helper

logger = logging.getLogger(__name__)

class DoctorService:

    def __init__(self, format, pretty=False):
        self.format = format            # Instance Variable
        self.pretty = pretty
        self.default_cm_service_names = 'cm-agent'
        self.default_cm_systemd_service_names = 'cloudera-scm-agent'
        self.default_freeipa_service_names = 'healthagent,dirsrv,tomcat,kadmind,named-pkcs11,krb5kdc,ipa-dnskeysyncd,ipa-custodia'
        self.default_freeipa_systemd_service_names = 'cdp-freeipa-healthagent,krb5kdc'    
        self.default_cb_service_names = 'nginx,sshd,sssd,salt-bootstrap,salt-minion,cdp-logging-agent'
        self.default_salt_master_services = 'salt-api,salt-master'
    
    def run(self, serviceNames):
        result=self.collect(serviceNames)
        if not self.is_not_json_format():
            logger.info(result)
            doctor_output.save_doctor_status(result, "service")

    def systemd_report(self, serviceNames, metricsDir):
        # todo:
        #function get_memory() {
 # mem_in_kb=$(ps v $1 | tail -1 | tr -s " " | cut -d " " -f9)
 # mem_in_bytes=$(($mem_in_kb*1024))
 # echo "$mem_in_bytes"
#}

#function get_children_memory() {
#  mem_in_kb=$(ps -o pid,rss --no-headers --ppid $1 | cut -d ' ' -f3 | awk 'NF{sum+=$1} END {print sum}')
#  if [[ -z "${mem_in_kb// }" ]]; then
#    echo "0"
#  else
#    mem_in_bytes=$(($mem_in_kb*1024))
#    echo "$mem_in_bytes"
#  fi
#}      
        hostname = socket.gethostname()
        systemd_services_list_str=serviceNames
        if not serviceNames:
            if self.is_freeipa_node(hostname):
                if self.is_not_json_format():
                    logger.info("[START Freeipa checks]")
                systemd_services_list_str=self.default_freeipa_systemd_service_names
            elif not serviceNames: 
                systemd_services=self.default_cm_systemd_service_names
                systemd_services_list_str=self.append_to_service_list(systemd_services, "cloudera-scm-server", "/etc/cloudera-scm-server/cm.settings")
        systemd_services_list = systemd_services_list_str.split(",")
        results=[]
        for systemd_service in systemd_services_list:
            active_cmd="systemctl is-active --quiet %s" % systemd_service
            logger.info("will run command: %s", active_cmd)
            ret=self.get_cmd_return_code(active_cmd)
            logger.info("result: %d" % ret)
            pass
    
    def collect(self, serviceNames):
        hostname = socket.gethostname()
        colored_hostname = telemetry_logger.header_message(hostname)
        binaryVersion=version.get_binary_version()
        servicesDetails=NodeStatus.ServicesDetails(cdpTelemetryVersion=binaryVersion)
        if self.is_not_json_format():
            logger.info("[START service checks for node {0}]\n".format(colored_hostname))
        else: 
            logger.debug("[START service checks for node {0}]\n".format(colored_hostname))
        if serviceNames is None:
            if self.is_not_json_format():
                logger.info("[START infra service checks]")
            else:
                logger.debug("[START infra service checks]")
            cb_services_str_list=self.default_cb_service_names
            cb_services_str_list=self.append_to_service_list(cb_services_str_list, self.default_salt_master_services, "/srv/salt")
            default_cb_services=self.get_service_report_by_list(cb_services_str_list)
            if not self.is_not_json_format():
                infraServices=self.create_servies_list_objects(default_cb_services)
                servicesDetails.infraServices.extend(infraServices)
            if self.is_freeipa_node(hostname):
                if self.is_not_json_format():
                    logger.info("[START Freeipa checks]")
                freeipa_services=self.get_service_report_by_list(self.default_freeipa_service_names)
                if not self.is_not_json_format():
                    freeIpaServices=self.create_servies_list_objects(freeipa_services)
                    servicesDetails.freeipaServices.extend(freeIpaServices)
            else:
                if self.is_not_json_format():
                    logger.info("[START CM checks]")
                cm_services_str_list=self.default_cm_service_names
                cm_services_str_list=self.append_to_service_list(cm_services_str_list, "cm-server", "/etc/cloudera-scm-server/cm.settings")
                cm_services_str_list=self.append_to_service_list(cm_services_str_list, "knox", "/var/log/knox")
                cm_services=self.get_service_report_by_list(cm_services_str_list)
                if not self.is_not_json_format():
                    cmServices=self.create_servies_list_objects(cm_services)
                    servicesDetails.cmServices.extend(cmServices)
        else:
            if self.is_not_json_format():
                logger.info("[START custom Service checks]")
            else:
                logger.debug("[START custom Service checks]")
            services_list=self.get_service_report_by_list(serviceNames)
            if not self.is_not_json_format():
                servicesList=self.create_servies_list_objects(services_list)
                servicesDetails.services.extend(servicesList)
        if self.is_not_json_format():
            logger.info("[END service checks]")
        else:
            nowTs=int(datetime.datetime.now().strftime('%s'))
            servicesDetails.timestamp=nowTs
            logger.debug("[END service checks]")
        servicesDetailsProtoDict = MessageToDict(servicesDetails)
        return response.ResponseHolder(map_content=servicesDetailsProtoDict).to_json(pretty=self.pretty, sortKeys=True)

    def is_freeipa_node(self, hostname) -> bool:
        #please make me smarter
        return "ipaserver" in hostname

    def append_to_service_list(self, service_name_list_str, service_names_to_add, path_to_check):
        if os.path.exists(path_to_check):
            service_name_list_str="%s,%s" % (service_name_list_str, service_names_to_add)
        return service_name_list_str

    def create_servies_list_objects(self, services_tuple_list):
        serviceStatuses=[]
        for service_tuple in services_tuple_list:
            serviceStatus=NodeStatus.ServiceStatus()
            serviceStatus.name=service_tuple[0]
            serviceStatus.status=NodeStatus.HealthStatus.Value(service_tuple[1])
            serviceStatuses.append(serviceStatus)
        return serviceStatuses

    def get_all_service_report(self):

        logger.info("Getting ALL service statuses")
        os.system('ps aux | grep sshd | grep -v grep | wc -l')

    def get_service_report_by_list(self, serviceNames):
        serviceArr = serviceNames.split(",")
        results = []
        for serviceName in serviceArr:
            result_pair = []
            result_pair.append(serviceName)
            result_pair.append(self.get_service_report_by_name(serviceName))
            results.append(result_pair)
        if self.is_not_json_format():
            self.print_table_format_with_format(results)
        return results

    def get_service_report_by_name(self, serviceName):
        result = self.get_concerete_service_status(serviceName)
        logger.debug("Result of service status check (%s): %s" % (serviceName, str(result)))
        if self.is_not_json_format():
            if result == 0:
                return telemetry_logger.success_message("[OK]")
            else:
                return telemetry_logger.fail_message("[NOK]")
        return "OK" if result == 0 else "NOK"


    def get_concerete_service_status(self, serviceName):
        cmd='ps aux | grep -v grep | grep -v cli.py | grep -v cdp-telemetry | grep -v cdp-doctor | grep -q ' +serviceName
        return self.get_cmd_return_code(cmd)

    def get_cmd_return_code(self, cmd):
        _,_, ret = command_helper.run_command(cmd, True)
        return ret

    def print_table_format_with_format(self, results):
        logger.info(telemetry_logger.format_table_with_format(results, ['ServiceName', 'Status'], self.format))
    
    def is_not_json_format(self):
        return not self.format or not self.format.lower() == "json"