#!/usr/bin/env python
import argparse
import sys
import logging
import os
import yaml
import shutil
import datetime
import socket
import time
from pid import PidFile
from utils import telemetry_logger
from concurrent.futures import ThreadPoolExecutor
from filecollector.config import Config
from filecollector.log_processor import LogProcessor
from filecollector.event_processor import EventProcessor
from filecollector.workspace import Workspace
from filecollector.collected_files import CollectedFilesHolder

#telemetry_logger.setup_logger('filecollector', file_name='filecollector.log')
logger=logging.getLogger(__name__)

def parse_args(args):
    parser = argparse.ArgumentParser(
        description='Python script to collect logs to specific folder')
    parser.add_argument('--config', type=str, required=True,
                        help='Path to logcollector configuration')
    parser.add_argument('--labels', type=str, required=False,
                            help='Comma separated list of labels for filtering files for collection')
    parser.add_argument('--start-time', type=float, required=False, dest="start_time",
                        help='Start (last modified) datestamp (epoh unix format) for the monitored logs')
    parser.add_argument('--end-time', type=float, required=False, dest="end_time",
                        help='End (creation) datestamp (epoh unix format) for the monitored logs (epoh unix format)')
    args = parser.parse_args(args)
    return args

def main(args):
    args = parse_args(args)
    filteredLabels = args.labels.split(',') if args.labels else []
    startTime=args.start_time
    endTime=args.end_time
    run(args.config, args.labels, args.start_time, args.end_time)

def run(configOpt, labelsOpt=None, startTimeOpt=None, endTimeOpt=None):
    execution_start_time = datetime.datetime.now()
    filteredLabels = labelsOpt.split(',') if labelsOpt else []
    startTime=startTimeOpt
    endTime=endTimeOpt
    with open(configOpt) as file:
        config = yaml.load(file, yaml.SafeLoader)
        if config and "collector" in config:
            binPath="/opt/cdp-telemetry/bin"
            fileCollectorBinaryLocation="%s/filecollector" % binPath
            if not ("fluentProcessor" in config["collector"]) and os.path.exists(fileCollectorBinaryLocation):
                parameters=""
                if labelsOpt:
                    parameters+=" --labels %s" % str(labelsOpt)
                if startTimeOpt:
                    parameters+=" --start-time %s" % str(int(startTimeOpt))
                if endTimeOpt:
                    parameters+=" --end-time %s" % str(int(endTimeOpt))
                ldLibPathVar="LD_LIBRARY_PATH"
                if ldLibPathVar in os.environ:
                    currentValue=os.environ[ldLibPathVar]
                    os.environ[ldLibPathVar]="%s:%s" % (currentValue, binPath)
                else:
                    os.environ[ldLibPathVar]=binPath
                exit_code=_run_system_command("%s collect --config %s %s" % (fileCollectorBinaryLocation, configOpt, parameters))
                sys.exit(exit_code)
            logger.debug("Run with config: %s" % (config["collector"]))
            outputLocation=config["collector"]["outputLocation"]
            files=config["collector"]["files"]
            collectorConfig=Config(config["collector"])
            outputScript=collectorConfig.get_str_key("outputScript")
            preProcessScript=collectorConfig.get_str_key("preProcessScript")
            processFileScript=collectorConfig.get_str_key("processFileScript")
            useFullPath=collectorConfig.get_bool_key("useFullPath", True)
            compressFormat=collectorConfig.get_str_key("compressFormat", "zip")
            workspaceLimitBytes=collectorConfig.get_int_key("workspaceLimitBytes")
            now = datetime.datetime.today()
            nTime = now.strftime("%Y-%m-%d-%H-%M-%S-%f")
            hostname=None
            if socket.gethostname().find('.')>=0:
                hostname=socket.gethostname()
            else:
                hostname=socket.gethostbyaddr(socket.gethostname())[0]
            zipfile_name = nTime + "-" + hostname.replace(".", "-")
            tmp_folder=os.path.abspath(os.path.join(outputLocation, "tmp", zipfile_name))

            fluentEventProcessor = None
            if "fluentProcessor" in config["collector"]:
                fluentProcessorConfig=Config(config["collector"]["fluentProcessor"])
                fluent_host=fluentProcessorConfig.get_str_key("host", "localhost")
                fluent_port=fluentProcessorConfig.get_int_key("port", 24224)
                fluent_tag=fluentProcessorConfig.get_str_key("tag")
                additional_fields=config["collector"]["fluentProcessor"]["additionalFields"] if "additionalFields" in config["collector"]["fluentProcessor"] else {}
                additional_fields_key=fluentProcessorConfig.get_str_key("additionalFieldsKey", "additionalContext")
                message_field=fluentProcessorConfig.get_str_key("messageField", "message")
                include_time=fluentProcessorConfig.get_bool_key("includeTime")
                fluentEventProcessor=EventProcessor(fluent_host, int(fluent_port), fluent_tag, additional_fields, additional_fields_key, message_field, include_time)
            if workspaceLimitBytes is None and not os.path.exists(tmp_folder):
                os.makedirs(tmp_folder)
            if not os.path.exists(outputLocation):
                os.makedirs(outputLocation)
            collectedFilesList=__get_collected_files_list(files, config, startTime, endTime)
            if not collectedFilesList:
                logger.info("Not found anything to process.")
                return
            __disk_check(collectedFilesList, outputLocation, collectorConfig)
            
            workspace_index=0
            workspaces=list()
            current_workspace=None
            if workspaceLimitBytes is None:
                logger.debug("Use only 1 workspace as 'workspaceLimitBytes' is not set.")
                workspace=Workspace(tmp_folder, zipfile_name, config)
                workspaces.append(workspace)
                current_workspace=workspace
            else:
                logger.debug("Configuration 'workspaceLimitBytes' is set, workspace index will be used for workspace folders.")
                filename_suffix="-" + str(workspace_index)
                workspace=Workspace(tmp_folder + filename_suffix, zipfile_name + filename_suffix, config, workspace_index)
                workspaces.append(workspace)
                current_workspace=workspace

            total_disk_size=0
            number_of_patterns=len(collectedFilesList)
            for idx, collectedFiles in enumerate(collectedFilesList):
                log_processor=LogProcessor(collectedFiles, config, filteredLabels, fluentEventProcessor)
                current_workspace.add_log_processor(log_processor)
                if (number_of_patterns - 1 > idx) and workspaceLimitBytes is not None:
                    disk_size=collectedFiles.get_total_size()
                    total_disk_size=total_disk_size+disk_size
                    if total_disk_size > workspaceLimitBytes:
                        workspace_index=workspace_index+1
                        filename_suffix="-" + str(workspace_index)
                        workspace=Workspace(tmp_folder + filename_suffix, zipfile_name + filename_suffix, config, workspace_index)
                        workspaces.append(workspace)
                        current_workspace=workspace
                        logger.debug("Increased workspace folder index will be used if needed: %s" % str(workspace_index))
                        total_disk_size=0

            workspaceWorkers=collectorConfig.get_int_key("workspaceWorkers", 1)
            with ThreadPoolExecutor(max_workers=workspaceWorkers) as workspaceExecutor:
                for workspace in workspaces:
                    workspaceExecutor.submit(workspace.process())

            if fluentEventProcessor:
                fluentEventProcessor.close()
    execution_end_time = datetime.datetime.now()
    logger.info('Execution time: {}'.format(execution_end_time - execution_start_time))

def __get_collected_files_list(files, config, startTime, endTime):
    collectedFilesList=[]
    for fileObj in files:
        fileObjConfig=Config(fileObj)
        sortFilesByDate=not "sortFilesByDate" in config["collector"] or bool(config["collector"]["sortFilesByDate"])
        holder=CollectedFilesHolder(fileObj, startTime, endTime, sortFilesByDate)
        if holder.files:
            logger.debug("Process collected files for label %s." % fileObj["label"])
            collectedFilesList.append(holder)
        else:
            logger.debug("Skipping adding collected files for label: %s" % fileObj["label"])
    return collectedFilesList

def __disk_check(collectedFilesList, outputLocation, config):
    checkDiskSpace=config.get_bool_key("checkDiskSpace", True)
    requiredDiskSpaceRatio=config.get_float_key("requiredDiskSpaceRatio", 1.0)
    if checkDiskSpace:
        logger.debug("disk space check is enabled")
        fullSize=0
        for collectedFiles in collectedFilesList:
            fullSize = fullSize + collectedFiles.get_total_size()
        freeSpace=0
        requiredFreeSpace=0
        try:
            total, used, freeSpace = shutil.disk_usage(outputLocation)
        except Exception as e:
            import psutil
            hdd = psutil.disk_usage(outputLocation)
            freeSpace=hdd.free
        requiredFreeSpace=int(fullSize * requiredDiskSpaceRatio)
        if fullSize > freeSpace:
            logger.error("there is not enough free space for file collection - free space: %d, required space: %d" % (freeSpace, requiredFreeSpace))
            sys.exit(1)
        else:
            logger.debug("free disk space: %d, required disk space: %d" % (freeSpace, requiredFreeSpace))
    else:
        logger.debug("disk space check is disabled")

def _run_system_command(command):
    return_value = os.system(command)
    return int(bin(return_value).replace("0b", "").rjust(16, '0')[:8], 2)

if __name__ == "__main__":
    pidfile=os.environ.get('FILECOLLECTOR_PIDFILE', 'filecollector-collector.pid')
    with PidFile(pidfile) as p:
        main(sys.argv[1:])