import logging
import os
import shutil
import subprocess
from filecollector.anonymizer import FileAnonymizer
from filecollector.archiver import Archiver 
from filecollector.config import Config
from filecollector.external_command import ExternalCommand

logger=logging.getLogger(__name__)

class LogProcessor:
    
    def __init__(self, collectedFiles, config, filteredLabels, fluentEventProcessor):
        self.collectedFiles=collectedFiles
        collectorConfig=Config(config["collector"])
        fileObjectConfig=collectedFiles.get_common_config()
        self.outputScript=collectorConfig.get_str_key("outputScript")
        self.usePerlScriptForAnonymization=collectorConfig.get_bool_key("usePerlScriptForAnonymization")
        self.useExternalCommandForCompression=collectorConfig.get_bool_key("useExternalCommandForCompression")
        self.preProcessScript=collectorConfig.get_str_key("preProcessScript")
        self.processFileScript=collectorConfig.get_str_key("processFileScript")
        self.useFullPath=collectorConfig.get_bool_key("useFullPath", True)
        self.compressFormat=collectorConfig.get_str_key("compressFormat", "zip")
        self.compress=fileObjectConfig.get_bool_key("compress", False)
        self.deleteProcessedTempFilesOneByOne=collectorConfig.get_bool_key("deleteProcessedTempFilesOneByOne")
        self.label=fileObjectConfig.get_str_key("label")
        self.path=fileObjectConfig.get_str_key("path")
        self.filteredLabels=filteredLabels
        self.folderPrefix=fileObjectConfig.get_str_key("folderPrefix", None)
        self.mandatory=fileObjectConfig.get_bool_key("mandatory", False)
        self.fluentEventProcessor=fluentEventProcessor
        self.skipLabelFromPath=fileObjectConfig.get_bool_key("skipLabelFromPath", False)
        self.useFullPathPerFile=fileObjectConfig.get_bool_key("useFullPath", True)
        self.skipAnonymization=fileObjectConfig.get_bool_key("skipAnonymization", False)
        self.rules=config["collector"]["rules"] if "rules" in config["collector"] and config["collector"]["rules"] else []
        self.fileSizeLimitInMB=collectorConfig.get_int_key("fileSizeLimitInMB", 50)
        self.numberOfLinesToKeepOnLimit=collectorConfig.get_int_key("numberOfLinesToKeepOnLimit", 1000000)
    
    def get_label(self):
        return self.label

    def process(self, workspace_folder):
        try:
            self.__process(workspace_folder)
            return (self.label, 0)
        except Exception as e:
            logger.error('Error during file processing.', exc_info=e)
            return (self.label, 1)

    def __process(self, workspace_folder):
        if not workspace_folder:
            raise Exception("Destionation (workspace) folder parameter is required!")
        if not self.mandatory and self.filteredLabels and self.label not in self.filteredLabels:
           return
        dest_folder=None
        labelInPath=self.label.lower()
        if self.skipLabelFromPath:
            labelInPath=""
        if self.folderPrefix:
            dest_folder=os.path.join(workspace_folder, self.folderPrefix, labelInPath)
        else:
            dest_folder=os.path.join(workspace_folder, labelInPath)
        allFileObjects=self.collectedFiles.get_files()
        for fileObj in allFileObjects:
            file=fileObj.get_path()
            absFilePath=fileObj.get_abs_path()
            logger.debug("process file: %s" % absFilePath)
            dest=os.path.join(dest_folder, os.path.abspath(file).lstrip(os.sep)) if self.useFullPathPerFile else os.path.join(dest_folder, os.path.basename(file))
            dest_parent=os.path.dirname(dest)
            if not os.path.exists(dest_parent):
                os.makedirs(dest_parent)
            if os.path.isfile(file):
                file_size=os.path.getsize(file)
                logger.debug("File '%s' size in Bytes is: %s Bytes" % (file, file_size))
                if file_size > self.fileSizeLimitInMB * (1024 * 1024):
                    logger.debug("File '%s' is larger than %s MB, only the last %s lines will be processed." % (file, str(self.fileSizeLimitInMB), str(self.numberOfLinesToKeepOnLimit)))
                    ExternalCommand("tail -n %s %s > %s" % (str(self.numberOfLinesToKeepOnLimit), str(file), str(dest))).run(doNotSplit=True)
                else:
                    shutil.copy(file, dest)
            if not self.skipAnonymization and self.rules:
                FileAnonymizer(dest, self.rules, self.label, self.usePerlScriptForAnonymization).redact()
            if self.processFileScript:
                subprocess.call([self.processFileScript, dest, self.label])
            if self.fluentEventProcessor:
                self.fluentEventProcessor.process(self.label, os.path.abspath(file), dest)
            if self.deleteProcessedTempFilesOneByOne:
                os.remove(dest)
        if self.compress:
            logger.debug("Compressionn is enabled for folder: %s" % dest_folder)
            archiver=Archiver(dest_folder, dest_folder, self.compressFormat, True, self.useExternalCommandForCompression)
            archiver.make_archive()