import os
import shutil
import logging
import subprocess
import json
import concurrent
from concurrent.futures import ThreadPoolExecutor
from filecollector.archiver import Archiver
from filecollector.config import Config
from filecollector.log_processor import LogProcessor

logger=logging.getLogger(__name__)

class Workspace:

    def __init__(self, workspace_folder, zipfile_name, config, workspace_index=0):
        self.workspace_folder=workspace_folder
        self.zipfile_name=zipfile_name
        self.workspace_index=workspace_index
        self.outputLocation=config["collector"]["outputLocation"]
        collectorConfig=Config(config["collector"])
        self.includeWorkspaceMetadata=collectorConfig.get_bool_key("includeWorkspaceMetadata", True)
        self.workspaceMetadataFileName=collectorConfig.get_str_key("workspaceMetadataFileName", "filecollector-metadata.json")
        self.workspaceMetadataExtraFields=collectorConfig.get_key("workspaceMetadataExtraFields", None)
        self.additionalWorkspaceMetadata=collectorConfig.get_key("additionalWorkspaceMetadata", None)
        self.additionalWorkspaceMetadataFileName=collectorConfig.get_str_key("additionalWorkspaceMetadataFileName", "bundle-info.json")
        self.useExternalCommandForCompression=collectorConfig.get_bool_key("useExternalCommandForCompression")
        self.compressFormat=collectorConfig.get_str_key("compressFormat", "zip")
        self.outputScript=collectorConfig.get_str_key("outputScript")
        self.preProcessScript=collectorConfig.get_str_key("preProcessScript")
        self.processFilesFolderScript=collectorConfig.get_str_key("processFilesFolderScript")
        self.skipCompress=not collectorConfig.get_bool_key("compress", True)
        self.keepProcessedFiles=not collectorConfig.get_bool_key("deleteProcessedTempFiles", True)
        self.deleteCompressedFile=collectorConfig.get_bool_key("deleteCompressedFile")
        self.logProcessorWorkers=collectorConfig.get_int_key("logProcessorWorkers", 1)
        self.log_processors=list()

    def get_workspace_folder(self):
        return self.workspace_folder

    def add_log_processor(self, log_processor):
        self.log_processors.append(log_processor)

    def process(self):
        if not os.path.exists(self.workspace_folder):
            os.makedirs(self.workspace_folder)
        if self.preProcessScript:
            subprocess.call([self.preProcessScript, self.workspace_folder])
        results = {}
        with ThreadPoolExecutor(max_workers=self.logProcessorWorkers) as logProcessorExecutor:
            futures = []
            for log_processor in self.log_processors:
                futures.append(logProcessorExecutor.submit(log_processor.process, self.workspace_folder))
            for future in concurrent.futures.as_completed(futures):
                try:
                    result = future.result()
                    outcome = "ERROR" if result[1] != 0 else "OK"
                    results[result[0]] = outcome
                except Exception as e:
                    logger.error('Error during gathering log processor results for workspace: %s', exc_info=e)
        if self.includeWorkspaceMetadata:
            logger.debug("Start processing workspace metadata files for %s" % self.workspace_folder)
            metadata = {}
            metadata["workspaceIndex"]=self.workspace_index
            metadata["workspaceFolder"]=self.workspace_folder
            labels=[]
            for log_processor in self.log_processors:
                labels.append(log_processor.get_label())
            metadata["labels"]=labels
            metadata["results"]=results
            if self.workspaceMetadataExtraFields:
                for k, v in self.workspaceMetadataExtraFields.items():
                    metadata[k]=v
            workspaceMetadataFile=os.path.join(self.workspace_folder, self.workspaceMetadataFileName)
            with open(workspaceMetadataFile, "w") as metadata_file:
                json.dump(metadata, metadata_file, indent=4, sort_keys=True)

        if self.additionalWorkspaceMetadata:
            additionalWorkspaceMetadataFile=os.path.join(self.workspace_folder, self.additionalWorkspaceMetadataFileName)
            logger.debug("Found additional workspace metadata, writing to %s" % additionalWorkspaceMetadataFile)
            jsonWorkspaceMetadataStr=json.dumps(self.additionalWorkspaceMetadata)
            valuesToReplace = {
                "workspaceFolder": self.workspace_folder,
                "workspaceIndex": self.workspace_index
            }
            for k, v in valuesToReplace.items():
                jsonWorkspaceMetadataStr = jsonWorkspaceMetadataStr.replace("${%s}" % k, str(v))
            jsonWorkspaceMetadataObj=json.loads(jsonWorkspaceMetadataStr)
            with open(additionalWorkspaceMetadataFile, "w") as add_metadata_file:
                json.dump(jsonWorkspaceMetadataObj, add_metadata_file, indent=4, sort_keys=True)

        if self.processFilesFolderScript:
            subprocess.call([self.processFilesFolderScript, self.workspace_folder])
        archiver=None
        if self.skipCompress:
            logger.info("skipping file compression")
        else:
            output_file=os.path.join(self.outputLocation, self.zipfile_name)
            archiver=Archiver(self.workspace_folder, output_file, self.compressFormat, False, self.useExternalCommandForCompression)
            archiver.make_archive()

        if self.keepProcessedFiles:
            logger.info("keep processed files in '%s' folder" % os.path.join(self.outputLocation, "tmp"))
        else:
            shutil.rmtree(os.path.join(self.outputLocation, "tmp"))

        if not self.keepProcessedFiles and self.outputScript and archiver is not None:
            output_compressed_file="%s.%s" % (output_file, archiver.get_extension())
            subprocess.call([self.outputScript, output_compressed_file])
            if self.deleteCompressedFile:
                os.remove(output_compressed_file)