import cm_client
import os
import json
import logging
import signal
import socket
import sys
import time
import datetime
import traceback
from abc import ABC, abstractmethod
from cm_client.rest import ApiException
from pprint import pprint
from threading import Event, Timer, Thread
from logging.handlers import RotatingFileHandler
from .system_metrics_collector import SystemMetricsCollectorAgent
from .metrics_logger import MetricsLogger, MetricsStatusLogger, NullProcessor
from nodestatus.generated import nodestatus_pb2 as NodeStatus
from google.protobuf.json_format import MessageToDict
from utils import version

import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
logging.getLogger("urllib3").setLevel(logging.ERROR)

logger=logging.getLogger(__name__)

class ClouderaManagerMetricsCollector:

    def __init__(self, config):
        self.config = config
        logger.debug("Load configuration: " + str(config))
        self.sharedDataHolder=SharedDataHolder(config)
        self._setup_termination_handler()
        if "systemMetricsCollector" in config:
            self.system_metrics_collector_agent = SystemMetricsCollectorAgent(config, False)
            self.system_metrics_collector_agent.start()
        else:
            self.system_metrics_collector_agent = None
        cm_client.configuration.username = config["clouderaManager"]["user"]
        cm_client.configuration.password = config["clouderaManager"]["password"]
        cm_client.configuration.verify_ssl = False
        self.cluster_type = config["clouderaManager"]["clusterType"]
        if socket.gethostname().find('.') >= 0:
            api_host = socket.gethostname()
        else:
            api_host = socket.gethostbyaddr(socket.gethostname())[0]
        protocol = config["clouderaManager"]["protocol"]
        port = config["clouderaManager"]["port"]
        api_version = config["clouderaManager"]["apiVersion"]
        api_url = protocol + "://" + api_host + ':' + port + '/api/' + api_version
        self.api_client = cm_client.ApiClient(api_url)
        logger.info("Cloudera Manager Api URL: " + api_url)
        metrics_api_url = protocol + "://" + api_host + ':' + port + '/api/v5'
        self.metrics_api_client = cm_client.ApiClient(metrics_api_url)
        logger.info("Cloudera Manager Metrics Api URL: " + metrics_api_url)
        availableClusterPeriod=self.config['config']['availableClusterUpdatePeriod']
        self.cluster_retriever=ClusterRetriever(self.sharedDataHolder, availableClusterPeriod, cm_client, self.api_client)
        self.services_retriever=ServicesRetriever(self.sharedDataHolder, self.config['config']['availableServicesUpdatePeriod'], cm_client, self.api_client)
        self.available_metrics_retriever=AvailableMetricsRetriever(self.sharedDataHolder, self.config['config']['availableMetricsUpdatePeriod'], cm_client, self.api_client, self.metrics_api_client)
        self.metrics_retriever=MetricsRetriever(self.sharedDataHolder, self.config['config']['metricsCollectPeriod'], cm_client, self.api_client, self.metrics_api_client)
        
        self.get_cluster_timer = Timer(availableClusterPeriod, self._get_cluster_with_retries)
        
        self.servicesRetrieverThread=Thread(target=self.services_retriever, args=[])
        self.servicesRetrieverThread.setDaemon(True)
        self.servicesRetrieverThread.name = "services-retriever"
        
        self.availableMetricsRetrieverThread=Thread(target=self.available_metrics_retriever, args=[])
        self.availableMetricsRetrieverThread.setDaemon(True)
        self.availableMetricsRetrieverThread.name = "available-metrics-retriever"

        self.metricsRetrieverThread=Thread(target=self.metrics_retriever, args=[])
        self.metricsRetrieverThread.setDaemon(True)
        self.metricsRetrieverThread.name = "metrics-retriever"

    def start(self):
        logger.info("Starting Cloudera Manager Metrics Collector daemon.")
        self._get_cluster_with_retries()
        self.services_retriever.retrieve_data()
        self.available_metrics_retriever.retrieve_data()
        self.metrics_retriever.retrieve_data()
        self.servicesRetrieverThread.start()
        self.availableMetricsRetrieverThread.start()
        self.metricsRetrieverThread.start()
        self.get_cluster_timer.start()

    def _get_cluster_with_retries(self):
        from .retry import retry
        retry(self.cluster_retriever.retrieve_data, logger=logger, context="Gather cluster details", delay=30)

    def _setup_termination_handler(self):
        def termination_handler(signum, frame):
            logger.debug("CM Metrics collector termination called.")
            if self.system_metrics_collector_agent:
                system_metric_timer=self.system_metrics_collector_agent.get_timer()
                system_metric_timer.cancel()
            if self.get_cluster_timer:
                self.get_cluster_timer.cancel()
            sys.exit(143)
        signal.signal(signal.SIGINT, termination_handler)
        signal.signal(signal.SIGTERM, termination_handler)

class SharedDataHolder(object):

    def __init__(self, config):
        self.config=config
        self.cluster=None
        self.services=None
        self.available_metrics=None
        self.cluster_details=NodeStatus.ClusterDetails()
        globalFields = config["globalFields"]
        if "clusterCrn" in globalFields:
            self.cluster_details.crn=globalFields["clusterCrn"]
        if "clusterName" in globalFields:
            self.cluster_details.name=globalFields["clusterName"]
        if "clusterVersion" in globalFields:
            self.cluster_details.version=globalFields["clusterVersion"]
        if "clusterType" in globalFields:
            self.cluster_details.type=NodeStatus.ClusterDetails.ClusterType.Value(globalFields["clusterType"].upper())
        if "platform" in globalFields:
            self.cluster_details.platform=NodeStatus.ClusterDetails.CloudPlatform.Value(globalFields["platform"])
        services = config["services"]
        self.service_metrics_map_from_config = {}
        for service in services:
            name = service["name"]
            metrics = service["metrics"]
            self.service_metrics_map_from_config[name] = metrics

    def get_cluster_details(self):
        return self.cluster_details

    def get_service_metrics_map_from_config(self):
        return self.service_metrics_map_from_config
    
    def get_config(self):
        return self.config

    def get_cluster(self):
        return self.cluster

    def set_cluster(self, cluster):
        self.cluster=cluster

    def get_services(self):
        return self.services

    def set_services(self, services):
        self.services=services

    def get_available_metrics(self):
        return self.available_metrics

    def set_available_metrics(self, available_metrics):
        self.available_metrics=available_metrics

class AbstractDataRetriever(ABC):

    def __init__(self, sharedDataHolder, sleepTime, cm_client, api_client, metrics_api_client=None):
        self.sharedDataHolder=sharedDataHolder
        self.sleepTime=sleepTime
        self.cm_client=cm_client
        self.api_client=api_client
        self.metrics_api_client=metrics_api_client
        self.setup_processors(sharedDataHolder.get_config())
    
    def __call__(self):
        e = Event()
        while True:
            self.retrieve_data()
            e.wait(self.sleepTime)

    @abstractmethod
    def retrieve_data(self):
        pass

    def setup_processors(self, config):
        pass

    def clear_cluster_details(self, event_list):
        for event in event_list:
            event.ClearField("clusterDetails")
        return event_list

class ClusterRetriever(AbstractDataRetriever):
    
    def retrieve_data(self):
        self._get_cluster()

    def _get_cluster(self):
        cluster_api_instance = cm_client.ClustersResourceApi(self.api_client)
        api_response = cluster_api_instance.read_clusters(view='SUMMARY',  _request_timeout = 20.0)
        clusterResponse = None
        for cluster in api_response.items:
            config=self.sharedDataHolder.get_config()
            cluster_type=config["clouderaManager"]["clusterType"]
            if cluster.cluster_type == cluster_type:
                clusterResponse = cluster
        if clusterResponse:
            self.sharedDataHolder.set_cluster(clusterResponse)
        else:
            raise Exception("Not found any clusters yet.")

class ServicesRetriever(AbstractDataRetriever):
    
    def retrieve_data(self):
        try:
            self._get_services()
        except Exception as e:
            logger.debug("Error occurred during get_services operation: %s" % str(traceback.format_exc()))

    def setup_processors(self, config):
        self._setup_health_check_metrics_processor(config)
        self._setup_health_check_metrics_status_processor(config)

    def _get_services(self):
        services_api_instance = cm_client.ServicesResourceApi(self.api_client)
        api_response = services_api_instance.read_services(self.sharedDataHolder.get_cluster().name, view='FULL')
        self.sharedDataHolder.set_services(api_response.items)
        if self.sharedDataHolder.get_services():
            health_checks_list=[]
            for service in self.sharedDataHolder.get_services():
                if service.health_checks:
                    for health_check in service.health_checks:
                        if "jsonHealthChecksLogger" in self.sharedDataHolder.get_config():
                            health_check_event=NodeStatus.CmServiceHealthCheckEvent(clusterDetails=self.sharedDataHolder.get_cluster_details())
                        else:
                            health_check_event=NodeStatus.CmServiceHealthCheckEvent()
                        health_check_event.serviceName = service.name
                        health_check_event.serviceType = service.type
                        if health_check.explanation:
                            health_check_event.explanation = health_check.explanation
                        health_check_event.name = health_check.name
                        health_check_event.value = health_check.summary
                        health_check_event.timestamp = int(datetime.datetime.now().strftime('%s'))
                        health_checks_list.append(health_check_event)
                        self.health_checks_processor.process(MessageToDict(health_check_event))
            if "jsonHealthChecksLogger" in self.sharedDataHolder.get_config():
                health_checks_list=self.clear_cluster_details(health_checks_list)
            if health_checks_list:
                binaryVersion=version.get_binary_version()
                cmHealthChecksReport=NodeStatus.CmHealthChecksReport(healthChecks=health_checks_list, clusterDetails=self.sharedDataHolder.get_cluster_details(), cdpTelemetryVersion=binaryVersion)
                self.health_checks_status_processor.process(MessageToDict(cmHealthChecksReport))

    def _setup_health_check_metrics_processor(self, config):
        if "jsonHealthChecksLogger" in config:
            name = config["jsonHealthChecksLogger"]['name']
            path = config["jsonHealthChecksLogger"]['path']
            max_bytes = config["jsonHealthChecksLogger"]['max_bytes']
            backup_count = config["jsonHealthChecksLogger"]['backup_count']
            self.health_checks_processor = MetricsLogger(name, path, max_bytes, backup_count)
        else:
            self.health_checks_processor = NullProcessor()
    
    def _setup_health_check_metrics_status_processor(self, config):
        if "statusReportFiles" in config and "healthchecksFileLocation" in config["statusReportFiles"]:
            self.health_checks_status_processor = MetricsStatusLogger(config["statusReportFiles"]["healthchecksFileLocation"])
        else:
            self.health_checks_status_processor = NullProcessor()

class AvailableMetricsRetriever(AbstractDataRetriever):
    
    def retrieve_data(self):
        try:
            self._get_available_metrics()
        except Exception as e:
            logger.debug("Error occurred during get_available_metrics operation: %s" % str(traceback.format_exc()))

    def _get_available_metrics(self):
       if self.sharedDataHolder.get_services():
          new_available_metrics_map = {}
          for service in self.sharedDataHolder.get_services():
              services_api_instance = cm_client.ServicesResourceApi(self.metrics_api_client)
              metrics_api_response = services_api_instance.get_metrics(self.sharedDataHolder.get_cluster().name, service.name)
              new_metrics_list = []
              for m in metrics_api_response.items:
                 if m.data:
                    new_metrics_list.append(m.name)
              new_available_metrics_map[service.name] = new_metrics_list
          self.sharedDataHolder.set_available_metrics(new_available_metrics_map)

class MetricsRetriever(AbstractDataRetriever):

    def retrieve_data(self):
        try:
            self._get_metrics()
        except Exception as e:
            logger.debug("Error occurred during get_metrics operation: %s" % str(traceback.format_exc()))

    def setup_processors(self, config):
        self._setup_metrics_processor(config)
        self._setup_metrics_status_processor(config)

    def _setup_metrics_processor(self, config):
        if "jsonMetricsLogger" in config:
            name = config["jsonMetricsLogger"]['name']
            path = config["jsonMetricsLogger"]['path']
            max_bytes = config["jsonMetricsLogger"]['max_bytes']
            backup_count = config["jsonMetricsLogger"]['backup_count']
            self.metrics_processor = MetricsLogger(name, path, max_bytes, backup_count)
        else:
            self.metrics_processor = NullProcessor()

    def _setup_metrics_status_processor(self, config):
        if "statusReportFiles" in config and "metricsFileLocation" in config["statusReportFiles"]:
            self.metrics_status_processor = MetricsStatusLogger(config["statusReportFiles"]["metricsFileLocation"])
        else:
            self.metrics_status_processor = NullProcessor()

    def _get_metrics(self):
        ts_api_instance = cm_client.TimeSeriesResourceApi(self.api_client)
        from_time = datetime.datetime.fromtimestamp(time.time() - 180000)
        to_time = datetime.datetime.fromtimestamp(time.time())
        services=self.sharedDataHolder.get_services()
        if services:
            metrics_data_holder=[]
            for service in services:
                metrics_list = self._get_common_service_metrics(service.name)
                if metrics_list:
                    tsquery = self._create_metrics_tsquery(service.name, metrics_list)
                    result = ts_api_instance.query_time_series(_from=from_time, query=tsquery, to=to_time)
                    if result:
                        ts_list = result.items[0]
                        for ts in ts_list.time_series:
                            if "jsonMetricsLogger" in self.sharedDataHolder.get_config():
                                metric_event=NodeStatus.CmServiceMetricEvent(clusterDetails=self.sharedDataHolder.get_cluster_details())
                            else:
                                metric_event=NodeStatus.CmServiceMetricEvent()
                            metric_event.serviceName=service.name
                            metric_event.serviceType=service.type
                            metric_event.name=ts.metadata.metric_name
                            for point in ts.data:
                                d = datetime.datetime.strptime(point.timestamp, "%Y-%m-%dT%H:%M:%S.%fZ")
                                unixtime = time.mktime(d.timetuple())
                                metric_event.timestamp=int(unixtime)
                                metric_event.value=point.value
                            metrics_data_holder.append(metric_event)
                            self.metrics_processor.process(MessageToDict(metric_event))
            if "jsonMetricsLogger" in self.sharedDataHolder.get_config():
                metrics_data_holder = self.clear_cluster_details(metrics_data_holder)
            if metrics_data_holder:
                binaryVersion=version.get_binary_version()
                cmMetricsReport=NodeStatus.CmMetricsReport(metrics=metrics_data_holder, clusterDetails=self.sharedDataHolder.get_cluster_details(), cdpTelemetryVersion=binaryVersion)
                self.metrics_status_processor.process(MessageToDict(cmMetricsReport))

    def _create_metrics_tsquery(self, service_name, metrics):
        joined_metrics=', '.join(metrics)
        return "select %s where serviceName=%s and category = \"SERVICE\"" % (joined_metrics, service_name)

    def _get_common_service_metrics(self, service_name):
        common_service_metrics = []
        serviceMetricsMapFromConfig=self.sharedDataHolder.get_service_metrics_map_from_config()
        if service_name in serviceMetricsMapFromConfig:
            service_metrics_from_config = serviceMetricsMapFromConfig[service_name]
            availableMetrics=self.sharedDataHolder.get_available_metrics()
            if service_name in availableMetrics:
                available_metrics_for_service = availableMetrics[service_name]
                if available_metrics_for_service:
                    for service_metric in service_metrics_from_config:
                         if service_metric in available_metrics_for_service:
                            common_service_metrics.append(service_metric)
        return common_service_metrics

    def _setup_metrics_processor(self, config):
        if "jsonMetricsLogger" in config:
            name = config["jsonMetricsLogger"]['name']
            path = config["jsonMetricsLogger"]['path']
            max_bytes = config["jsonMetricsLogger"]['max_bytes']
            backup_count = config["jsonMetricsLogger"]['backup_count']
            self.metrics_processor = MetricsLogger(name, path, max_bytes, backup_count)
        else:
            self.metrics_processor = NullProcessor()

    def _setup_metrics_status_processor(self, config):
        if "statusReportFiles" in config and "metricsFileLocation" in config["statusReportFiles"]:
            self.metrics_status_processor = MetricsStatusLogger(config["statusReportFiles"]["metricsFileLocation"])
        else:
            self.metrics_status_processor = NullProcessor()