import psutil
import logging
import os
import datetime

from utils import version
from nodestatus.generated import nodestatus_pb2 as NodeStatus

class HostMetricsCollector(object):

    def __init__(self, verbose=False, filters=None):
        self.verbose=verbose
        self.filters=filters # todo

    def cpu_metrics(self):
        return self.cpu

    def collect_cpu_metrics(self):
        cpu_metrics=None
        cpu_times=NodeStatus.CpuTimes()
        self.__get_and_set_attributes_for_proto(psutil.cpu_times_percent(interval=1), cpu_times, ["user", "nice", "system", "idle"])
        if self.verbose:
            cpu_stats=NodeStatus.CpuStats()
            self.__get_and_set_attributes_for_proto(psutil.cpu_stats(), cpu_stats, ["ctx_switches", "interrupts", "soft_interrupts", "syscalls"])
            cpu_freq=NodeStatus.CpuFrequency()
            self.__get_and_set_long_attributes_for_proto(psutil.cpu_freq(), cpu_freq, ["current", "min", "max"])
            cpu_metrics=NodeStatus.CpuMetrics(times=cpu_times, stats=cpu_stats, freq=cpu_freq)
        else:
            cpu_metrics=NodeStatus.CpuMetrics(times=cpu_times)
        cpu_count=psutil.cpu_count(logical=False)
        cpu_logical_count=psutil.cpu_count()
        cpu_load_avg_resp=psutil.getloadavg()
        cpu_metrics.count=cpu_count
        cpu_metrics.logical_count=cpu_logical_count
        cpu_metrics.load_avg.extend(list(cpu_load_avg_resp))
        self.cpu=cpu_metrics

    def network_metrics(self):
        return self.network

    def collect_network_metrics(self):
        network_metrics=None
        networkIO=NodeStatus.NetworkIO()
        if self.verbose:
            self.__get_and_set_attributes_for_proto(psutil.net_io_counters(pernic=False), networkIO, ["bytes_sent", "bytes_recv", "packets_sent", "packets_recv", "errin", "errout", "dropin", "dropout"])
        networkConnections=NodeStatus.NetworkConnections()
        try:
            status_maps={}
            for conn in psutil.net_connections():
                if str(conn.status) == "NONE":
                    continue
                if conn.status not in status_maps:
                    status_maps[conn.status]=1
                else:
                    status_maps[conn.status]=status_maps[conn.status]+1
            for connField, connValue in status_maps.items():
                if hasattr(networkConnections, connField):
                    setattr(networkConnections, connField, connValue)
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass
        if self.verbose:
            network_metrics=NodeStatus.NetworkMetrics(connections=networkConnections, io=networkIO)
        else:
            network_metrics=NodeStatus.NetworkMetrics(connections=networkConnections)
        self.network=network_metrics

    def disk_metrics(self):
        return self.disk

    def collect_disk_metrics(self):
        disk_partitions_resp=psutil.disk_partitions()
        disk_partitions=[]
        if disk_partitions_resp:
            for disk_part in disk_partitions_resp:
                disk_usage=NodeStatus.DiskUsage()
                self.__get_and_set_attributes_for_proto(psutil.disk_usage(disk_part.mountpoint), disk_usage, ["total", "used", "free", "percent"])
                disk_partition=NodeStatus.DiskPartition(usage=disk_usage)
                self.__get_and_set_attributes_for_proto(disk_part, disk_partition, ["device", "mountpoint", "fstype", "opts", "maxfile", "maxpath"])
                disk_partitions.append(disk_partition)
        disk_metrics=None
        topVarLogs=self.__get_varlog_top_dirs()
        if self.verbose:
            diskIO=NodeStatus.DiskIO()
            self.__get_and_set_attributes_for_proto(psutil.disk_io_counters(perdisk=False), diskIO, ["read_count", "write_count", "read_bytes", "write_bytes", "read_time", "write_time"])
            disk_metrics=NodeStatus.DiskMetrics(partitions=disk_partitions, top_varlog=topVarLogs, io=diskIO)
        else:
            disk_metrics=NodeStatus.DiskMetrics(partitions=disk_partitions, top_varlog=topVarLogs)
        self.disk=disk_metrics

    def __get_varlog_top_dirs(self, top=10):
        top_varlog=[]
        varlog_path=os.path.join("/var", "log")
        if os.path.exists(varlog_path):
            alldirs=self.__listdirs(varlog_path)
            tuple_list=list()
            for currentdir in alldirs:
                abs_curr_dir=os.path.join(varlog_path, currentdir)
                tuple_list.append((abs_curr_dir, self.__dir_size(abs_curr_dir)))
            tuple_list.sort(key=lambda x: x[1], reverse=True)
            for tup in tuple_list[:top]:
                logObj=NodeStatus.VmLog(path=tup[0], size=self.sizeof_fmt(tup[1]))
                top_varlog.append(logObj)
        return top_varlog
    
    def __listdirs(self, folder):
        return [d for d in os.listdir(folder) if os.path.isdir(os.path.join(folder, d))]

    def __dir_size(self, directory):
        total_size = 0 
        for path, dirs, files in os.walk(directory):
            for f in files:
                fp = os.path.join(path, f)
                total_size += os.path.getsize(fp)
        return total_size
    
    def sizeof_fmt(self, num, suffix='B'):
        for unit in ['','K','M','G','T','P','E','Z']:
            if abs(num) < 1024.0:
                return "%3.1f %s%s" % (num, unit, suffix)
            num /= 1024.0
        return "%.1f %s%s" % (num, 'Y', suffix)

    def memory_metrics(self):
        return self.memory

    def collect_memory_metrics(self):
        memory_metrics=NodeStatus.MemoryMetrics()
        virtual_memory=NodeStatus.VirtualMemory()
        self.__get_and_set_attributes_for_proto(psutil.virtual_memory(), virtual_memory, ["total", "used", "free", "available", "percent", "active", "inactive", "wired", "buffers", "cached", "shared", "slab"])
        swap_memory=NodeStatus.SwapMemory()
        self.__get_and_set_attributes_for_proto(psutil.swap_memory(), swap_memory, ["total", "used", "free", "percent", "sin", "sout"])
        self.memory=NodeStatus.MemoryMetrics(virtual_memory=virtual_memory, swap_memory=swap_memory)

    def collect_all(self):
        self.collect_cpu_metrics()
        self.collect_disk_metrics()
        self.collect_memory_metrics()
        self.collect_network_metrics()
        nowTs=int(datetime.datetime.now().strftime('%s'))
        binaryVersion=version.get_binary_version()
        system_metrics=NodeStatus.SystemMetrics(cpu=self.cpu_metrics(), disk=self.disk_metrics(), memory=self.memory_metrics(), network=self.network_metrics(), timestamp=nowTs, cdpTelemetryVersion=binaryVersion)
        return system_metrics

    def __get_attributes_as_map(self, object, field_list=[]):
        map_resp={}
        for field in field_list:
            if hasattr(object, field):
                map_resp[field]=getattr(object, field)
        return map_resp

    def __get_and_set_attributes_for_proto(self, object, proto_obj, field_list=[]):
        for field in field_list:
            if hasattr(object, field):
                fieldVal=getattr(object, field)
                if hasattr(proto_obj, field):
                    setattr(proto_obj, field, fieldVal)
        return proto_obj

    def __get_and_set_long_attributes_for_proto(self, object, proto_obj, field_list=[]):
        for field in field_list:
            if hasattr(object, field):
                fieldVal=getattr(object, field)
                if hasattr(proto_obj, field):
                    setattr(proto_obj, field, int(fieldVal))
        return proto_obj