import logging
import json
import signal
from threading import Timer
from .metrics_logger import MetricsLogger, MetricsStatusLogger, NullProcessor
from .host_metrics_collector import HostMetricsCollector
from google.protobuf.json_format import MessageToDict

logger=logging.getLogger(__name__)

class SystemMetricsCollectorAgent:

    def __init__(self, config, standalone=True):
        self.config=config
        self.get_system_metrics_timer = None
        self.host_metrics_collector = HostMetricsCollector()
        if standalone:
            self._setup_termination_handler()
        self._setup_system_metrics_processor(config)
        self._setup_system_metrics_status_processor(config)
        self.globalFields = config["globalFields"]

    def start(self):
        logger.info("Starting CDP System Metrics Collector daemon.")
        self._get_system_metrics_periodically(self.config['config']['systemMetricsCollectPeriod'])

    def get_timer(self):
        return self.get_system_metrics_timer
    
    def _get_system_metrics_periodically(self, period):
        from .retry import retry
        retry(self._get_system_metrics, logger=logger, context="Gather system details", delay=30)
        self.get_system_metrics_timer = Timer(period, self._get_system_metrics_periodically, args=[period])
        self.get_system_metrics_timer.start()

    def _get_system_metrics(self):
        metrics_result=MessageToDict(self.host_metrics_collector.collect_all())
        self.system_metrics_status_processor.process(metrics_result)
        self.system_metrics_processor.process(self._append_global_fields(metrics_result))

    def _setup_system_metrics_processor(self, config):
        if "jsonMetricsLogger" in config["systemMetricsCollector"]:
            name = config["systemMetricsCollector"]["jsonMetricsLogger"]['name']
            path = config["systemMetricsCollector"]["jsonMetricsLogger"]['path']
            max_bytes = config["systemMetricsCollector"]["jsonMetricsLogger"]['max_bytes']
            backup_count = config["systemMetricsCollector"]["jsonMetricsLogger"]['backup_count']
            self.system_metrics_processor = MetricsLogger(name, path, max_bytes, backup_count)
        else:
            self.system_metrics_processor = NullProcessor()

    def _setup_system_metrics_status_processor(self, config):
        if "statusReportFiles" in config["systemMetricsCollector"] and "metricsFileLocation" in config["systemMetricsCollector"]["statusReportFiles"]:
            self.system_metrics_status_processor = MetricsStatusLogger(config["systemMetricsCollector"]["statusReportFiles"]["metricsFileLocation"])
        else:
            self.system_metrics_status_processor = NullProcessor()

    def _setup_termination_handler(self):
        def termination_handler(signum, frame):
            logger.debug("System Metrics collector termination called.")
            if self.get_system_metrics_timer:
                self.get_system_metrics_timer.cancel()
        signal.signal(signal.SIGINT, termination_handler)
        signal.signal(signal.SIGTERM, termination_handler)

    def _append_global_fields(self, obj):
        if self.globalFields:
            for key in self.globalFields:
                obj[key]=self.globalFields[key]
        return obj