import os
import datetime
import logging
import yaml
import time
import hashlib
from pid import PidFile
from threading import Thread, Event
from utils import command_helper
from flask import Flask
from flask_httpauth import HTTPBasicAuth
from nodestatus.salt_report import generate_salt_service_ping_healthcheck

logger=logging.getLogger('werkzeug')

def start(config):
    try:
        run(config)
    except KeyboardInterrupt:
        logger.debug("Node status monitor stopped.")

def run(config_path):
    config=NodeStatusConfig(config_path)
    if config.scheduledSaltUpdates:
        commandsThread = Thread(target=CommandsWorker(), args=(config.firstSleep, config.sleepTime, config.commands))
        commandsThread.daemon=True
        commandsThread.start()
    
    saltPingConfig=config.get_salt_ping_config()
    if saltPingConfig and saltPingConfig.is_enabled():
        logger.info("Periodic PING for local salt master and remote salt minions are enabled.")
        saltPingCmdThread = Thread(target=CommandsWorker(), args=(saltPingConfig.firstSleep, saltPingConfig.sleepTime, saltPingConfig.commands))
        saltPingCmdThread.daemon=True
        saltPingCmdThread.start()

    abs_folder=os.path.abspath(config.folder)
    abs_data_folder=os.path.abspath(config.dataFolder)
    app = Flask(__name__, static_url_path='', static_folder=abs_folder)
    auth = HTTPBasicAuth()
    users={}
    if config.useBasicAuth:
        users[config.username]=config.password
    saltPingCache = SaltPingCache(0, config.saltPingCacheEvictSeconds)

    @auth.verify_password
    def verify(username, password):
        if not config.useBasicAuth:
            return True
        else:
            result=False
            if username in users:
                if config.hashType == "sha256":
                    if encrypt_string(password) == users[username]:
                        result=True
                else:
                    if password == users[username]:
                        result=True
            return result
    
    @app.route("/api/v1/metering")
    @auth.login_required
    def metering():
        return app.send_static_file('metering_report.json')
        
    @app.route("/api/v1/services")
    @auth.login_required
    def services():
        return app.send_static_file('services_report.json')
        
    @app.route("/api/v1/network")
    @auth.login_required
    def netowrk():
        return app.send_static_file('network_report.json')

    @app.route("/api/v1/salt")
    @auth.login_required
    def salt():
        return app.send_static_file('salt_report.json')

    @app.route("/api/v1/salt/ping")
    @auth.login_required
    def salt_ping():
        if saltPingConfig and saltPingConfig.is_enabled():
            nowTs=int(datetime.datetime.now().strftime('%s'))
            if saltPingCache.should_be_updated(nowTs):
                generate_salt_service_ping_healthcheck(abs_data_folder, False, False)
        return app.send_static_file('salt_report.json')
    
    @app.route("/api/v1/salt/master")
    @auth.login_required
    def salt_local_master():
        return app.send_static_file('salt_master_local_report.json')

    @app.route("/api/v1/system/metrics")
    @auth.login_required
    def system_metrics():
        return app.send_static_file('system_metrics_report.json')
    
    @app.route("/api/v1/salt/minions")
    @auth.login_required
    def salt_minions():
        return app.send_static_file('salt_minions_report.json')

    @app.route("/api/v1/salt/keys")
    @auth.login_required
    def salt_keys():
        return app.send_static_file('salt_keys_report.json')
    
    @app.route("/api/v1/cm_services/metrics")
    @auth.login_required
    def cm_metrics():
        return app.send_static_file('cm_metrics_report.json')

    @app.route("/api/v1/cm_services/health")
    @auth.login_required
    def cm_health_checks():
        return app.send_static_file('cm_health_report.json')

    from waitress import serve
    serve(app, host=config.host, port=config.port)

def update(config_path):
    config=NodeStatusConfig(config_path)
    for cmd in config.commands:
        logger.debug("Running command: %s" % cmd)
        out,err,code=command_helper.run_command(cmd)
        if code != 0:
            logger.debug("Exit code of the command was: %s" % str(code))
            logger.error(str(err))

class NodeStatusConfig(object):

    def __init__(self, config_path):
        with open(config_path) as file:
            config = yaml.load(file, yaml.SafeLoader)
            self.folder=config["server"]["folder"] if "folder" in config["server"] else "/var/lib/cdp-nodestatus/report"
            self.dataFolder=config["server"]["dataFolder"] if "dataFolder" in config["server"] else "/var/lib/cdp-nodestatus/"
            self.port=int(config["server"]["port"]) if "port" in config["server"] else 61888
            self.host=config["server"]["host"] if "host" in config["server"] else "0.0.0.0"
            self.commands=config["commands"] if "commands" in config else ["cdp-telemetry nodestatus collect"]
            self.firstSleep=int(config["firstSleep"]) if "firstSleep" in config else 300
            self.sleepTime=int(config["sleepTime"]) if "sleepTime" in config else 3600
            self.scheduledSaltUpdates=bool(config["scheduledSaltUpdates"]) if "scheduledSaltUpdates" in config else True
            if "auth" in config["server"] and "username" in config["server"]["auth"] and "password" in config["server"]["auth"]:
                self.useBasicAuth=True
                self.username=config["server"]["auth"]["username"]
                self.password=config["server"]["auth"]["password"]
                self.hashType=config["server"]["auth"]["hash"] if "hash" in config["server"]["auth"] else "simple"
            else:
                self.useBasicAuth=False
            self.saltPingCacheEvictSeconds=int(config["saltPingCacheEvictSeconds"]) if "saltPingCacheEvictSeconds" in config else 30
            self.saltPingConfig=None
            if "saltPing" in config:
                self.saltPingConfig=SaltPingConfig(config["saltPing"])
    
    def get_salt_ping_config(self):
        return self.saltPingConfig

    def folder(self):
        return self.folder
    
    def dataFolder(self):
        return self.dataFolder

    def port(self):
        return self.port
    
    def host(self):
        return self.host

    def commands(self):
        return self.commands
    
    def firstSleep(self):
        return self.firstSleep
    
    def sleepTime(self):
        return self.sleepTime
    
    def useBasicAuth(self):
        return self.useBasicAuth

    def username(self):
        return self.usernaame

    def password(self):
        return self.password
    
    def hashType(self):
        return self.hashType

class CommandsWorker(object):

    def __call__(self, firstSleep=5, sleepTime=5, commands=[]):
        e = Event()
        e.wait(firstSleep)
        while True:
            for cmd in commands:
                logger.debug("Running command: %s" % cmd)
                out,err,code=command_helper.run_command(cmd)
                if code != 0:
                    logger.debug("Exit code of the command was: %s" % str(code))
                    logger.error(str(err))
            e.wait(sleepTime)

def encrypt_string(hash_string):
    sha_signature = \
        hashlib.sha256(hash_string.encode()).hexdigest()
    return sha_signature

class SaltPingConfig(object):

    def __init__(self, config):
        self.enabled = bool(config["enabled"]) if "enabled" in config else False
        self.firstSleep=int(config["firstSleep"]) if "firstSleep" in config else 300
        self.sleepTime=int(config["sleepTime"]) if "sleepTime" in config else 600
        self.commands=list()
        self.commands.append("cdp-nodestatus salt-ping all")

    def is_enabled(self):
        return self.enabled
    
    def firstSleep(self):
        return self.firstSleep
    
    def sleepTime(self):
        return self.sleepTime
    
    def commands(self):
        return self.commands

class SaltPingCache(object):

    def __init__(self, timestamp, evictTimeinSec=30):
        self.timestamp=timestamp
        self.evictTimeinSec=evictTimeinSec

    def should_be_updated(self, input_timestamp):
        result=False
        if input_timestamp > self.timestamp:
            time_difference = input_timestamp - self.timestamp
            if time_difference <= self.evictTimeinSec:
                logger.info("Time difference (seconds) between ping executions is: %s, it needs to be larger than %s for a fresh response" % (str(time_difference), str(self.evictTimeinSec)))
            else:
                logger.info("Time difference (seconds) between ping execution is: %s. Refreshing ping cache..." % str(time_difference))
                self.timestamp=input_timestamp
                result=True
        return result