
import datetime
import os
import logging
import sys
import json
import traceback
from utils import telemetry_logger, version
from doctor.salt import run_salt_command, run_salt_key_command
from doctor.service import DoctorService
from nodestatus.generated import nodestatus_pb2 as NodeStatus
from google.protobuf.json_format import ParseDict, MessageToDict

logger=logging.getLogger(__name__)

def generate_salt_service_ping_healthcheck(heathcheck_dir, exitOnError=True, pretty=False):
    saltMasterHealth=generate_local_salt_master_healthcheck(heathcheck_dir, exitOnError, pretty)
    saltMinionsHealth=generate_salt_minions_healthcheck(heathcheck_dir, exitOnError, pretty)
    saltKeys=generate_salt_keys_check(heathcheck_dir, exitOnError, pretty)
    concat_master_and_minions_and_keys(heathcheck_dir, saltMasterHealth, saltMinionsHealth, saltKeys, exitOnError, pretty)

def concat_master_and_minions_and_keys(heathcheck_dir, saltMasterHealth, saltMinionsHealth, saltKeys, exitOnError=True, pretty=False):
    output_file=os.path.join(heathcheck_dir, "report", "salt_report.json")
    nodeStatusHealthReport=None
    try:
        timestamp=int(datetime.datetime.now().strftime('%s'))
        binaryVersion=version.get_binary_version()
        nodeStatusHealthReport=NodeStatus.SaltHealthReport(master=saltMasterHealth, minions=saltMinionsHealth, timestamp=timestamp, keys=saltKeys, cdpTelemetryVersion=binaryVersion)
        with open(output_file, 'w') as report_output:
            if pretty:
                json.dump(MessageToDict(nodeStatusHealthReport), report_output, indent=2, sort_keys=True)
            else:
                json.dump(MessageToDict(nodeStatusHealthReport), report_output, sort_keys=True)
    except Exception as e:
            logger.error("Error occurred during write json file '%s': %s" % (output_file, str(traceback.format_exc())))
            if exitOnError:
                sys.exit(1)
    return nodeStatusHealthReport

def generate_salt_minions_healthcheck(heathcheck_dir, exitOnError=True, pretty=False):
    output_file=os.path.join(heathcheck_dir, "salt", "salt_minions_healthcheck_result.json")
    salt_minion_report_file=os.path.join(heathcheck_dir, "report", "salt_minions_report.json")
    run_and_log_salt_command("test.ping --out=json --out-file=%s --out-indent=-1" % output_file, True)
    saltMinionsHealth=None
    if os.path.exists(output_file):
        try:
            with open(output_file, 'r') as json_file:
                json_list = list(json_file)
            data={}
            for json_str in json_list:
                if json_str:
                    json_data=json.loads(json_str)
                    for key, json_dict_elem in json_data.items():
                        data[key]=json_dict_elem
            saltMinionsHealth=NodeStatus.SaltMinionsHealth()
            minionPingStatuses=[]
            for instanceKey in data:
                pingStatusResponse=NodeStatus.StatusDetails()
                pingStatusResponse.host=instanceKey
                instanceResponse=data[instanceKey]
                if str(instanceResponse) == "True":
                    pingStatusResponse.status=NodeStatus.HealthStatus.OK
                else:
                    pingStatusResponse.status=NodeStatus.HealthStatus.NOK
                    pingStatusResponse.statusReason=instanceResponse
                minionPingStatuses.append(pingStatusResponse)
            with open(salt_minion_report_file, "w") as min_out_file:
                saltMinionsHealth.pingResponses.extend(minionPingStatuses)
                saltMinionsHealth.timestamp=int(datetime.datetime.now().strftime('%s'))
                if pretty:
                    json.dump(MessageToDict(saltMinionsHealth), min_out_file, indent=2, sort_keys=True)
                else:
                    json.dump(MessageToDict(saltMinionsHealth), min_out_file, sort_keys=True)
        except Exception as e:
            logger.error("Error occurred during load json file '%s': %s" % (output_file, str(traceback.format_exc())))
            if exitOnError:
                sys.exit(1)
    return saltMinionsHealth

def generate_salt_keys_check(heathcheck_dir, exitOnError=True, pretty=False):
    output_file=os.path.join(heathcheck_dir, "salt", "salt_keys_check_result.json")
    salt_keys_report_file=os.path.join(heathcheck_dir, "report", "salt_keys_report.json")
    run_and_log_salt_key_command(output_file)
    # response: {"minions": [], "minions_pre": [], "minions_rejected": [], "minions_denied": []}
    saltKeys=None
    if os.path.exists(output_file):
        try:
            with open(output_file, "r") as salt_keys_json_file:
                salt_keys_json_data = json.load(salt_keys_json_file)
                accepted=list()
                unaccepted=list()
                denied=list()
                rejected=list()
                if "minions" in salt_keys_json_data:
                    accepted=salt_keys_json_data["minions"]
                if "minions_pre" in salt_keys_json_data:
                    unaccepted=salt_keys_json_data["minions_pre"]
                if "minions_rejected" in salt_keys_json_data:
                    rejected=salt_keys_json_data["minions_rejected"]
                if "minions_denied" in salt_keys_json_data:
                    denied=salt_keys_json_data["minions_denied"]
                timestamp=int(datetime.datetime.now().strftime('%s'))
                saltKeys=NodeStatus.SaltKeys(unaccepted=unaccepted, accepted=accepted, rejected=rejected, denied=denied, timestamp=timestamp)
                try:
                    with open(salt_keys_report_file, "w") as salt_keys_report_out:
                        if pretty:
                            json.dump(MessageToDict(saltKeys), salt_keys_report_out, indent=2, sort_keys=True)
                        else:
                            json.dump(MessageToDict(saltKeys), salt_keys_report_out, sort_keys=True)
                except Exception as e:
                    logger.error("Error occurred during load json file '%s': %s" % (salt_keys_report_file, str(traceback.format_exc())))
                    if exitOnError:
                        sys.exit(1)
        except Exception as e:
            logger.error("Error occurred during load json file '%s': %s" % (output_file, str(traceback.format_exc())))
            if exitOnError:
                sys.exit(1)
    return saltKeys

def generate_local_salt_master_healthcheck(heathcheck_dir, exitOnError=True, pretty=False):
    output_file=os.path.join(heathcheck_dir, "salt", "salt_local_master_healthcheck_result.json")
    salt_master_local_report_file=os.path.join(heathcheck_dir, "report", "salt_master_local_report.json")
    doctorService=DoctorService("json")
    servicesDoctorRespponse=doctorService.collect("salt-master,salt-api,salt-bootstrap")
    saltMasterHealth=None
    try:
        with open(output_file, "w") as local_salt_master_json_file_for_write:
            json.dump(json.loads(servicesDoctorRespponse), local_salt_master_json_file_for_write, indent=2, sort_keys=True)
    except Exception as e:
        logger.error("Error occurred during write json file '%s': %s" % (output_file, str(traceback.format_exc())))
        if exitOnError:
            sys.exit(1)
    if os.path.exists(output_file):
        try:
            saltMasterHealth=NodeStatus.SaltMasterHealth()
            with open(output_file, "r") as local_salt_json_file:
                local_salt_json_data = json.load(local_salt_json_file)
                salt_service_states=[]
                if "services" in local_salt_json_data:
                    service_list=local_salt_json_data["services"]
                    for service_obj in service_list:
                        if service_obj["name"] == "salt-master" or service_obj["name"] == "salt-api" or service_obj["name"] == "salt-bootstrap":
                            serviceStatusValue=NodeStatus.HealthStatus.Value(service_obj["status"])
                            serviceStatus=NodeStatus.ServiceStatus(name=service_obj["name"], status=serviceStatusValue)
                            salt_service_states.append(serviceStatus)
                with open(salt_master_local_report_file, "w") as salt_local_report_out:
                    saltMasterHealth.services.extend(salt_service_states)
                    saltMasterHealth.timestamp=int(datetime.datetime.now().strftime('%s'))
                    if pretty:
                        json.dump(MessageToDict(saltMasterHealth), salt_local_report_out, indent=2, sort_keys=True)
                    else:
                        json.dump(MessageToDict(saltMasterHealth), salt_local_report_out, sort_keys=True)
        except Exception as e:
            logger.error("Error occurred during load json file '%s': %s" % (output_file, str(traceback.format_exc())))
            if exitOnError:
                sys.exit(1)
    return saltMasterHealth

def generate_salt_healthcheck(heathcheck_dir, healthCheckFilters=None, databusEndpoint=None, databusS3Endpoint=None, cloudProvider=None, pretty=False):
    salt_folder=os.path.join(heathcheck_dir, "salt")
    status_folder=os.path.join(heathcheck_dir, "report")
    if not os.path.exists(salt_folder):
        os.mkdir(salt_folder)
    if not os.path.exists(status_folder):
        os.mkdir(status_folder)
    if not healthCheckFilters or "metering" in healthCheckFilters:
        generate_metering_salt_healthcheck(heathcheck_dir, pretty)
    if not healthCheckFilters or "services" in healthCheckFilters:
        generate_services_salt_healthcheck(heathcheck_dir, pretty)
    if not healthCheckFilters or "network" in healthCheckFilters:
        generate_network_salt_healthcheck(heathcheck_dir, databusEndpoint, databusS3Endpoint, cloudProvider, pretty)
    if not healthCheckFilters or "system-metrics" in healthCheckFilters:
        generate_system_metrics_healthcheck(heathcheck_dir, pretty)

def generate_services_salt_healthcheck(services_healthcheck_dir, pretty):
    salt_output_file=os.path.join(services_healthcheck_dir, "salt", "salt_services_healthcheck_result.json")
    report_output_file=os.path.join(services_healthcheck_dir, "report", "services_report.json")
    salt_command="cmd.run 'cdp-doctor service status --format json' timeout=120 --out=json --out-file=%s --out-indent=-1" % salt_output_file
    saltCheckConfig=SaltStatusCheckConfig("service", salt_command, salt_output_file, report_output_file)
    __generate_status_salt_healthcheck(saltCheckConfig, pretty)

def generate_network_salt_healthcheck(network_healthcheck_dir, databus_url, databus_s3_url, cloud_provider, pretty):
    salt_output_file=os.path.join(network_healthcheck_dir, "salt", "salt_network_healthcheck_result.json")
    report_output_file=os.path.join(network_healthcheck_dir, "report", "network_report.json")
    cloud_provider_params="-c %s" % cloud_provider if cloud_provider else ""
    databus_s3_url_params="--databus-s3-url %s " % databus_s3_url if databus_s3_url else " "
    salt_command="cmd.run 'cdp-doctor network status --format json -s --databus-url %s %s%s' timeout=120 --out=json --out-file=%s --out-indent=-1" % (databus_url, databus_s3_url_params, cloud_provider_params, salt_output_file)
    saltCheckConfig=SaltStatusCheckConfig("network", salt_command, salt_output_file, report_output_file)
    __generate_status_salt_healthcheck(saltCheckConfig, pretty)

def generate_metering_salt_healthcheck(metering_healthcheck_dir, pretty):
    salt_output_file=os.path.join(metering_healthcheck_dir, "salt", "salt_metering_healthcheck_result.json")
    report_output_file=os.path.join(metering_healthcheck_dir, "report", "metering_report.json")
    salt_command="cmd.run 'cdp-doctor metering status --format json' timeout=120 --out=json --out-file=%s --out-indent=-1" % salt_output_file
    saltCheckConfig=SaltStatusCheckConfig("metering", salt_command, salt_output_file, report_output_file)
    __generate_status_salt_healthcheck(saltCheckConfig, pretty)

def generate_system_metrics_healthcheck(system_metrics_dir, pretty):
    salt_output_file=os.path.join(system_metrics_dir, "salt", "salt_system_metrics_result.json")
    report_output_file=os.path.join(system_metrics_dir, "report", "system_metrics_report.json")
    salt_command="cmd.run 'cdp-doctor system metrics --format json' timeout=120 --out=json --out-file=%s --out-indent=-1" % salt_output_file
    saltCheckConfig=SaltStatusCheckConfig("system-metrics", salt_command, salt_output_file, report_output_file)
    __generate_status_salt_healthcheck(saltCheckConfig, pretty)

def __generate_status_salt_healthcheck(salt_status_check_config, pretty=False):
    salt_output_file=salt_status_check_config.get_salt_output()
    status_report_output_file=salt_status_check_config.get_status_report_output()
    run_and_log_salt_command(salt_status_check_config.get_salt_command(), True)
    status_check_type=salt_status_check_config.get_check_type()
    if os.path.exists(salt_output_file):
        try:
            with open(salt_output_file, 'r') as json_file:
                json_list = list(json_file)
            data={}
            for json_str in json_list:
                if json_str:
                    json_data=json.loads(json_str)
                    for key, json_dict_elem in json_data.items():
                        data[key]=json_dict_elem
            nodeStatusList=[]
            for instanceKey in data:
                jsonStr=data[instanceKey]
                commonNodeStatusData=NodeStatus.StatusDetails()
                commonNodeStatusData.host=instanceKey
                nodeStatus=NodeStatus.NodeStatus()
                servicesDetails=None
                meteringDetails=None
                networkDetails=None
                systemMetrics=None
                try:
                    check_details=json.loads(jsonStr)
                    if status_check_type == "service":
                        servicesDetails=ParseDict(check_details, NodeStatus.ServicesDetails(), ignore_unknown_fields=True)
                    if status_check_type == "network":
                        networkDetails=ParseDict(check_details, NodeStatus.NetworkDetails(), ignore_unknown_fields=True)
                    if status_check_type == "metering":
                        meteringDetails=ParseDict(check_details, NodeStatus.MeteringDetails(), ignore_unknown_fields=True)
                    if status_check_type == "system-metrics":
                        systemMetrics=ParseDict(check_details, NodeStatus.SystemMetrics(), ignore_unknown_fields=True)
                    commonNodeStatusData.status=NodeStatus.HealthStatus.OK
                except Exception as e:
                    logger.error("Error occurred during read json string: %s" % str(traceback.format_exc()))
                    commonNodeStatusData.status=NodeStatus.HealthStatus.NOK
                nodeStatus=NodeStatus.NodeStatus(statusDetails=commonNodeStatusData, servicesDetails=servicesDetails, networkDetails=networkDetails, meteringDetails=meteringDetails, systemMetrics=systemMetrics)
                nodeStatusList.append(nodeStatus)
            
            with open(status_report_output_file, "w") as stat_out_file:
                nowTs=int(datetime.datetime.now().strftime('%s'))
                binaryVersion=version.get_binary_version()
                nodeStatusReport=NodeStatus.NodeStatusReport(cdpTelemetryVersion=binaryVersion)
                nodeStatusReport.nodes.extend(nodeStatusList)
                nodeStatusReport.timestamp=nowTs
                if pretty:
                    json.dump(MessageToDict(nodeStatusReport), stat_out_file, indent=2, sort_keys=True)
                else:
                    json.dump(MessageToDict(nodeStatusReport), stat_out_file, sort_keys=True)
        except Exception as e:
            logger.error("Error occurred during load json file '%s': %s" % (salt_output_file, str(traceback.format_exc())))
            sys.exit(1)

def run_and_log_salt_command(command, doNotSplit=False):
    logger.debug("Running salt command: %s" % command)
    out, err, exit_code = run_salt_command(command, doNotSplit=doNotSplit)
    if (str(exit_code).rstrip() == "0"):
        logger.debug(telemetry_logger.success_message("[OK]"))
    else:
        logger.error(telemetry_logger.fail_message("[FAILED]"))
        logger.debug("StdOut: \n%s" % out)
        logger.debug("StdErr: \n%s" % err)
        logger.debug("Exit code: \n%s" % str(exit_code))

def run_and_log_salt_key_command(output_file):
    command="-L --out=json --out-file=%s --out-indent=-1" % output_file
    logger.debug("Running salt-key command: %s" % command)
    out, err, exit_code = run_salt_key_command(command)
    if (str(exit_code).rstrip() == "0"):
        logger.debug(telemetry_logger.success_message("[OK]"))
    else:
        logger.error(telemetry_logger.fail_message("[FAILED]"))
        logger.debug("StdOut: \n%s" % out)
        logger.debug("StdErr: \n%s" % err)
        logger.debug("Exit code: \n%s" % str(exit_code))

class SaltStatusCheckConfig(object):

    def __init__(self, checkType, saltCommand, saltOutput, statusReportOutput):
        self.checkType=checkType
        self.saltCommand=saltCommand
        self.saltOutput=saltOutput
        self.statusReportOutput=statusReportOutput

    def get_salt_output(self):
        return self.saltOutput

    def get_status_report_output(self):
        return self.statusReportOutput

    def get_salt_command(self):
        return self.saltCommand

    def get_check_type(self):
        return self.checkType