import boto3
import logging
import os
import glob
import errno
from boto3.s3.transfer import S3Transfer
from botocore.exceptions import ClientError
from boto3.exceptions import S3UploadFailedError

logger = logging.getLogger(__name__)

def get_lifecycle(bucket, region):
    s3 = boto3.client('s3', region_name=region)
    lifecycle_config = s3.get_bucket_lifecycle_configuration(Bucket=bucket)
    logger.info(lifecycle_config)

def upload_file(transfer, bucket, file, upload_file_path, enc):
    extra_args={}
    if enc:
        try:
            extra_args={"ServerSideEncryption": "AES256"}
            transfer.upload_file(file, bucket, upload_file_path, extra_args=extra_args)
        except S3UploadFailedError as e:
            logger.debug(e)
            logger.info("As AES256 default encryption failed, try to upload with KMS based encryption.")
            extra_args={"ServerSideEncryption": "aws:kms"}
            transfer.upload_file(file, bucket, upload_file_path, extra_args=extra_args)
    else:
        transfer.upload_file(file, bucket, upload_file_path, extra_args=extra_args)

def upload(bucket, file, paths, location, region, enc):
    s3 = boto3.client('s3', region_name=region)
    transfer = S3Transfer(s3)
    extra_args={}
    if paths:
        logger.debug("Use matcher to search for files: %s" % (paths))
        files = glob.glob(paths)
        for file_to_upload in files:
            logger.debug("File to upload: %s" % (file_to_upload))
            base_file=os.path.basename(file_to_upload)
            upload_file_path = location + base_file if location.endswith("/") else location + "/" + base_file
            logger.debug("Upload destination: %s" % (upload_file_path))
            upload_file(transfer, bucket, file_to_upload, upload_file_path, enc)
    else:
        logger.debug("File to upload: %s" % (file))
        upload_file_path=location + os.path.basename(file) if location.endswith("/") else location
        logger.debug("Upload destination: %s" % (upload_file_path))
        upload_file(transfer, bucket, file, upload_file_path, enc)

def check_encryption(bucket, region):
    s3 = boto3.client('s3', region_name=region)
    get_encryption(s3, bucket)

def get_encryption(s3, bucket):
    try:
        enc = s3.get_bucket_encryption(Bucket=bucket)
        rules = enc['ServerSideEncryptionConfiguration']['Rules']
        logger.info('Bucket: %s, Encryption: %s' % (bucket, rules))
    except ClientError as e:
        if e.response['Error']['Code'] == 'ServerSideEncryptionConfigurationNotFoundError':
            logger.info('Bucket: %s, no server-side encryption' % (bucket))
        else:
            logger.info("Bucket: %s, unexpected error: %s" % (bucket, e))

def pull(bucket, prefix, local_dir, region):
    s3 = boto3.client('s3', region_name=region)
    keys = []
    dirs = []
    next_token = ''
    base_kwargs = {
        'Bucket':bucket,
        'Prefix':prefix,
    }
    while next_token is not None:
        kwargs = base_kwargs.copy()
        if next_token != '':
            kwargs.update({'ContinuationToken': next_token})
        results = s3.list_objects_v2(**kwargs)
        contents = results.get('Contents')
        for i in contents:
            k = i.get('Key')
            if k[-1] != '/':
                keys.append(k)
            else:
                dirs.append(k)
        next_token = results.get('NextContinuationToken')
    for d in dirs:
        dest_pathname = os.path.join(local_dir, d)
        if not os.path.exists(os.path.dirname(dest_pathname)):
            logger.debug("Create directory as it does not exist: %s" % (dest_pathname))
            os.makedirs(os.path.dirname(dest_pathname))
    for k in keys:
        dest_pathname = os.path.join(local_dir, k)
        if not os.path.exists(os.path.dirname(dest_pathname)):
            os.makedirs(os.path.dirname(dest_pathname))
        logger.debug("Download file from S3 - bucket: %s, object key: %s, destination: %s" % (bucket, k, dest_pathname))
        s3.download_file(bucket, k, dest_pathname)