from asn1crypto import keys, pem
from pure25519 import eddsa
from urllib.parse import urlsplit
import base64
import hashlib
import email.utils
import json
import collections
from ecdsa import SigningKey
from ecdsa.util import sigencode_der

class V1Signer(object):
    ERROR_MESSAGE = \
       "Failed to import private key from: '%s'. The private key is " \
       "corrupted or not in the right format. The private key " \
       "was extracted either from 'env' (environment variables), " \
       "'shared-credentials-file' (a profile in the shared " \
       "credential file, by default under ~/.altus/credentials), or " \
       "'auth-config-file' (a file containing the credentials whose " \
       "location was supplied on the command line.)"

    def __init__(self, access_key_id, private_key, auth_method, logger, headers={}):
        self.access_key_id = access_key_id
        self.private_key = private_key
        self.auth_method = auth_method
        self.logger = logger
        self.headers=headers
        if not headers or 'Content-Type' not in headers:
            self.headers['Content-Type'] = 'application/json'

    def _raw_sign_string(self, string_to_sign):
        raise NotImplementedError("Implement _raw_sign_string")

    def _sign_string(self, string_to_sign):
        """
        Sign the supplied string using the credentials and return the base64
        encoded signature in UTF-8 format.
        :param string_to_sign: String to sign
        :return: Signature as string
        """
        signature = self._raw_sign_string(string_to_sign)
        return base64.urlsafe_b64encode(signature).strip().decode('utf-8')

    def _canonical_standard_headers(self, headers):
        interesting_headers = ['content-type', 'x-altus-date']
        hoi = []
        if 'x-altus-date' in headers:
            raise Exception("x-altus-date found in headers!")
        headers['x-altus-date'] = self._get_date()
        for ih in interesting_headers:
            found = False
            for key in headers:
                lk = key.lower()
                if headers[key] is not None and lk == ih:
                    hoi.append(headers[key].strip())
                    found = True
            if not found:
                hoi.append('')
        return '\n'.join(hoi)

    def _canonical_string(self, method, split, headers):
        cs = method.upper() + '\n'
        cs += self._canonical_standard_headers(headers) + '\n'
        cs += split.path + '\n'
        cs += self.auth_method
        return cs

    def _get_signature(self, method, split, headers):
        string_to_sign = self._canonical_string(method, split, headers)
        self.logger.debug('StringToSign:\n%s', string_to_sign)
        return self._sign_string(string_to_sign)

    def add_auth(self, request_method, request_url):
        self.logger.debug("Calculating signature using %s." % self.auth_method)
        self.logger.debug('HTTP request method: %s', request_method)
        split = urlsplit(request_url)
        signature = self._get_signature(request_method,
                                        split,
                                        self.headers)
        self._inject_signature(signature)
        return self.headers

    def _get_date(self):
        return email.utils.formatdate(usegmt=True)

    def _inject_signature(self, signature):
        if 'x-altus-auth' in self.headers:
            raise Exception("x-altus-auth found in headers!")
        self.headers['x-altus-auth'] = self._get_signature_header(signature)

    def _get_signature_header(self, signature):
        auth_params = collections.OrderedDict()
        auth_params['access_key_id'] = self.access_key_id
        auth_params['auth_method'] = self.auth_method
        encoded_auth_params = json.dumps(auth_params).encode('utf-8')
        return "%s.%s" % (
            base64.urlsafe_b64encode(encoded_auth_params).strip().decode('utf-8'),
            signature)

class Ed25519v1Auth(V1Signer):
    """
    Ed25519 signing with a SHA-512 hash returning a base64 encoded signature.
    """
    AUTH_METHOD_NAME = 'ed25519v1'
    ED25519_SEED_LENGTH = 32
    ED25519_BASE64_SEED_LENGTH = 44

    def __init__(self, access_key_id, private_key, logger, headers={}):
        super(Ed25519v1Auth, self).__init__(access_key_id, private_key, self.AUTH_METHOD_NAME, logger, headers)

    @classmethod
    def detect_private_key(cls, key):
        return len(key) == cls.ED25519_BASE64_SEED_LENGTH

    def _raw_sign_string(self, string_to_sign):
        """
        Sign the supplied string using the credentials and return the raw signature.
        :param string_to_sign: String to sign
        :return: Raw signature as string
        """
        try:
            # We expect the private key to be a base64 formatted string.
            seed = base64.b64decode(self.private_key)
            if len(seed) != self.ED25519_SEED_LENGTH:
                raise Exception('Not an Ed25519 private key: %s' %
                                self.private_key)

            pk = eddsa.publickey(seed)
            signature = eddsa.signature(string_to_sign.encode('utf-8'), seed, pk)
            return signature
        except Exception:
            message = self.ERROR_MESSAGE % self.credentials.method
            self.logger.debug(message, exc_info=True)
            raise Exception(message)

class ECDSAv1Auth(V1Signer):
    """
    ECDSA signing with a SHA-512 hash returning a base64 encoded signature.
    """
    AUTH_METHOD_NAME = 'ecdsav1'

    def __init__(self, access_key_id, private_key, logger, headers={}):
        super(ECDSAv1Auth, self).__init__(access_key_id, private_key, self.AUTH_METHOD_NAME, logger, headers)
   
    def _parse_private_key(self, private_key, error_message="Failed to import private key."):
        """
        Parse the ECDSA private key.
        :param private_key: The private key.
        :return: The SigningKey.
        :raise: Exception if the private key can not be parsed.
        """
        try:
            # We expect the private key to be a PKCS8 pem formatted string.
            pem_bytes = private_key.encode('utf-8')
            if pem.detect(pem_bytes):
                _, _, der_bytes = pem.unarmor(pem_bytes)
                # In PKCS8 the key is wrapped in a container that describes it
                info = keys.PrivateKeyInfo.load(der_bytes, strict=True)
                # Directly unwrap the private key. The asn1crypto library stopped
                # offering an API call for this in their 1.0.0 release but their
                # official answer of using a separate native-code-dependent
                # library to do one line of work is unreasonable. Of course, this
                # line might break in the future...
                unwrapped = info['private_key'].parsed
                # The unwrapped key is equivalent to pkcs1 contents
                return SigningKey.from_der(unwrapped.dump())
            else:
                raise Exception('Not a PEM file')
        except Exception:
            self.logger.debug(error_message, exc_info=True)
            raise Exception(error_message)

    @classmethod
    def detect_private_key(private_key):
        """
        Try to see if the private key is an ECDSA key by parsing it.
        :param private_key: The key to check.
        :return: True if the key is an ECDSA key, False otherwise.
        """
        try:
            ECDSAv1Auth._parse_private_key(private_key)
            return True
        except Exception:
            return False

    def _raw_sign_string(self, string_to_sign):
        """
        Sign the supplied string using the credentials and return the raw signature.
        :param string_to_sign: String to sign
        :return: Raw signature as string
        """
        formatted_private_key=self.private_key.replace("\\n", "\n")
        sk = self._parse_private_key(
            formatted_private_key,
            self.ERROR_MESSAGE % 'ecdsav1')
        # Sign the hash.
        signature = sk.sign_deterministic(
            string_to_sign.encode('utf-8'),
            hashfunc=hashlib.sha512,
            sigencode=sigencode_der)
        return signature