import logging
import os
import re
from .proxy_settings import ProxySettings

logger=logging.getLogger(__name__)

TUNNEL_FILE_CONFIG = "/cdp/bin/reverse-tunnel-values-GATEWAY.sh"
# TUNNEL_FILE_CONFIG = "examples/ccm/reverse-tunnel-values-GATEWAY.sh"
CCM_V2_CONFIG = "/cdp/bin/ccmv2/config.toml"
# CCM_V2_CONFIG = "examples/ccm/config.toml"
CCM_V2_JUMPGATE_CONFIG =  "/etc/jumpgate/config.toml"
# CCM_V2_JUMPGATE_CONFIG = "examples/ccm/jumpgate/config.toml"
CCM_V2_JUMPGATE_PROXY_CONFIG = "/etc/cdp/proxy.env"
# CCM_V2_JUMPGATE_PROXY_CONFIG = "examples/ccm/jumpgate/proxy.env"

class CcmConfigHolder(object):

    def __init__(self, ccmV2Config=CCM_V2_CONFIG, ccmV2JumpgateConfig=CCM_V2_JUMPGATE_CONFIG, ccmV2JumpgateProxyConfig=CCM_V2_JUMPGATE_PROXY_CONFIG):
        self.ccm_v2_config_map={}
        self.ccm_v1_config_map={}
        self.ccm_v2_config_location=ccmV2Config
        self.ccm_v2_jumpgate_config_location=ccmV2JumpgateConfig
        self.ccm_v2_jumpgate_proxy_config_location=ccmV2JumpgateProxyConfig
        self.is_jumpgate=os.path.exists(ccmV2JumpgateConfig)
        self.ccm_v2_config_map=self.__resolve_ccm_v2_configs()
        self.ccm_v1_config_map=self.__resolve_ccm_v1_configs()

    def get_ccm_v2_config_filename(self):
        return self.ccm_v2_config_location

    def get_ccm_v2_jumpgate_config_filename(self):
        return self.ccm_v2_jumpgate_config_location

    def get_ccm_v1_config_filename(self):
        return TUNNEL_FILE_CONFIG

    def is_ccm_enabled(self):
        return True if (self.ccm_v2_config_map or self.ccm_v1_config_map) else False

    def is_v2(self):
        return True if self.ccm_v2_config_map else False

    def is_jumpgate_config(self):
        return True if self.is_jumpgate else False
    
    def get_configs(self):
        return self.ccm_v2_config_map if self.ccm_v2_config_map else self.ccm_v1_config_map

    def get_host_and_port(self):
        if self.ccm_v2_config_map:
            return self.ccm_v2_config_map["ccm_host"], self.ccm_v2_config_map["ccm_port"]
        elif self.ccm_v1_config_map:
            return self.ccm_v1_config_map["CCM_HOST"], self.ccm_v1_config_map["CCM_SSH_PORT"]
        else:
            return "", ""

    def get_ccm_v2_url(self):
        if self.is_jumpgate:
            return self.ccm_v2_config_map["relayServer"] if "relayServer" in self.ccm_v2_config_map else None
        else:
            return self.ccm_v2_config_map["proxy"] if "proxy" in self.ccm_v2_config_map else None

    def get_ccm_v2_proxy_settings(self):
        proxySettings=None
        if self.is_jumpgate:
            if "https_proxy" in self.ccm_v2_config_map:
                proxySettings=ProxySettings(full_proxy_url=self.ccm_v2_config_map["https_proxy"])
        elif "httpProxyURL" in self.ccm_v2_config_map:
            proxySettings=ProxySettings(full_proxy_url=self.ccm_v2_config_map["httpProxyURL"])
        return proxySettings

    def get_ccm_v2_certs(self):
        cacert, cert, key = None, None, None
        trusted_server_cert_name=self.__get_trusted_server_cert_config_name()
        if trusted_server_cert_name in self.ccm_v2_config_map:
            cacert = self.ccm_v2_config_map[trusted_server_cert_name]
            if cacert.startswith("-----"):
                logger.debug("Cacert value is multiline for %s. Writing content to /tmp/%s-tmp.crt." % (trusted_server_cert_name, trusted_server_cert_name))
                cacert_file = "/tmp/%s-tmp.crt" % trusted_server_cert_name
                self.write_cert(cacert, cacert_file)
                cacert = cacert_file
            elif not os.path.exists(cacert):
                cacert = None
        client_auth_cert_name=self.__get_client_auth_cert_config_name()
        if client_auth_cert_name in self.ccm_v2_config_map:
            cert = self.ccm_v2_config_map[client_auth_cert_name]
            if cert.startswith("-----"):
                logger.debug("Cert value is multiline for %s. Writing content to /tmp/%s-tmp.crt." % (client_auth_cert_name, client_auth_cert_name))
                cert_file = "/tmp/%s-tmp.crt" % client_auth_cert_name
                self.write_cert(cert, cert_file)
                cert = cert_file
            elif not os.path.exists(cert):
                cert = None
        client_auth_key_name=self.__get_client_auth_key_config_name()
        if client_auth_key_name in self.ccm_v2_config_map:
            key = self.ccm_v2_config_map[client_auth_key_name]
            if key.startswith("-----"):
                logger.debug("Key value is multiline for %s. Writing content to /tmp/%s-tmp.key." % (client_auth_key_name, client_auth_key_name))
                key_file = "/tmp/%s-tmp.key" % client_auth_key_name
                self.write_cert(key, key_file)
                key = key_file
            elif not os.path.exists(key):
                key = None
        return cacert, cert, key

    def write_cert(self, cert, target_file):
        splitted_cert=cert.split("-----")
        with open(target_file, "w") as text_file:
            if len(splitted_cert) > 2:
                text_file.write("-----" + splitted_cert[1] + "-----\n")
                for chunk in self.chunks(splitted_cert[2], 79):
                    text_file.write(chunk + "\n")
                text_file.write("-----" + splitted_cert[3] + "-----")
            else:
                text_file.write(cert)
    
    def chunks(self, string, n):
        for start in range(0, len(string), n):
            yield string[start:start+n]

    def cleanup_jumpgate_tmp_certs(self):
        if self.is_jumpgate_config():
            trusted_server_cert_name=self.__get_trusted_server_cert_config_name()
            client_auth_cert_name=self.__get_client_auth_cert_config_name()
            client_auth_key_name=self.__get_client_auth_key_config_name()
            cacert_file = "/tmp/%s-tmp.crt" % trusted_server_cert_name
            cert_file = "/tmp/%s-tmp.crt" % client_auth_cert_name
            key_file = "/tmp/%s-tmp.key" % client_auth_key_name
            if trusted_server_cert_name and os.path.exists(cacert_file):
                logger.debug("Deleting temporal %s file." % cacert_file)
                os.remove(cacert_file)
            if client_auth_cert_name and os.path.exists(cert_file):
                logger.debug("Deleting temporal %s file." % cert_file)
                os.remove(cert_file)
            if client_auth_key_name and os.path.exists(key_file):
                logger.debug("Deleting temporal %s file." % key_file)
                os.remove(key_file)
    
    def __get_trusted_server_cert_config_name(self):
        return "relayServerCertificate" if self.is_jumpgate else "trustedServerCertificatePath"

    def __get_client_auth_cert_config_name(self):
        return "clientAuthenticationCertificate" if self.is_jumpgate else "certificatePath"

    def __get_client_auth_key_config_name(self):
        return "clientAuthenticationKey" if self.is_jumpgate else "keyPath"

    def __get_relay_server_config_name(self):
        return "relayServer" if self.is_jumpgate else "proxy"

    def __get_http_proxy_url_config_name(self):
        return "https_proxy" if self.is_jumpgate else "httpProxyURL"

    def get_ccm_v1_proxy_settings(self):
        proxy_host=self.ccm_v1_config_map["PROXY_HOST"] if "PROXY_HOST" in self.ccm_v1_config_map else None
        proxy_port=self.ccm_v1_config_map["PROXY_PORT"] if "PROXY_PORT" in self.ccm_v1_config_map else None
        proxy_username=self.ccm_v1_config_map["PROXY_USER"] if "PROXY_USER" in self.ccm_v1_config_map else None
        proxy_password=self.ccm_v1_config_map["PROXY_PASS"] if "PROXY_PASS" in self.ccm_v1_config_map else None
        return proxy_host, proxy_port, proxy_username, proxy_password

    # todo: refactor resolve config functions as there is a lot of duplication in those
    def __resolve_ccm_v2_configs(self):
        if self.is_jumpgate:
            return self.__resolve_ccm_v2_jumpgate_configs()
        else:
            return self.__resolve_ccm_v2_old_configs()
    
    def __resolve_ccm_v2_jumpgate_configs(self):
        result={}
        if os.path.exists(self.ccm_v2_jumpgate_config_location):
            with open(self.ccm_v2_jumpgate_config_location) as f:
                current_key=""
                full_value=""
                for line in f:
                    lineStripped=line.strip()
                    if current_key:
                        if lineStripped.endswith('"""'):
                            full_value=full_value + lineStripped.replace('"', '')
                            result[current_key]=full_value
                            current_key=""
                        else:
                            full_value=full_value + lineStripped
                            result[current_key]=lineStripped
                        continue
                    splitted=lineStripped.split("=")
                    if len(splitted) > 1:
                        if splitted[1].strip().startswith('"""'):
                            current_key=splitted[0].strip()
                            logger.debug("Found multiline config with name %s" % current_key)
                            full_value=splitted[1].strip().replace('"', '')
                            result[current_key]=full_value
                        relayServerKey=self.__get_relay_server_config_name()
                        if splitted[0].strip() == relayServerKey:
                            self.__resolve_splitted_proxy_host(result, splitted, relayServerKey)
                        else:
                            result[splitted[0].strip()]=splitted[1].strip().replace('"', '')
        if os.path.exists(self.ccm_v2_jumpgate_proxy_config_location):
            with open(self.ccm_v2_jumpgate_proxy_config_location) as f:
                for line in f:
                    lineStripped=line.strip()
                    splitted=lineStripped.split("=")
                    if len(splitted) > 1:
                        http_proxy_url_config_name=self.__get_http_proxy_url_config_name()
                        if splitted[0].strip() == http_proxy_url_config_name:
                            self.__resolve_splitted_http_proxy_url(result, splitted, http_proxy_url_config_name)
                        elif splitted[0].strip() == "no_proxy":
                            no_proxy_host_list=splitted[1].strip().replace('"', '')
                            result["no_proxy"]=no_proxy_host_list
                            if no_proxy_host_list:
                                logger.debug("Setting no proxy environment variables as it is found in the configuration: %s" % no_proxy_host_list)
                                os.environ["NO_PROXY"]=no_proxy_host_list
                                os.environ["no_proxy"]=no_proxy_host_list
                        else:
                            result[splitted[0].strip()]=splitted[1].strip().replace('"', '')
        return result
    
    def __resolve_ccm_v2_old_configs(self):
        result={}
        if os.path.exists(self.ccm_v2_config_location):
            with open(self.ccm_v2_config_location) as f:
                for line in f:
                    lineStripped=line.strip()
                    splitted=lineStripped.split("=")
                    if len(splitted) > 1:
                        relayServerKey=self.__get_relay_server_config_name()
                        http_proxy_url_config_name=self.__get_http_proxy_url_config_name()
                        if splitted[0].strip() == relayServerKey:
                            self.__resolve_splitted_proxy_host(result, splitted, relayServerKey)
                        elif splitted[0].strip() == http_proxy_url_config_name:
                            self.__resolve_splitted_http_proxy_url(result, splitted, http_proxy_url_config_name)
                        else:
                            result[splitted[0].strip()]=splitted[1].strip().replace('"', '')
        return result

    def __resolve_ccm_v1_configs(self):
        result={}
        if not self.is_v2():
            if os.path.exists(TUNNEL_FILE_CONFIG):
                with open(TUNNEL_FILE_CONFIG) as tunnel_values_sh:
                    for line in tunnel_values_sh:
                        name, var = line.partition("=")[::2]
                        result[name.rstrip()] = str(var.rstrip()) # e.g.: CCM_HOST, CCM_SSH_PORT
                proxy_host, proxy_port = self.__get_ccm_proxy_host_port()
                proxy_user, proxy_pass = self.__get_ccm_proxy_user_pass()
                result["PROXY_HOST"]=proxy_host
                result["PROXY_PORT"]=proxy_port
                result["PROXY_USER"]=proxy_user
                result["PROXY_PASS"]=proxy_pass
        return result

    def __get_ccm_proxy_user_pass(self):
        proxy_auth_file="/root/.ssh/proxy_auth"
        if os.path.exists(proxy_auth_file):
            try:
                with open(proxy_auth_file) as configssh:
                    user_pass = configssh.read().rstrip().split(":")
                    if len(user_pass) > 0:
                        if len(user_pass) > 1:
                            return user_pass[0], user_pass[1]
                        else:
                            return user_pass[0], ""
            except PermissionError as ex:
                return "", ""
            except FileNotFoundError as ex:
                return "", ""
        else:
            return "", ""

    def __get_ccm_proxy_host_port(self):
        try:
            ssh_config="/root/.ssh/config"
            if os.path.exists(ssh_config):
                with open(ssh_config) as configssh:
                    regex = re.findall("ProxyCommand /usr/bin/corkscrew ([^\s]+) ([^\s]+)", configssh.read())
                    if (len(regex) > 0):
                        if (len(regex[0]) > 1):
                            return regex[0][0], regex[0][1]
            else:
                return "", ""
        except PermissionError as ex:
            return "", ""
        except FileNotFoundError as ex:
            return "", ""

    def __resolve_splitted_proxy_host(self, result, line_parts, proxy_key):
        stripped_proxy_str=line_parts[1].strip().replace('"', '')
        result[proxy_key]=stripped_proxy_str
        split_by_http_scheme=stripped_proxy_str.split("http://")
        split_by_https_scheme=stripped_proxy_str.split("https://")
        address=None
        result["ccm_scheme"]="https"
        if len(split_by_http_scheme) > 1:
            address=split_by_http_scheme[1]
            result["ccm_port"]="80"
            result["ccm_scheme"]="http"
        elif len(split_by_https_scheme) > 1:
             address=split_by_https_scheme[1]
             result["ccm_port"]="443"
        else:
            address=split_by_https_scheme[0]
        splitAddress=address.split("/")[0].split(":")
        result["ccm_host"]=splitAddress[0]
        if len(splitAddress) > 1:
            result["ccm_port"]=splitAddress[1]
        return result

    def __resolve_splitted_http_proxy_url(self, result, line_parts, http_proxy_url):
        stripped_http_proxy_str=line_parts[1].strip().replace('"', '')
        result[http_proxy_url]=stripped_http_proxy_str
        split_by_http_scheme=stripped_http_proxy_str.split("http://")
        split_by_https_scheme=stripped_http_proxy_str.split("https://")
        addressWithoutScheme=None
        result["http_proxy_scheme"]="http"
        if len(split_by_http_scheme) > 1:
            addressWithoutScheme=split_by_http_scheme[1]
        elif len(split_by_https_scheme) > 1:
            addressWithoutScheme=split_by_https_scheme[1]
            result["http_proxy_scheme"]="https"
        else:
            addressWithoutScheme=split_by_http_scheme[0]
        addressWithAuth=addressWithoutScheme.split("@")
        address=addressWithAuth[0]
        if len(addressWithAuth) > 1:
            address=addressWithAuth[1]
            user_auth=addressWithAuth[0].split(":")
            if len(user_auth) > 1:
                result["http_proxy_username"]=user_auth[0]
                result["http_proxy_password"]=user_auth[1]
            address_with_port=address.split(":")
            result["http_proxy_host"]=address_with_port[0]
            if len(address_with_port) > 1:
                result["http_proxy_port"]=address_with_port[1]
            elif len(split_by_https_scheme) > 1:
                result["http_proxy_port"]="443"
            else:
                result["http_proxy_port"]="80"
        return result

