
import os
import sys
import requests
import logging
import urllib3
from .proxy_settings import ProxySettings

urllib3.disable_warnings(urllib3.exceptions.SecurityWarning)

logger=logging.getLogger(__name__)

NO_PROXY_FILE="/etc/cdp-telemetry/conf/proxy-whitelist.txt"
# NO_PROXY_FILE="examples/network/proxy-whitelist.txt"

def smart_connection_check(url, timeout=60, proxySettings=None, cacert=None, cert_key_pair=(None, None)):
    verify_cert=get_requests_ca_bundle_from_cert(cacert)
    has_any_proxy=True if get_any_http_proxy() else False
    result=False
    if has_any_proxy:
        should_bypass=requests.utils.should_bypass_proxies(url, None)
        if should_bypass:
            logger.debug("%s should bypass proxies" % url)
            result=connect(url, timeout, verify_cert, cert_key_pair)
        else:
            url_with_https=True if url.startswith("https://") else False
            result=check_proxy_connections_by_url_scheme(url_with_https, url, timeout, cacert, verify_cert, cert_key_pair)
    else:
        result=connect(url, timeout, verify_cert, cert_key_pair)
    return result

def check_proxy_connections_by_url_scheme(is_https, url, timeout, cacert, verify_cert, cert_key_pair):
    result=False
    possible_proxies=requests.utils.get_environ_proxies(url)
    selected_proxy=requests.utils.select_proxy(url, possible_proxies)
    generatedProxySettings=ProxySettings()
    full_proxy_url=generatedProxySettings.get_full_proxy_url()
    proxy_url=generatedProxySettings.get_url()
    use_https_proxy_first=True if full_proxy_url.startswith("https://") or cacert else False
    scheme_msg="https" if is_https else "http"
    proxy_env_msg="https_proxy" if is_https else "http_proxy"
    if selected_proxy:
        logger.debug("Found selected %s proxy. Skipping connection check without proxy." % (scheme_msg))
    else:
        logger.debug("Not found %s proxy for %s. Trying without proxy first ..." % (scheme_msg, url))
        result=connect(url, timeout, verify_cert, cert_key_pair)
        if not result:
            logger.debug("Now try proxies by setting %s" % proxy_env_msg)
    if not result:
        use_proxy(full_proxy_url, proxy_url, is_https, use_https_proxy_first)
        result=connect(url, timeout, verify_cert, cert_key_pair)
    if not result:
        use_proxy(full_proxy_url, proxy_url, is_https, not use_https_proxy_first)
        result=connect(url, timeout, verify_cert, cert_key_pair)
    return result

def use_proxy(full_proxy_url, proxy_url, use_https_proxy, use_https_proxy_url):
    final_proxy_url=None
    final_full_proxy_url=None
    proxy_to_use=None
    if use_https_proxy:
        proxy_to_use="https_proxy"
        delete_env_var("http_proxy")
    else:
        proxy_to_use="http_proxy"
        delete_env_var("https_proxy")
    if use_https_proxy_url:
        final_full_proxy_url=full_proxy_url.replace("http://", "https://") if full_proxy_url.startswith("http://") else full_proxy_url
        final_proxy_url=proxy_url.replace("http://", "https://") if full_proxy_url.startswith("http://") else proxy_url
    else:
        final_full_proxy_url=full_proxy_url.replace("https://", "http://") if full_proxy_url.startswith("https://") else full_proxy_url
        final_proxy_url=proxy_url.replace("https://", "http://") if full_proxy_url.startswith("https://") else proxy_url
    os.environ[proxy_to_use]=final_full_proxy_url
    os.environ[proxy_to_use.upper()]=final_full_proxy_url
    logger.debug("Proxy - use %s=%s" % (proxy_to_use, final_proxy_url))

def delete_env_var(env_var):
    if env_var in os.environ:
        del os.environ[env_var]
    if env_var.upper() in os.environ:
        del os.environ[env_var.upper()]

def connect(url, timeout=60, verify_cert=None, cert_key_pair=(None, None)):
    try:
        if cert_key_pair[0] and cert_key_pair[1]:
            _ = requests.get(url, timeout=timeout, headers={'Connection':'close'}, verify=verify_cert, cert=cert_key_pair)
        else:
            _ = requests.get(url, timeout=timeout, headers={'Connection':'close'}, verify=verify_cert)
        return True
    except requests.exceptions.ReadTimeout:
        try:
            logger.debug("Read timeout occurred, 1 more try...")
            if cert_key_pair[0] and cert_key_pair[1]:
                _ = requests.get(url, timeout=timeout, headers={'Connection':'close'}, verify=verify_cert, cert=cert_key_pair)
            else:
                _ = requests.get(url, timeout=timeout, headers={'Connection':'close'}, verify=verify_cert)
            return True
        except requests.ReadTimeout as rt_ex:
            logger.debug("Read timeout during connection check: %s" % str(rt_ex))
            return False
        except requests.exceptions.Timeout as ext:
            logger.debug("Timeout error during connection check: %s" % str(ext))
            return False
        except requests.ConnectionError as ex2:
            logger.debug("Connection error during connection check: %s" % str(ex2))
            return False
    except requests.ConnectionError as ex:
        logger.debug("Connection error during connection check: %s" % str(ex))
        return False
    except requests.exceptions.Timeout as ext:
        logger.debug("Timeout error during connection check: %s" % str(ext))
        return False
    return True

def get_requests_ca_bundle_from_cert(cacert=None):
    if cacert:
        logger.debug("URL connection check will use %s cacert file." % cacert)
        os.environ["REQUESTS_CA_BUNDLE"]=cacert
    else:
        if not os.getenv("REQUESTS_CA_BUNDLE"):
            cacert=get_ssl_cert_file()
            if cacert:
                logger.debug("URL connection check will use %s cacert file." % cacert)
                os.environ["REQUESTS_CA_BUNDLE"]=cacert
        else:
            logger.debug("URL connection check will use %s cacert file." % os.getenv("REQUESTS_CA_BUNDLE"))
    verify_cert=False
    if "REQUESTS_CA_BUNDLE" in os.environ:
        verify_cert=cacert
    return verify_cert

def connected(url, timeout=5, proxySettings=None, clearProxy=False, cacert=None, cert_key_pair=(None, None)):
    init_no_proxy()
    if proxySettings:
        proxy_url=proxySettings.get_full_proxy_url()
        if url.startswith("https") or cacert:
            logger.debug("Use HTTPS_PROXY for checking %s" % url)
            os.environ["HTTPS_PROXY"]=proxy_url
            os.environ["https_proxy"]=proxy_url
        else:
            logger.debug("Use HTTP_PROXY for checking %s" % url)
            os.environ["HTTP_PROXY"]=proxy_url
            os.environ["http_proxy"]=proxy_url
    if clearProxy:
        if "HTTPS_PROXY" in os.environ:
            logger.debug("HTTPS proxy will be cleared for checking %s " % url)
            del os.environ["HTTPS_PROXY"]
        if "https_proxy" in os.environ:
            logger.debug("https proxy will be cleared for checking %s " % url)
            del os.environ["https_proxy"]
        if "HTTP_PROXY" in os.environ:
            logger.debug("HTTP proxy will be cleared for checking %s " % url)
            del os.environ["HTTP_PROXY"]
        if "http_proxy" in os.environ:
            logger.debug("http proxy will be cleared for checking %s " % url)
            del os.environ["http_proxy"]
    log_proxies(url)
    return smart_connection_check(url, timeout, proxySettings, cacert, cert_key_pair)

def check_connection(url):
    proxy=get_any_http_proxy()
    proxy_url_message=None
    if proxy:
        proxySettings=ProxySettings(proxy)
        proxy_url=proxySettings.get_url()
        proxy_url_message=proxy.replace("https://", "http(s)://") if proxy_url.startswith("https://") else proxy_url.replace("http://", "http(s)://")
    url_helper_message="url: %s, proxy: %s" % (url, proxy_url_message) if proxy else "url: %s" % url
    if connected(url):
        logger.info("Connection successful. (%s)" % (url_helper_message))
    else:
        logger.error("Connection failed. (%s)" % (url_helper_message))
        sys.exit(1)
        
def get_ssl_cert_file():
    certs_to_find=['/etc/ssl/certs/ca-certificates.crt', '/etc/pki/tls/certs/ca-bundle.crt', '/usr/share/ssl/certs/ca-bundle.crt', '/usr/local/share/certs/ca-root-nss.crt', '/etc/ssl/cert.pem']
    for cert in certs_to_find:
        if os.path.exists(cert):
            return cert
    return None

def log_proxies(url):
    env_vars=["HTTPS_PROXY", "https_proxy", "HTTP_PROXY", "http_proxy"]
    for env_var in env_vars:
        if env_var in os.environ:
            logger.debug("%s is set for %s" % (env_var, url))

def get_any_http_proxy():
    env_vars=["HTTPS_PROXY", "HTTP_PROXY", "https_proxy", "http_proxy"]
    found_proxy=None
    for env_var in env_vars:
        if env_var in os.environ:
            found_proxy=os.getenv(env_var)
    return found_proxy

def init_no_proxy():
    if "no_proxy" in os.environ:
        logger.debug("Found no proxy settings: %s" % os.environ["no_proxy"])
    elif "NO_PROXY" in os.environ:
        logger.debug("Found no proxy settings: %s" % os.environ["NO_PROXY"])
    elif os.path.exists(NO_PROXY_FILE):
        no_proxy_lines=[]
        with open(NO_PROXY_FILE, "r") as no_proxy_list_file:
            no_proxy_lines = [line.strip() for line in no_proxy_list_file if line.strip()]
        if no_proxy_lines:
            logger.debug("Found whitlisted address patterns in whitelist file (%s): %s"% (NO_PROXY_FILE, str(no_proxy_lines)))
            os.environ["NO_PROXY"]=",".join(no_proxy_lines)
            os.environ["no_proxy"]=",".join(no_proxy_lines)