package filecollector

import (
	"io/ioutil"
	"path/filepath"
	"strings"

	pcre "github.com/gijsbers/go-pcre"
	yaml "github.com/goccy/go-yaml"
)

type FileCollectorConfig struct {
	Collector map[string]interface{} `yaml:collector",omitempty"`
}

type CollectorConfig struct {
	OutputLocation                      string
	FileRelativePrefix                  string
	Files                               []CollectedFile
	Rules                               []AnyonymizationRule
	FileLimitBytes                      int64
	NumberOfLinesToKeepOnLimit          int64
	BytesBufferSizeForAnonymiziation    int64
	FreePageCacheOnStart                bool
	FreePageCacheOnFinish               bool
	WorkspaceLimitBytes                 int64
	WorkspaceWorkers                    int64
	WorkspaceMetadataFileName           string
	WorkspaceMetadataExtraFields        map[string]string
	AdditionalWorkspaceMetadata         map[string]interface{}
	AdditionalWorkspaceMetadataFileName string
	IncludeWorkspaceMetadata            bool
	Compress                            bool
	CompressFormat                      string
	UseFullPath                         bool
	RequiredDiskSpaceRatio              float64
	RequiredFreeMemoryRatio             float64
	DeleteProcessedTempFiles            bool
	DeleteCompressedFile                bool
	LogProcessorWorkers                 int64
}

type CollectedFile struct {
	Label             string
	Path              string
	Excludes          []string
	Compress          bool
	CompressAllFiles  bool
	UseFullPath       bool
	FolderPrefix      string
	Mandatory         bool
	SkipAnonymization bool
	SkipLabelFromPath bool
}

type AnyonymizationRule struct {
	Pattern     string
	Regex       pcre.Regexp
	Replacement []byte
	Labels      []string
}

type UserInput struct {
	Labels   []string
	FromDate int64
	ToDate   int64
}

type Workspace struct {
	FilesPerLabel map[string][]FileObject
	Name          string
	Size          int64
	RootWorkDir   string
	TempDir       string
}

type FileObject struct {
	FileName       string
	DestFileName   string
	TempFileName   string
	WorkDirectory  string
	LabelDirectory string
	FileConfigs    CollectedFile
	Rules          []AnyonymizationRule
}

type FileProcessorJob struct {
	File      FileObject
	Workspace Workspace
	Config    *CollectorConfig
}

type FileProcessorJobResult struct {
	FileName      string
	Label         string
	Result        string
	WorkspaceName string
}

type LabelArchiverJob struct {
	Folder string
}

type LabelArchiverJobResult struct {
	LabelFolder string
	Result      string
}

type WorkspaceArchiverJob struct {
	WorkspaceName        string
	RootWorkFolder       string
	TempFolder           string
	DeleteProcessedFiles bool
}

type WorkspaceArchiverJobResult struct {
	WorkspaceName string
	Result        string
}

func ConvertToCollectorConfig(collectorMap map[string]interface{}) CollectorConfig {
	collectorConfig := CollectorConfig{}
	collectorConfig.Rules = ConvertToRules(collectorMap)
	collectorConfig.OutputLocation = ConvertToString(collectorMap, "outputLocation", "")
	collectorConfig.FileRelativePrefix = ConvertToString(collectorMap, "fileRelativePrefix", "")
	collectorConfig.RequiredDiskSpaceRatio = ConvertToFloat64(collectorMap, "requiredDiskSpaceRatio", float64(1.0))
	collectorConfig.RequiredFreeMemoryRatio = ConvertToFloat64(collectorMap, "requiredFreeMemoryRatio", float64(1.3))
	collectorConfig.WorkspaceWorkers = ConvertToInt64(collectorMap, "workspaceWorkers", int64(1))
	collectorConfig.FileLimitBytes = ConvertToInt64(collectorMap, "fileLimitBytes", int64(104857600))
	collectorConfig.NumberOfLinesToKeepOnLimit = ConvertToInt64(collectorMap, "numberOfLinesToKeepOnLimit", int64(1000000))
	collectorConfig.BytesBufferSizeForAnonymiziation = ConvertToInt64(collectorMap, "bytesBufferSizeForAnonymiziation", int64(10485760))
	collectorConfig.WorkspaceLimitBytes = ConvertToInt64(collectorMap, "workspaceLimitBytes", int64(0))
	collectorConfig.WorkspaceMetadataFileName = ConvertToString(collectorMap, "workspaceMetadataFileName", "filecollector-metadata.json")
	// because of a typo in cloudbreak:
	collectorConfig.AdditionalWorkspaceMetadataFileName = ConvertToString(collectorMap, "additionalWorkspaceMetadataFileNmae", "")
	if len(collectorConfig.AdditionalWorkspaceMetadataFileName) == 0 {
		collectorConfig.AdditionalWorkspaceMetadataFileName = ConvertToString(collectorMap, "additionalWorkspaceMetadataFileName", "bundle-info.json")
	}
	collectorConfig.IncludeWorkspaceMetadata = ConvertToBool(collectorMap, "includeWorkspaceMetadata", true)
	collectorConfig.FreePageCacheOnStart = ConvertToBool(collectorMap, "freePageCacheOnStart", false)
	collectorConfig.FreePageCacheOnFinish = ConvertToBool(collectorMap, "freePageCacheOnFinish", false)
	collectorConfig.Compress = ConvertToBool(collectorMap, "compress", true)
	collectorConfig.CompressFormat = ConvertToString(collectorMap, "compressFormat", "zip")
	collectorConfig.DeleteCompressedFile = ConvertToBool(collectorMap, "deleteCompressedFile", false)
	collectorConfig.DeleteProcessedTempFiles = ConvertToBool(collectorMap, "deleteProcessedTempFiles", true)
	collectorConfig.LogProcessorWorkers = ConvertToInt64(collectorMap, "logProcessorWorkers", int64(1))
	collectorConfig.UseFullPath = ConvertToBool(collectorMap, "useFullPath", true)
	collectorConfig.Files = ConvertToFiles(collectorMap, &collectorConfig)
	workspaceMetadataExtraFields := make(map[string]string)
	if wmefVal, ok := collectorMap["workspaceMetadataExtraFields"]; ok {
		wmefItem := wmefVal.(map[string]interface{})
		for k, v := range wmefItem {
			workspaceMetadataExtraFields[k] = v.(string)
		}
	}
	collectorConfig.WorkspaceMetadataExtraFields = workspaceMetadataExtraFields
	if awmVal, ok := collectorMap["additionalWorkspaceMetadata"]; ok {
		collectorConfig.AdditionalWorkspaceMetadata = awmVal.(map[string]interface{})
	}
	return collectorConfig
}

func ConvertToInt64(valueMap map[string]interface{}, key string, defaultValue int64) int64 {
	retVal := defaultValue
	if itVal, ok := valueMap[key]; ok {
		retVal = int64(itVal.(uint64))
	}
	return retVal
}

func ConvertToFloat64(valueMap map[string]interface{}, key string, defaultValue float64) float64 {
	retVal := defaultValue
	if itVal, ok := valueMap[key]; ok {
		retVal = itVal.(float64)
	}
	return retVal
}

func ConvertToString(valueMap map[string]interface{}, key string, defaultValue string) string {
	retVal := defaultValue
	if itVal, ok := valueMap[key]; ok {
		retVal = itVal.(string)
	}
	return retVal
}

func ConvertToBool(valueMap map[string]interface{}, key string, defaultValue bool) bool {
	retVal := defaultValue
	if itVal, ok := valueMap[key]; ok {
		retVal = itVal.(bool)
	}
	return retVal
}

func ConvertToFiles(valueMap map[string]interface{}, collectorConfig *CollectorConfig) []CollectedFile {
	retVal := make([]CollectedFile, 0)
	if itVal, ok := valueMap["files"]; ok {
		filesList := itVal.([]interface{})
		for _, fileI := range filesList {
			fileItem := fileI.(map[string]interface{})
			collectedFile := CollectedFile{}
			if labelVal, ok := fileItem["label"]; ok {
				collectedFile.Label = labelVal.(string)
			}
			if pathVal, ok := fileItem["path"]; ok {
				collectedFile.Path = pathVal.(string)
			}
			if compressVal, ok := fileItem["compress"]; ok {
				collectedFile.Compress = compressVal.(bool)
			}
			if compressAllFilesVal, ok := fileItem["compressAllFiles"]; ok {
				collectedFile.CompressAllFiles = compressAllFilesVal.(bool)
			}
			if useFullPathVal, ok := fileItem["useFullPath"]; ok {
				collectedFile.UseFullPath = useFullPathVal.(bool)
			} else {
				collectedFile.UseFullPath = collectorConfig.UseFullPath
			}
			if mandatoryVal, ok := fileItem["mandatory"]; ok {
				collectedFile.Mandatory = mandatoryVal.(bool)
			}
			if skipAnonymizationVal, ok := fileItem["skipAnonymization"]; ok {
				collectedFile.SkipAnonymization = skipAnonymizationVal.(bool)
			}
			if skipLabelFromPathVal, ok := fileItem["skipLabelFromPath"]; ok {
				collectedFile.SkipLabelFromPath = skipLabelFromPathVal.(bool)
			}
			if folderPrefixVal, ok := fileItem["folderPrefix"]; ok {
				collectedFile.FolderPrefix = folderPrefixVal.(string)
			}
			if excludesVal, ok := fileItem["excludes"]; ok {
				excludesIfArr := excludesVal.([]interface{})
				excludes := make([]string, 0)
				for _, excludesI := range excludesIfArr {
					excludes = append(excludes, excludesI.(string))
				}
				collectedFile.Excludes = excludes
			}
			retVal = append(retVal, collectedFile)
		}
	}
	return retVal
}

func ConvertToRules(valueMap map[string]interface{}) []AnyonymizationRule {
	retVal := make([]AnyonymizationRule, 0)
	if itVal, ok := valueMap["rules"]; ok {
		ruleList, parseOk := itVal.([]interface{})
		if !parseOk {
			return retVal
		}
		for _, ruleI := range ruleList {
			anyonymizationRule := AnyonymizationRule{}
			ruleItem := ruleI.(map[string]interface{})
			if patternVal, ok := ruleItem["pattern"]; ok {
				anyonymizationRule.Pattern = patternVal.(string)
				re, _ := pcre.Compile(anyonymizationRule.Pattern, 0)
				anyonymizationRule.Regex = re
			}
			if replacementVal, ok := ruleItem["replacement"]; ok {
				anyonymizationRule.Replacement = []byte(replacementVal.(string))
			} else {
				anyonymizationRule.Replacement = []byte("[REDACTED]")
			}
			if labelsVal, ok := ruleItem["labels"]; ok {
				labelsIfArr := labelsVal.([]interface{})
				labels := make([]string, 0)
				for _, labelI := range labelsIfArr {
					labels = append(labels, labelI.(string))
				}
				anyonymizationRule.Labels = labels
			}
			retVal = append(retVal, anyonymizationRule)
		}
	}
	return retVal
}

func ParseCollectorConfig(filePath string, executableDir string) (*CollectorConfig, error) {
	b, err := ioutil.ReadFile(filePath)
	if err != nil {
		return nil, err
	}
	var filecollectorConfig FileCollectorConfig
	readErr := yaml.Unmarshal(b, &filecollectorConfig)
	if readErr != nil {
		return nil, readErr
	}
	collectorConfig := ConvertToCollectorConfig(filecollectorConfig.Collector)
	if len(collectorConfig.FileRelativePrefix) > 0 {
		executableDirParent, execDirErr := filepath.Abs(filepath.Dir(executableDir))
		if execDirErr != nil {
			return nil, execDirErr
		}
		configFileFolder, confFileFolderErr := filepath.Abs(filepath.Dir(filePath))
		if confFileFolderErr != nil {
			return nil, confFileFolderErr
		}
		configFileFolderParent, confFileFoldeParentrErr := filepath.Abs(filepath.Dir(configFileFolder))
		if confFileFoldeParentrErr != nil {
			return nil, confFileFoldeParentrErr
		}
		collectorConfig.FileRelativePrefix = strings.Replace(collectorConfig.FileRelativePrefix, "${configFileFolder}", configFileFolder, 1)
		collectorConfig.FileRelativePrefix = strings.Replace(collectorConfig.FileRelativePrefix, "${configFileFolderParent}", configFileFolderParent, 1)
		collectorConfig.FileRelativePrefix = strings.Replace(collectorConfig.FileRelativePrefix, "${executableDirParent}", executableDirParent, 1)
	}
	return &collectorConfig, nil
}
