package filecollector

import (
	"archive/tar"
	"compress/gzip"
	models "filecollector/models"
	"fmt"
	"io"
	"os"
	"path/filepath"
	"strings"
	"sync"
)

func ArchiveLabelProcessor(id int, wg *sync.WaitGroup, results *[]models.LabelArchiverJobResult, mutex *sync.RWMutex, labelArchiverChannel <-chan models.LabelArchiverJob) {
	defer wg.Done()
	for labelArchiverWorker := range labelArchiverChannel {
		result := archiveLabelFolderProcess(id, labelArchiverWorker)
		mutex.Lock()
		*results = append(*results, result)
		mutex.Unlock()
	}
}

func archiveLabelFolderProcess(id int, job models.LabelArchiverJob) models.LabelArchiverJobResult {
	err := ArchiveLabelFolder(job.Folder)
	if err != nil {
		return models.LabelArchiverJobResult{Result: fmt.Sprintf("ERROR %s", err), LabelFolder: job.Folder}
	}
	return models.LabelArchiverJobResult{Result: "OK", LabelFolder: job.Folder}
}

func ArchiveWorkspaceProcessor(id int, wg *sync.WaitGroup, results *[]models.WorkspaceArchiverJobResult, mutex *sync.RWMutex, workspaceArchiverChannel <-chan models.WorkspaceArchiverJob) {
	defer wg.Done()
	for workspaceArchiverWorker := range workspaceArchiverChannel {
		result := archiveWorkspaceProcess(id, workspaceArchiverWorker)
		mutex.Lock()
		*results = append(*results, result)
		mutex.Unlock()
	}
}

func archiveWorkspaceProcess(id int, job models.WorkspaceArchiverJob) models.WorkspaceArchiverJobResult {
	err := ArchiveWorkspace(job.TempFolder, job.RootWorkFolder, job.DeleteProcessedFiles)
	if err != nil {
		return models.WorkspaceArchiverJobResult{Result: fmt.Sprintf("ERROR %s", err), WorkspaceName: job.WorkspaceName}
	}
	return models.WorkspaceArchiverJobResult{Result: "OK", WorkspaceName: job.WorkspaceName}
}

func ArchiveWorkspace(folderSrc string, targetFolder string, deleteSrc bool) error {
	files := make(map[string]string)
	err := filepath.Walk(folderSrc,
		func(path string, info os.FileInfo, err error) error {
			if err != nil {
				return err
			}
			if !info.IsDir() {
				splitted := strings.Split(path, folderSrc+string(os.PathSeparator))
				filePathKey := filepath.Join(filepath.Base(folderSrc), splitted[1])
				files[filePathKey] = path
			}
			return nil
		})
	if err != nil {
		return err
	}
	folderBaseName := filepath.Base(folderSrc)
	compressedFileName := filepath.Join(targetFolder, folderBaseName)
	out, err := os.Create(fmt.Sprintf("%s.tar.gz", compressedFileName))
	if err != nil {
		log.Error("Error writing archive:", err)
		return err
	}
	defer out.Close()
	err = createArchiveFromMultipleFiles(files, out, true)
	if err != nil {
		log.Error("Error creating archive:", err)
		return err
	}
	if deleteSrc {
		rmErr := os.RemoveAll(folderSrc)
		if rmErr != nil {
			log.Error(fmt.Sprintf("Error during removing label folder %s: %s", folderSrc, rmErr))
		}
	}
	return nil
}

func ArchiveLabelFolder(folderSrc string) error {
	files := make(map[string]string)
	err := filepath.Walk(folderSrc,
		func(path string, info os.FileInfo, err error) error {
			if err != nil {
				return err
			}
			if !info.IsDir() {
				splitted := strings.Split(path, folderSrc+string(os.PathSeparator))
				filePathKey := filepath.Join(filepath.Base(folderSrc), splitted[1])
				files[filePathKey] = path
			}
			return nil
		})
	if err != nil {
		return err
	}
	out, err := os.Create(fmt.Sprintf("%s.tar.gz", folderSrc))
	if err != nil {
		log.Error("Error writing archive:", err)
		return err
	}
	defer out.Close()
	err = createArchiveFromMultipleFiles(files, out, false)
	if err != nil {
		log.Error("Error creating archive:", err)
		return err
	}
	rmErr := os.RemoveAll(folderSrc)
	if rmErr != nil {
		log.Error(fmt.Sprintf("Error during removing label folder %s: %s", folderSrc, rmErr))
	}
	return nil
}

func ArchiveSingleFile(filePath string, deleteSrc bool) error {
	out, err := os.Create(fmt.Sprintf("%s.tar.gz", filePath))
	if err != nil {
		log.Error("Error writing archive:", err)
		return err
	}
	defer out.Close()
	err = createArchiveFromSingleFile(filePath, out, false)
	if err != nil {
		log.Error("Error creating archive:", err)
		return err
	}
	if deleteSrc {
		rmErr := os.Remove(filePath)
		if rmErr != nil {
			log.Error(fmt.Sprintf("Error during removing %s: %s", filePath, rmErr))
		}
	}
	return nil
}

func createArchiveFromMultipleFiles(files map[string]string, buf io.Writer, skipRoot bool) error {
	gw := gzip.NewWriter(buf)
	defer gw.Close()
	tw := tar.NewWriter(gw)
	defer tw.Close()
	for basefileWithPath, file := range files {
		err := addToArchive(tw, file, basefileWithPath, skipRoot)
		if err != nil {
			return err
		}
	}
	return nil
}

func createArchiveFromSingleFile(file string, buf io.Writer, skipRoot bool) error {
	gw := gzip.NewWriter(buf)
	defer gw.Close()
	tw := tar.NewWriter(gw)
	defer tw.Close()
	err := addToArchive(tw, file, filepath.Base(file), skipRoot)
	if err != nil {
		return err
	}
	return nil
}

func addToArchive(tw *tar.Writer, filename string, headername string, skipRoot bool) error {
	file, err := os.Open(filename)
	if err != nil {
		return err
	}
	defer file.Close()
	info, err := file.Stat()
	if err != nil {
		return err
	}
	header, err := tar.FileInfoHeader(info, info.Name())
	if err != nil {
		return err
	}
	if skipRoot {
		splitted := strings.SplitN(headername, string(os.PathSeparator), 2)
		if len(splitted) > 1 {
			headername = splitted[1]
		} else {
			headername = splitted[0]
		}
	}
	header.Name = headername
	err = tw.WriteHeader(header)
	if err != nil {
		return err
	}
	_, err = io.Copy(tw, file)
	if err != nil {
		return err
	}
	return nil
}
