package filecollector

import (
	"errors"
	models "filecollector/models"
	"fmt"
	"io"
	"os"
	"os/exec"
	"sync"
)

func FileProcessor(id int, wg *sync.WaitGroup, results *[]models.FileProcessorJobResult, mutex *sync.RWMutex, fileProcessorChannel <-chan models.FileProcessorJob) {
	defer wg.Done()
	for fileWorker := range fileProcessorChannel {
		result := process(id, fileWorker)
		mutex.Lock()
		*results = append(*results, result)
		mutex.Unlock()
	}
}

func process(id int, processor models.FileProcessorJob) models.FileProcessorJobResult {
	log.Debug(fmt.Sprintf(" %d - file processing: %s", id, processor.File.FileName))
	fileObject := processor.File
	config := processor.Config
	compress := fileObject.FileConfigs.Compress
	fullFileName := fileObject.FileName
	copyDestination := fileObject.DestFileName
	tempDestination := fileObject.TempFileName
	result := "OK"
	checkFileErr := preCheckFileExists(fullFileName)
	if checkFileErr != nil {
		result = fmt.Sprintf("ERROR %s - %s", fullFileName, checkFileErr)
	}
	if result == "OK" {
		err := processFile(fileObject, fullFileName, copyDestination, tempDestination, config)
		if err != nil {
			result = fmt.Sprintf("ERROR %s - %s", fullFileName, err)
		}
	}
	cleanupErr := cleanupPreProcessedFiles(tempDestination)
	if cleanupErr != nil {
		log.Debug(fmt.Sprintf("Cleanup error: %s", cleanupErr))
	}
	if result == "OK" && compress {
		archErr := ArchiveSingleFile(copyDestination, true)
		if archErr != nil {
			log.Debug(fmt.Sprintf("Archive error for %s: %s", copyDestination, archErr))
		}
	}
	return models.FileProcessorJobResult{FileName: fullFileName, Result: result, Label: fileObject.FileConfigs.Label, WorkspaceName: processor.Workspace.Name}
}

func processFile(fileObj models.FileObject, filePath string, destPath string, tempPath string, config *models.CollectorConfig) error {
	limitBytes := config.FileLimitBytes
	numberOfLinesToKeep := config.NumberOfLinesToKeepOnLimit
	if len(tempPath) > 0 {
		log.Info(fmt.Sprintf("Copying %s to %s", filePath, tempPath))
		err := copyOrTail(filePath, tempPath, limitBytes, numberOfLinesToKeep)
		if err != nil {
			return err
		}
		redactErr := RedactFile(fileObj, config.BytesBufferSizeForAnonymiziation)
		if redactErr != nil {
			return redactErr
		}
	} else {
		log.Info(fmt.Sprintf("Copying %s to %s", filePath, destPath))
		err := copyOrTail(filePath, destPath, limitBytes, numberOfLinesToKeep)
		if err != nil {
			return err
		}
	}
	return nil
}

func copyOrTail(src, dst string, limitBytes int64, keepLines int64) error {
	osFileStat, osFileStatErr := os.Stat(src)
	if osFileStatErr != nil {
		return osFileStatErr
	}
	fileSize := osFileStat.Size()
	if limitBytes != 0 && fileSize > limitBytes {
		cmd := exec.Command("/bin/bash", "-c", fmt.Sprintf("tail -n %d %s > %s", keepLines, src, dst))
		err := cmd.Run()
		if err != nil {
			return err
		}
	} else {
		_, err := copy(src, dst)
		if err != nil {
			return err
		}
	}
	return nil
}

func copy(src, dst string) (int64, error) {
	sourceFileStat, err := os.Stat(src)
	if err != nil {
		return 0, err
	}

	if !sourceFileStat.Mode().IsRegular() {
		return 0, fmt.Errorf("%s is not a regular file", src)
	}

	source, err := os.Open(src)
	if err != nil {
		return 0, err
	}
	defer source.Close()

	destination, err := os.Create(dst)
	if err != nil {
		return 0, err
	}
	defer destination.Close()
	nBytes, err := io.Copy(destination, source)
	return nBytes, err
}

func preCheckFileExists(path string) error {
	if _, err := os.Stat(path); os.IsExist(err) {
		return errors.New(fmt.Sprintf("It could be file '%s' does not exist!", path))
	}
	return nil
}

func cleanupPreProcessedFiles(path string) error {
	if len(path) == 0 {
		return nil
	}
	if _, err := os.Stat(path); os.IsNotExist(err) {
		return nil
	} else {
		log.Debug(fmt.Sprintf("Removing file: %s", path))
		rmErr := os.Remove(path)
		if rmErr != nil {
			return rmErr
		}
		return nil
	}
}
