package filecollector

import (
	"encoding/json"
	"errors"
	logging "filecollector/logging"
	models "filecollector/models"
	"fmt"
	"io/ioutil"
	"os"
	"os/exec"
	"path"
	"path/filepath"
	"runtime"
	"strings"
	"sync"
	"time"

	"github.com/djherbis/times"
	"github.com/mackerelio/go-osstat/memory"
	"golang.org/x/sys/unix"
)

var log = logging.NewLogger("filecollector.log")

func CollectFiles(collectorConfig *models.CollectorConfig, userInput models.UserInput) error {
	labels := userInput.Labels
	osSeparator := string(os.PathSeparator)
	outputLocation := appendRelativePrefix(collectorConfig.OutputLocation, collectorConfig.FileRelativePrefix, osSeparator)
	allSize := int64(0)
	hostname, hostNameErr := os.Hostname()
	if hostNameErr != nil {
		return hostNameErr
	}
	currentWorkspaceIndex := int32(0)
	currentWorkspaceSize := int64(0)
	fileSizeLimitBytes := collectorConfig.FileLimitBytes
	workspaces := make([]models.Workspace, 0)
	t := time.Now()
	timestamp := strings.Replace(t.Format("2006-01-02-15-04-05.000000"), ".", "-", 1)
	currentWorkspace := createWorkspace(hostname, currentWorkspaceIndex, timestamp, outputLocation)
	labelsSize := len(collectorConfig.Files)
	currentFileObjectsMapForWorkspace := make(map[string][]models.FileObject)
	foundAnyFiles := false
	for labelIndex, fileToProcess := range collectorConfig.Files {
		label := fileToProcess.Label
		log.Debug(fmt.Sprintf("Checking files with label %s", label))
		if len(labels) > 0 && !fileToProcess.Mandatory {
			foundLabel := contains(labels, label)
			if !foundLabel {
				log.Debug(fmt.Sprintf("Label '%s' will be skipped from processing.", label))
				continue
			}
		}
		filePathPattern := appendRelativePrefix(fileToProcess.Path, collectorConfig.FileRelativePrefix, osSeparator)
		matches, mathErr := filepath.Glob(filePathPattern)
		if mathErr != nil {
			return mathErr
		}
		excludes, excludesErr := getExcludes(fileToProcess.Excludes, label, collectorConfig.FileRelativePrefix, osSeparator)
		if excludesErr != nil {
			log.Error(excludesErr)
		}
		if len(matches) == 0 {
			log.Debug(fmt.Sprintf("Not found any matches for path pattern: %s (label: %s)", filePathPattern, label))
		}
		allSizeForLabel := int64(0)
		files := make([]models.FileObject, 0)
		for _, fileMatch := range matches {
			if !foundAnyFiles {
				foundAnyFiles = true
			}
			_, fileExcludeExists := excludes[fileMatch]
			if fileExcludeExists {
				log.Debug(fmt.Sprintf("File '%s' will be excluded from label '%s'", fileMatch, label))
				continue
			}
			log.Debug(fmt.Sprintf("Found file: %s", fileMatch))
			osFileStat, osFileStatErr := os.Stat(fileMatch)
			if osFileStatErr != nil {
				log.Error(osFileStatErr)
				continue
			}
			expectedFileSize := osFileStat.Size()
			if fileSizeLimitBytes != 0 && expectedFileSize > fileSizeLimitBytes {
				expectedFileSize = fileSizeLimitBytes
			}
			allSizeForLabel += expectedFileSize
			fileStat, fileStatErr := times.Stat(fileMatch)
			if fileStatErr != nil {
				log.Error(fileStatErr)
				continue
			}
			excludeByDate, excludeByDateErr := excludeFileByDate(fileMatch, fileStat, userInput.FromDate, userInput.ToDate)
			if excludeByDateErr != nil {
				log.Error(excludeByDateErr)
			}
			if excludeByDate {
				log.Debug(fmt.Sprintf("File '%s' will be excluded as out of date range", fileMatch))
				continue
			}
			log.Info(fmt.Sprintf("Adding file '%s' to workspace '%s'", fileMatch, currentWorkspace.Name))
			fileObj := creaateFileObject(fileMatch, fileToProcess, collectorConfig, currentWorkspace.TempDir)
			files = append(files, fileObj)
		}
		if len(files) > 0 {
			currentFileObjectsMapForWorkspace[label] = files
		}
		currentWorkspace.FilesPerLabel = currentFileObjectsMapForWorkspace
		currentWorkspaceSize += allSizeForLabel
		currentWorkspace.Size = currentWorkspaceSize
		if labelsSize == labelIndex+1 {
			if len(files) > 0 {
				workspaces = append(workspaces, currentWorkspace)
			}
		} else if currentWorkspaceSize > collectorConfig.WorkspaceLimitBytes {
			currentWorkspaceIndex += 1
			currentWorkspaceSize = 0
			workspaces = append(workspaces, currentWorkspace)
			currentWorkspace = createWorkspace(hostname, currentWorkspaceIndex, timestamp, outputLocation)
			currentFileObjectsMapForWorkspace = make(map[string][]models.FileObject)
		}
		allSize += allSizeForLabel
	}
	if !foundAnyFiles {
		return errors.New("Not found any files to process!")
	}
	workdirIsNotExists := isNotExist(outputLocation)
	if workdirIsNotExists {
		os.MkdirAll(outputLocation, 0750)
	}
	os := getOsRuntime()
	if len(os) > 0 && os != "windows" {
		validationErr := ValidateSystemResources(collectorConfig, allSize, outputLocation, os)
		if validationErr != nil {
			return validationErr
		}
	}
	return ProcessWorkspaces(workspaces, outputLocation, collectorConfig)
}

func ValidateSystemResources(collectorConfig *models.CollectorConfig, dataSize int64, workdir string, os string) error {
	var stat unix.Statfs_t
	unix.Statfs(workdir, &stat)
	availableBytes := float64(stat.Bavail * uint64(stat.Bsize))
	requiredDiskSize := float64(dataSize) * collectorConfig.RequiredDiskSpaceRatio
	if requiredDiskSize > availableBytes {
		return errors.New(fmt.Sprintf("Not enough free disk space for file collection - free: %d, required: %d", int64(availableBytes), int64(requiredDiskSize)))
	}
	if collectorConfig.FreePageCacheOnStart && os == "linux" {
		fpce := freePageCache()
		if fpce != nil {
			log.Debug(fpce)
		}
	}
	memory, memErr := memory.Get()
	if memErr != nil {
		return memErr
	}
	freeMemory := memory.Free
	log.Debug(fmt.Sprintf("Available free memory: %d", freeMemory))
	if len(collectorConfig.Rules) > 0 {
		logProcessorWorkers := collectorConfig.LogProcessorWorkers
		possibleBytesBuffer := collectorConfig.BytesBufferSizeForAnonymiziation
		memoryRatio := collectorConfig.RequiredFreeMemoryRatio
		requiredMemory := float64(logProcessorWorkers*possibleBytesBuffer) * memoryRatio
		if float64(freeMemory) > requiredMemory {
			log.Debug(fmt.Sprintf("Required memory for file processing: %f", requiredMemory))
		} else {
			log.Debug(fmt.Sprintf("Not enough free memory. Required memory for file processing: %d", int64(requiredMemory)))
			minimumRequiredMemory := float64(possibleBytesBuffer) * memoryRatio
			workersFit := float64(freeMemory) / minimumRequiredMemory
			if 1.0 > workersFit {
				return errors.New(fmt.Sprintf("Not enough free memory for file collection - free: %d, minimum required: %d", freeMemory, int64(minimumRequiredMemory)))
			} else {
				newNumberOfLogProcessorWorkers := int64(workersFit)
				log.Debug(fmt.Sprintf("Setting a new number of log processor workers (as not enough memory): %d", newNumberOfLogProcessorWorkers))
				collectorConfig.LogProcessorWorkers = newNumberOfLogProcessorWorkers
			}
		}
	}
	return nil
}

func ProcessWorkspaces(workspaces []models.Workspace, outputLocation string, config *models.CollectorConfig) error {
	for _, workspace := range workspaces {
		log.Info(fmt.Sprintf("Processing workspace: %s", workspace.Name))
		log.Debug(fmt.Sprintf("Creating temp forlder for workspace: %s", workspace.TempDir))
		os.MkdirAll(workspace.TempDir, 0750)
		for key, fileObjLabel := range workspace.FilesPerLabel {
			log.Debug(fmt.Sprintf("Creating required directories for %s", key))
			for _, file := range fileObjLabel {
				os.MkdirAll(file.WorkDirectory, 0750)
			}
		}
	}
	if len(config.AdditionalWorkspaceMetadataFileName) > 0 {
		for _, workspace := range workspaces {
			fileNameToWrite := filepath.Join(workspace.TempDir, config.AdditionalWorkspaceMetadataFileName)
			file, wmErr := json.MarshalIndent(config.AdditionalWorkspaceMetadata, " ", " ")
			if wmErr != nil {
				log.Debug(fmt.Sprintf("Error during marshalling content: %s , %s", config.AdditionalWorkspaceMetadata, wmErr))
			}
			writeErr := ioutil.WriteFile(fileNameToWrite, file, 0644)
			if writeErr != nil {
				log.Debug(fmt.Sprintf("Error during writing file: %s , %s", fileNameToWrite, writeErr))
			}
		}
	}
	fileProcessorJobs := make([]models.FileProcessorJob, 0)
	for _, workspace := range workspaces {
		for key, fileObjLabel := range workspace.FilesPerLabel {
			log.Info(fmt.Sprintf("Adding log processor jobs for label %s", key))
			for _, file := range fileObjLabel {
				fileProcessorJob := models.FileProcessorJob{File: file, Workspace: workspace, Config: config}
				fileProcessorJobs = append(fileProcessorJobs, fileProcessorJob)
			}
		}
	}
	var (
		fileProcessingResults []models.FileProcessorJobResult
		fileProcessingMutex   = sync.RWMutex{}
		wg                    sync.WaitGroup
		jobChannel            = make(chan models.FileProcessorJob)
	)
	numberOfWorkers := 1
	if config.LogProcessorWorkers > 1 {
		numberOfWorkers = int(config.LogProcessorWorkers)
	}
	wg.Add(numberOfWorkers)
	for i := 0; i < numberOfWorkers; i++ {
		go FileProcessor(i, &wg, &fileProcessingResults, &fileProcessingMutex, jobChannel)
	}
	for _, fileProcessorJob := range fileProcessorJobs {
		jobChannel <- fileProcessorJob
	}
	close(jobChannel)
	wg.Wait()
	if len(fileProcessingResults) > 0 && config.IncludeWorkspaceMetadata {
		processedWorkspaceResult := make(map[string][]map[string]string)
		for _, workspace := range workspaces {
			processedWorkspaceResult[workspace.Name] = make([]map[string]string, 0)
		}
		for _, result := range fileProcessingResults {
			mapArr := processedWorkspaceResult[result.WorkspaceName]
			mapResult := make(map[string]string)
			mapResult["label"] = result.Label
			mapResult["result"] = result.Result
			mapResult["file"] = result.FileName
			mapArr = append(mapArr, mapResult)
			processedWorkspaceResult[result.WorkspaceName] = mapArr
		}
		baseFileName := config.WorkspaceMetadataFileName
		for _, workspace := range workspaces {
			workspaceData := make(map[string]interface{})
			workspaceData["workspace"] = workspace.Name
			workspaceData["results"] = processedWorkspaceResult[workspace.Name]
			fileNameToWrite := filepath.Join(workspace.TempDir, baseFileName)
			file, _ := json.MarshalIndent(workspaceData, " ", " ")
			_ = ioutil.WriteFile(fileNameToWrite, file, 0644)
		}
	}

	labelFolders := make(map[string]bool)
	for _, workspace := range workspaces {
		for _, fileObjLabel := range workspace.FilesPerLabel {
			for _, file := range fileObjLabel {
				if file.FileConfigs.Compress && len(file.LabelDirectory) > 0 {
					labelFolders[file.LabelDirectory] = true
				}
			}
		}
	}
	workspaceWorkers := 1
	if config.WorkspaceWorkers > 1 {
		workspaceWorkers = int(config.WorkspaceWorkers)
	}
	if len(labelFolders) > 0 {
		labelArchiverJobs := make([]models.LabelArchiverJob, 0)
		for labelFolder, _ := range labelFolders {
			labelArchiverJobs = append(labelArchiverJobs, models.LabelArchiverJob{Folder: labelFolder})
		}
		var (
			labelArchiverResults    []models.LabelArchiverJobResult
			labelArchiverMutex      = sync.RWMutex{}
			wgLabelArch             sync.WaitGroup
			labelArchiverJobChannel = make(chan models.LabelArchiverJob)
		)
		wgLabelArch.Add(workspaceWorkers)
		for i := 0; i < workspaceWorkers; i++ {
			go ArchiveLabelProcessor(i, &wgLabelArch, &labelArchiverResults, &labelArchiverMutex, labelArchiverJobChannel)
		}
		for _, labelArchiverJob := range labelArchiverJobs {
			labelArchiverJobChannel <- labelArchiverJob
		}
		close(labelArchiverJobChannel)
		wgLabelArch.Wait()
	}

	if config.Compress {
		workspaceArchJobs := make([]models.WorkspaceArchiverJob, 0)
		for _, workspace := range workspaces {
			workspaceArchJobs = append(workspaceArchJobs, models.WorkspaceArchiverJob{WorkspaceName: workspace.Name, RootWorkFolder: workspace.RootWorkDir, TempFolder: workspace.TempDir, DeleteProcessedFiles: config.DeleteProcessedTempFiles})
		}
		var (
			workspaceArchiverResults []models.WorkspaceArchiverJobResult
			workspaceArchiverMutex   = sync.RWMutex{}
			wgWorkspaceArch          sync.WaitGroup
			workspaceJobChannel      = make(chan models.WorkspaceArchiverJob)
		)
		wgWorkspaceArch.Add(workspaceWorkers)
		for i := 0; i < workspaceWorkers; i++ {
			go ArchiveWorkspaceProcessor(i, &wgWorkspaceArch, &workspaceArchiverResults, &workspaceArchiverMutex, workspaceJobChannel)
		}
		for _, workspaceArchJob := range workspaceArchJobs {
			workspaceJobChannel <- workspaceArchJob
		}
		close(workspaceJobChannel)
		wgWorkspaceArch.Wait()
	}
	if config.DeleteProcessedTempFiles {
		os.RemoveAll(filepath.Join(outputLocation, "tmp"))
	}
	if config.FreePageCacheOnFinish && getOsRuntime() == "linux" {
		fpce := freePageCache()
		if fpce != nil {
			log.Debug(fpce)
		}
	}
	return nil
}

func getExcludes(excludes []string, label string, relativePrefix string, osSeparator string) (map[string]bool, error) {
	results := make(map[string]bool)
	for _, excludePattern := range excludes {
		excludePattern = appendRelativePrefix(excludePattern, relativePrefix, osSeparator)
		matches, mathErr := filepath.Glob(excludePattern)
		if mathErr != nil {
			return results, mathErr
		}
		for _, fileMatch := range matches {
			results[fileMatch] = true
			log.Debug(fmt.Sprintf("File '%s' will be filtered out from %s", fileMatch, label))
		}
	}
	return results, nil
}

func excludeFileByDate(fileName string, fileStat times.Timespec, startTime int64, endTime int64) (bool, error) {
	result := false
	lastModifiedTime := fileStat.ModTime().Unix()
	log.Debug(fmt.Sprintf("File '%s' last modified timestamp: %d", fileName, lastModifiedTime))
	if startTime != 0 && (lastModifiedTime*1000) < startTime {
		log.Debug(fmt.Sprintf("Skipping processing file '%s' as creation time is larger than 'start-time' parameter", fileName))
		result = true
	} else if fileStat.HasBirthTime() {
		creationTime := fileStat.BirthTime().Unix()
		log.Debug(fmt.Sprintf("File '%s' creation timestamp: %d", fileName, creationTime))
		if endTime != 0 && endTime < (creationTime*1000) {
			log.Debug(fmt.Sprintf("Skipping processing file '%s' as creation time is larger than 'end-time' parameter", fileName))
			result = true
		}
	} else {
		log.Debug(fmt.Sprintf("File '%s' is in the right date range.", fileName))
	}
	return result, nil
}

func contains(slice []string, val string) bool {
	for _, item := range slice {
		if item == val {
			return true
		}
	}
	return false
}

func createWorkspace(hostname string, index int32, timestamp string, outputLocation string) models.Workspace {
	replacedHostname := strings.ReplaceAll(hostname, ".", "-")
	workspaceName := fmt.Sprintf("%s-%s-%d", timestamp, replacedHostname, index)
	workspaceTmpDir := filepath.Join(outputLocation, "tmp", workspaceName)
	return models.Workspace{Name: workspaceName, TempDir: workspaceTmpDir, RootWorkDir: outputLocation}
}

func creaateFileObject(filePath string, config models.CollectedFile, collectorConfig *models.CollectorConfig, tempDir string) models.FileObject {
	fileObj := models.FileObject{FileName: filePath, FileConfigs: config}
	fileBaseDir := path.Dir(filePath)
	fileBaseName := filepath.Base(filePath)
	folderPrefix := config.FolderPrefix
	pathWithFolderPrefix := filepath.Join(tempDir, folderPrefix)
	workdirForLabel := pathWithFolderPrefix
	if !config.SkipLabelFromPath {
		workdirForLabel = filepath.Join(pathWithFolderPrefix, config.Label)
		fileObj.LabelDirectory = workdirForLabel
	}
	var copyDestination string
	if config.UseFullPath {
		dirWithFullPath := filepath.Join(workdirForLabel, fileBaseDir)
		fileObj.WorkDirectory = dirWithFullPath
		copyDestination = filepath.Join(dirWithFullPath, fileBaseName)
	} else {
		fileObj.WorkDirectory = workdirForLabel
		copyDestination = filepath.Join(workdirForLabel, fileBaseName)
	}
	fileObj.DestFileName = copyDestination
	if !config.SkipAnonymization {
		fileObj.Rules = collectorConfig.Rules
		if len(fileObj.Rules) > 0 {
			fileObj.TempFileName = copyDestination + ".preprocessed"
		}
	}
	return fileObj
}

func appendRelativePrefix(path string, relativePrefix string, osSeparator string) string {
	result := path
	if len(relativePrefix) > 0 && !strings.HasPrefix(path, osSeparator) {
		result = filepath.Join(relativePrefix, result)
	}
	return result
}

func isNotExist(path string) bool {
	result := false
	if _, err := os.Stat(path); os.IsNotExist(err) {
		result = true
	}
	return result
}

func freePageCache() error {
	cmd := exec.Command("/bin/bash", "-c", "echo 1 > /proc/sys/vm/drop_caches")
	err := cmd.Run()
	if err != nil {
		return err
	}
	return nil
}

func getOsRuntime() string {
	return runtime.GOOS
}
