package proxy

import (
	"context"
	"encoding/base64"
	"encoding/json"
	"fmt"
	"net/http"
	"strconv"
	"strings"
	"time"

	log "github.com/sirupsen/logrus"
	"github.infra.cloudera.com/cdp-ums-auth-proxy/generated/usermanagement"
)

func (client *UMSClient) AuthN(r *http.Request) (*CdpAuthResponse, string) {
	validated, validateError, authString, token, msgString := validateRequest(r)
	if !validated {
		return &CdpAuthResponse{Success: false, StatusCode: http.StatusBadRequest, Message: validateError}, ""
	}
	requestId := GetRequestIdFromHeader(r)
	ctx := ContextWithInternalActor(context.Background())
	if len(requestId) > 0 {
		log.Debugf("Processing request with id: %s", requestId)
		ctx = ContextWithRequestId(ctx, requestId)
	}
	if len(token) > 0 {
		machineUserCrn, found := client.GetFromAuthNCache(token)
		if !found {
			log.Debugf("Token is not found yet in the cache. Token: [REDACTED]")
			msg, statusCode, failFound := client.GetFromAuthNFailureCache(token)
			if failFound {
				return &CdpAuthResponse{Success: false, StatusCode: statusCode, Message: msg}, ""
			}
			log.Debugf("Token is not found yet in the failure cache. Token: [REDACTED]")
		} else {
			log.Debugf("Got token from cache for %v", machineUserCrn)
			return &CdpAuthResponse{Success: true}, machineUserCrn
		}
		umc := client.userManagementClient
		decodedToken, err := base64.URLEncoding.DecodeString(token)
		if err != nil {
			log.Errorf("Cannot decode token: %v", err)
			return GetErrorResponseWithAddToAuthNFailureCache(client, token, http.StatusBadRequest, "Cannot decode token.")
		}
		log.Debugf("Decoded token: %s", string(decodedToken))
		var tokenObj tokenstr
		json.Unmarshal(decodedToken, &tokenObj)
		validToken, err := checkTokenValidity(token, tokenObj)
		if err != nil {
			return GetErrorResponseWithAddToAuthNFailureCache(client, token, http.StatusBadRequest, fmt.Sprintf("Token expire string is not valid: %s", tokenObj.ExpiredAt))
		}
		if !validToken {
			return GetErrorResponseWithAddToAuthNFailureCache(client, token, http.StatusBadRequest, fmt.Sprintf("Token expire string is not in the right date range. (timestamp: %s)", tokenObj.ExpiredAt))
		}
		splitted := strings.Split(authString, ".")
		newAuthString := fmt.Sprintf("%s.%s", splitted[0], tokenObj.SignedStr)
		tokenAuthResponse, err := umsAuthenticate(umc, ctx, newAuthString, tokenObj.ExpiredAt)
		if err != nil {
			authError := fmt.Sprintf("Auth error (by using auth token) - %v", err)
			return GetErrorResponseWithAddToAuthNFailureCache(client, token, http.StatusUnauthorized, fmt.Sprintf("Authentication failure with the token against from the user management service: %v", authError))
		}
		if tokenAuthResponse != nil && len(tokenAuthResponse.ActorCrn) > 0 {
			log.Infof("Successful authentication (with token) for machine user with crn '%v'", tokenAuthResponse)
			client.AddToAuthNCache(token, tokenAuthResponse.ActorCrn)
			return &CdpAuthResponse{Success: true}, tokenAuthResponse.ActorCrn
		}
		return GetErrorResponseWithAddToAuthNFailureCache(client, token, http.StatusUnauthorized, "Authentication failure")
	} else {
		msgToVerify, err := base64.URLEncoding.DecodeString(msgString)
		if err != nil {
			log.Errorf("Cannot decode msg: %v", err)
			return &CdpAuthResponse{Success: false, StatusCode: http.StatusBadRequest, Message: "Cannot decode X-Altus-Msg header"}, ""
		}
		umc := client.userManagementClient
		authResponse, err := umsAuthenticate(umc, ctx, authString, string(msgToVerify))
		if err != nil {
			authError := fmt.Sprintf("Auth error (by using auth string) - %v", err)
			return &CdpAuthResponse{Success: false, StatusCode: http.StatusUnauthorized, Message: fmt.Sprintf("UMS authentication error: %v", authError)}, ""
		}
		if authResponse != nil && len(authResponse.ActorCrn) > 0 {
			log.Infof("Successful authentication (with auth string) for machine user with crn '%v'", authResponse)
			return &CdpAuthResponse{Success: true}, authResponse.ActorCrn
		}
		return &CdpAuthResponse{Success: false, StatusCode: http.StatusUnauthorized, Message: "Authentication failure"}, ""
	}
}

func validateRequest(r *http.Request) (bool, string, string, string, string) {
	authString := GetAuthString(r)
	token := GetAuthToken(r)
	msgString := ""
	if len(authString) == 0 {
		return false, "Auth string header is required.", "", "", ""
	}
	if len(token) == 0 {
		msgString = GetAuthMsg(r)
		if len(msgString) == 0 {
			return false, "Auth token or msg header is required", "", "", ""
		}
	}
	return true, "", authString, token, msgString
}

func umsAuthenticate(umc usermanagement.UserManagementClient, ctx context.Context, authHeader string, msgToVerify string) (*usermanagement.AuthenticateResponse, error) {
	request := usermanagement.AccessKeyRequestSigningV1AuthRequest{
		AuthHeader:  authHeader,
		MsgToVerify: msgToVerify,
	}
	authMethod := usermanagement.AuthenticateRequest_AccessKeyV1AuthRequest{
		AccessKeyV1AuthRequest: &request,
	}
	authRequest := usermanagement.AuthenticateRequest{
		AuthenticationMethod: &authMethod,
	}
	return umc.Authenticate(ctx, &authRequest)
}

type tokenstr struct {
	AccessKeyId string `json:"access_key_id"`
	ExpiredAt   string `json:"expiredAt"`
	SignedStr   string `json:"signedStr"`
}

func GetErrorResponseWithAddToAuthNFailureCache(client *UMSClient, token string, statusCode int, msg string) (*CdpAuthResponse, string) {
	AddToFailureCache(client.authNFailureCache, token, msg, statusCode)
	return &CdpAuthResponse{Success: false, StatusCode: statusCode, Message: msg}, ""
}

func checkTokenValidity(originalToken string, token tokenstr) (bool, error) {
	expiredAt, err := stringToTime(token.ExpiredAt)
	if err != nil {
		return false, err
	}
	timeNowPlus := time.Now().Add(time.Minute * 65)
	timeNowMinus := time.Now().Add(-time.Minute * 65)
	if expiredAt.After(timeNowPlus) || expiredAt.Before(timeNowMinus) {
		log.Printf("Token validation failed for: %s", originalToken)
		return false, nil
	}
	return true, nil
}

func stringToTime(s string) (time.Time, error) {
	sec, err := strconv.ParseInt(s, 10, 64)
	if err != nil {
		return time.Time{}, err
	}
	return time.Unix(sec, 0), nil
}
