package proxy

import (
	"context"
	"fmt"
	"net/http"
	"strings"

	log "github.com/sirupsen/logrus"
	"github.infra.cloudera.com/cdp-ums-auth-proxy/generated/usermanagement"
)

func (client *UMSClient) AuthR(r *http.Request, actorCrn string) *CdpAuthResponse {
	numOfRigthsToCheck := len(client.checkRights)
	if numOfRigthsToCheck == 0 && !client.checkAccountID {
		log.Debugf("Authorization is disabled.")
		return &CdpAuthResponse{Success: true}
	}
	token := GetAuthToken(r)
	requestId := GetRequestIdFromHeader(r)
	ctx := ContextWithInternalActor(context.Background())
	if len(requestId) > 0 {
		log.Debugf("Processing authorization request with id: %s", requestId)
		ctx = ContextWithRequestId(ctx, requestId)
	}
	if len(token) > 0 {
		actorCrn, found := client.GetFromAuthRCache(token)
		if found {
			log.Debugf("Use token for athorization from cache for actorCrn: %v", actorCrn)
			return &CdpAuthResponse{Success: true}
		} else {
			msg, statusCode, failFound := client.GetFromAuthRFailureCache(token)
			if failFound {
				return &CdpAuthResponse{Success: false, StatusCode: statusCode, Message: msg}
			}
		}
	}
	if client.checkAccountID && len(actorCrn) > 0 {
		log.Debugf("Verify accountId for %v", actorCrn)
		if len(actorCrn) > 0 {
			accountIdCheckResult := checkAccountID(r, actorCrn, client)
			if accountIdCheckResult {
				if numOfRigthsToCheck == 0 {
					return &CdpAuthResponse{Success: true}
				}
			} else {
				errorMsg := fmt.Sprintf("Machine user %s cannot write into account with this request path: %s", actorCrn, r.URL.Path)
				if len(token) > 0 {
					return GetErrorResponseWithAddToAuthRFailureCache(client, token, http.StatusForbidden, errorMsg)
				}
				return &CdpAuthResponse{Success: false, StatusCode: http.StatusForbidden, Message: errorMsg}
			}
		}
	}
	authResponse, err := umsCheckRights(client.authorizationClient, ctx, actorCrn, client.checkRights)
	if err != nil {
		authRErr := fmt.Sprintf("%v", err)
		return &CdpAuthResponse{Success: false, Message: fmt.Sprintf("Authorization RPC call failed - %v", authRErr), StatusCode: http.StatusUnauthorized}
	}
	if authResponse != nil {
		hasAllRights := true
		for _, hasRight := range authResponse.Result {
			if !hasRight {
				hasAllRights = false
				break
			}
		}
		if hasAllRights {
			log.Infof("Athorization is successful for actorCrn: %v", actorCrn)
			if len(token) > 0 {
				client.AddToAuthRCache(token, actorCrn)
			}
			return &CdpAuthResponse{Success: true}
		}
		requiredRights := strings.Join(client.checkRights, ",")
		errorMsg := fmt.Sprintf("Required rights (%v) missing to execute operation for actorCrn: %v", requiredRights, actorCrn)
		if len(token) > 0 {
			return GetErrorResponseWithAddToAuthRFailureCache(client, token, http.StatusForbidden, errorMsg)
		}
		return &CdpAuthResponse{Success: false, StatusCode: http.StatusForbidden, Message: errorMsg}
	}
	return &CdpAuthResponse{Success: false, StatusCode: http.StatusForbidden, Message: "Unauthorized"}
}

func umsCheckRights(authClient usermanagement.AuthorizationClient, ctx context.Context, actorCrn string, rightsInput []string) (*usermanagement.HasRightsResponse, error) {
	checkRights := make([]*usermanagement.RightCheck, 0)
	for _, right := range rightsInput {
		checkRights = append(checkRights, &usermanagement.RightCheck{Right: right})
	}
	hasRightsRequest := usermanagement.HasRightsRequest{ActorCrn: actorCrn, Check: checkRights}
	hasRightsResponse, err := authClient.HasRights(ctx, &hasRightsRequest)
	if err != nil {
		return nil, err
	}
	return hasRightsResponse, nil
}

func checkAccountID(r *http.Request, crn string, client *UMSClient) bool {
	splitted := strings.Split(crn, ":")
	originalUri := r.Header.Get(client.originalUriHeaderName)
	if len(splitted) > 4 && len(originalUri) > 0 {
		accountId := splitted[4]
		log.Debugf("Check account id '%s' against %s", accountId, originalUri)
		return strings.Contains(originalUri, accountId)
	} else if len(originalUri) == 0 {
		log.Debugf("%s header is missing from the auth request", client.originalUriHeaderName)
	}
	return false
}

func GetErrorResponseWithAddToAuthRFailureCache(client *UMSClient, token string, statusCode int, msg string) *CdpAuthResponse {
	AddToFailureCache(client.authRFailureCache, token, msg, statusCode)
	return &CdpAuthResponse{Success: false, StatusCode: statusCode, Message: msg}
}
