package proxy

import (
	"path/filepath"
	"strings"

	"github.com/spf13/viper"
)

type UMSAuthProxyConfig struct {
	Port                int64
	UMSInfo             *UMSInfo
	UMSCache            *UMSCache
	LoggerConfig        *LoggerConfig
	AuthorizationConfig *AuthorizationConfig
}

type UMSInfo struct {
	Host                           string
	Port                           int64
	ConnectTimeoutSeconds          int64
	RequestTimeoutThresholdSeconds int64
}

type UMSCache struct {
	TtlMinutes        int64
	TtlFailureSeconds int64
	MaxSizeMB         int64
	Shards            int64
	CleanTimeMinutes  int64
}

type LoggerConfig struct {
	Level string
	File  bool
}

type AuthorizationConfig struct {
	Enabled               bool
	VerifyAccountID       bool
	OriginalUriHeaderName string
	Rights                []string
}

func CreateUMSAuthProxyConfig(configLocation string) (*UMSAuthProxyConfig, error) {
	v := viper.GetViper()
	file := filepath.Base(configLocation)
	parentDir := filepath.Dir(configLocation)
	v.SetConfigName(fileNameWithoutExtTrimSuffix(file))
	v.SetConfigType("yaml")
	v.AddConfigPath(parentDir)
	v.AutomaticEnv()
	v.SetEnvKeyReplacer(strings.NewReplacer(`.`, `_`))
	err := v.ReadInConfig()
	if err != nil {
		return nil, err
	}
	config := UMSAuthProxyConfig{}
	v.SetDefault("port", 61000)
	config.Port = v.GetInt64("port")
	v.SetDefault("ums.connectTimeoutSeconds", 30)
	v.SetDefault("ums.requestTimeoutThresholdSeconds", 5)
	ums := UMSInfo{
		Host:                           v.GetString("ums.host"),
		Port:                           v.GetInt64("ums.port"),
		ConnectTimeoutSeconds:          v.GetInt64("ums.connectTimeoutSeconds"),
		RequestTimeoutThresholdSeconds: v.GetInt64("ums.requestTimeoutThresholdSeconds"),
	}
	config.UMSInfo = &ums
	v.SetDefault("cache.ttlMinutes", 90)
	v.SetDefault("cache.ttlFailureSeconds", 20)
	v.SetDefault("cache.maxSizeMB", 8912)
	v.SetDefault("cache.shards", 256)
	v.SetDefault("cache.cleanTimeMinutes", 5)
	umsCache := UMSCache{
		TtlMinutes:        v.GetInt64("cache.ttlMinutes"),
		TtlFailureSeconds: v.GetInt64("cache.ttlFailureSeconds"),
		MaxSizeMB:         v.GetInt64("cache.maxSizeMB"),
		Shards:            v.GetInt64("cache.shards"),
		CleanTimeMinutes:  v.GetInt64("cache.cleanTimeMinutes"),
	}
	config.UMSCache = &umsCache
	v.SetDefault("logger.level", "info")
	loggerConfig := LoggerConfig{
		File:  v.GetBool("logger.file"),
		Level: v.GetString("logger.level"),
	}
	config.LoggerConfig = &loggerConfig
	v.SetDefault("authorization.verifyAccountId", true)
	v.SetDefault("authorization.originalUriHeaderName", "X-Original-URI")
	v.SetDefault("authorization.rights", make([]string, 0))
	authorizationConfig := AuthorizationConfig{
		Enabled:               v.GetBool("authorization.enabled"),
		VerifyAccountID:       v.GetBool("authorization.verifyAccountId"),
		OriginalUriHeaderName: v.GetString("authorization.originalUriHeaderName"),
		Rights:                v.GetStringSlice("authorization.rights"),
	}
	config.AuthorizationConfig = &authorizationConfig
	return &config, nil
}

func fileNameWithoutExtTrimSuffix(fileName string) string {
	return strings.TrimSuffix(fileName, filepath.Ext(fileName))
}
