package proxy

import (
	"encoding/json"
	"fmt"
	"net/http"
	"os"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promauto"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	log "github.com/sirupsen/logrus"
)

var (
	responseCounter = promauto.NewCounterVec(prometheus.CounterOpts{
		Name: "ums_auth_proxy_http_requests_total",
		Help: "HTTP auth response of the proxy",
	}, []string{"status_code"})
	internalResponseCounter = promauto.NewCounterVec(prometheus.CounterOpts{
		Name: "ums_auth_proxy_internal_http_requests_total",
		Help: "Internal HTTP auth response of the proxy",
	}, []string{"status_code"})
)

func (c *UMSClient) checkAuth(w http.ResponseWriter, r *http.Request) {
	headerInfo := GetRequestHeaderInfo(r)
	authNResp, machineUser := c.AuthN(r)
	if !authNResp.Success {
		log.Printf("Handling authN request failed. internal status code: %d - status code: %d - msg: %v%v", authNResp.StatusCode, http.StatusUnauthorized, authNResp.Message, headerInfo)
		responseCounter.WithLabelValues(fmt.Sprint(http.StatusUnauthorized)).Inc()
		internalResponseCounter.WithLabelValues(fmt.Sprint(authNResp.StatusCode)).Inc()
		http.Error(w, "Authenticated: false", http.StatusUnauthorized)
		return
	}
	authRResp := c.AuthR(r, machineUser)
	if !authRResp.Success {
		log.Printf("Handling authR request failed. internal status code: %d - status code: %d - msg: %v%v", authRResp.StatusCode, http.StatusForbidden, authRResp.Message, headerInfo)
		responseCounter.WithLabelValues(fmt.Sprint(http.StatusForbidden)).Inc()
		internalResponseCounter.WithLabelValues(fmt.Sprint(authRResp.StatusCode)).Inc()
		http.Error(w, "Authorized: false", http.StatusForbidden)
		return
	}
	log.Debugf("Handling auth request successful.%v", headerInfo)
	w.WriteHeader(http.StatusOK)
	responseCounter.WithLabelValues(fmt.Sprint(http.StatusOK)).Inc()
	internalResponseCounter.WithLabelValues(fmt.Sprint(http.StatusOK)).Inc()
	return
}

func Serve(configLocation string) {
	conf, err := CreateUMSAuthProxyConfig(configLocation)
	if err != nil {
		fmt.Println(err)
		os.Exit(1)
	}
	InitLogger(conf)
	confJson, jsonWriteErr := json.Marshal(conf)
	if jsonWriteErr != nil {
		fmt.Println(jsonWriteErr)
		os.Exit(1)
	}
	log.Printf("Application config: %v", string(confJson))
	umsClient, err := NewUMSClient(conf)
	if err != nil {
		log.Error(fmt.Sprintf("Cannot create new UMS client: %v", err))
		os.Exit(1)
	}
	promauto.NewGaugeFunc(prometheus.GaugeOpts{
		Name:        "ums_auth_proxy_cache_item_count",
		Help:        "successful/failed UMS authentication/authorization cache size",
		ConstLabels: prometheus.Labels{"cache_type": "authn_success"},
	}, func() float64 {
		return float64(umsClient.authNtokenCache.cache.Len())
	})
	promauto.NewGaugeFunc(prometheus.GaugeOpts{
		Name:        "ums_auth_proxy_cache_item_count",
		Help:        "successful/failed UMS authentication/authorization cache size",
		ConstLabels: prometheus.Labels{"cache_type": "authn_failure"},
	}, func() float64 {
		return float64(umsClient.authNFailureCache.cache.Len())
	})
	promauto.NewGaugeFunc(prometheus.GaugeOpts{
		Name:        "ums_auth_proxy_cache_item_count",
		Help:        "successful/failed UMS authentication/authorization cache size",
		ConstLabels: prometheus.Labels{"cache_type": "authr_success"},
	}, func() float64 {
		return float64(umsClient.authRtokenCache.cache.Len())
	})
	promauto.NewGaugeFunc(prometheus.GaugeOpts{
		Name:        "ums_auth_proxy_cache_item_count",
		Help:        "successful/failed UMS authentication/authorization cache size",
		ConstLabels: prometheus.Labels{"cache_type": "authr_failure"},
	}, func() float64 {
		return float64(umsClient.authRFailureCache.cache.Len())
	})
	http.Handle("/metrics", promhttp.Handler())
	http.HandleFunc("/authenticated", umsClient.checkAuth)
	log.Fatal(http.ListenAndServe(fmt.Sprintf(":%d", conf.Port), nil))
}
