package proxy

import (
	"context"
	"fmt"
	"time"

	bigcache "github.com/allegro/bigcache/v3"
	"github.com/google/uuid"

	log "github.com/sirupsen/logrus"
	"github.infra.cloudera.com/cdp-ums-auth-proxy/generated/usermanagement"
	"google.golang.org/grpc"
	"google.golang.org/grpc/metadata"
)

const (
	DefaultRequestTimeThreshold        = time.Second * 5
	internalActorCrn            string = "crn:altus:iam:us-west-1:altus:user:__internal__actor__"
	authNTokenCacheName         string = "authn-token-cache"
	authNFailureCacheName       string = "authn-token-cache"
	authRTokenCacheName         string = "authn-token-cache"
	authRFailureCacheName       string = "authn-token-cache"
)

type UMSClient struct {
	clientConn            *grpc.ClientConn
	authorizationClient   usermanagement.AuthorizationClient
	userManagementClient  usermanagement.UserManagementClient
	authNtokenCache       *CdpTokenCache
	authNFailureCache     *CdpTokenCache
	authRtokenCache       *CdpTokenCache
	authRFailureCache     *CdpTokenCache
	checkRights           []string
	checkAccountID        bool
	originalUriHeaderName string
}

type CdpToken struct {
	Token       string
	AccessKeyId string
	ExpireAt    int64
	SignedStr   string
}

type CdpTokenCache struct {
	identifier string
	cache      *bigcache.BigCache
	cacheTTL   time.Duration
}

type CdpAuthResponse struct {
	StatusCode int
	Message    string
	Success    bool
}

func NewUMSClient(config *UMSAuthProxyConfig) (*UMSClient, error) {
	log.Info("Creating UMSClient ...")
	ret := UMSClient{}
	connString := fmt.Sprintf("%s:%d", config.UMSInfo.Host, config.UMSInfo.Port)
	conn, err := grpc.Dial(connString, grpc.WithInsecure(),
		grpc.WithTimeout(time.Duration(config.UMSInfo.ConnectTimeoutSeconds)*time.Second),
		WithCdpMetadataInterceptor(),
		slowLogInterceptorFromConf(config))
	if err != nil {
		log.Error(fmt.Sprintf("UMSClient cannot establish connection. err=%v", err))
		return nil, err
	}
	authorizationClient := usermanagement.NewAuthorizationClient(conn)
	userManagementClient := usermanagement.NewUserManagementClient(conn)

	ret.clientConn = conn
	ret.authorizationClient = authorizationClient
	ret.userManagementClient = userManagementClient
	ttlForSuccessCache := time.Minute * time.Duration(config.UMSCache.TtlMinutes)
	ttlFailureCache := time.Duration(config.UMSCache.TtlFailureSeconds) * time.Second
	cleanTime := time.Duration(config.UMSCache.CleanTimeMinutes) * time.Minute
	shards := int(config.UMSCache.Shards)
	succAuthNConfig := bigcache.Config{
		Shards:      shards,
		LifeWindow:  ttlForSuccessCache,
		CleanWindow: cleanTime,
		OnRemove: func(key string, element []byte) {
			log.Debugf("REMOVE %s from '%s' cache.", key, authNTokenCacheName)
		},
	}
	succAuthRConfig := bigcache.Config{
		Shards:      shards,
		LifeWindow:  ttlForSuccessCache,
		CleanWindow: cleanTime,
		OnRemove: func(key string, element []byte) {
			log.Debugf("REMOVE %s from '%s' cache.", key, authRTokenCacheName)
		},
	}
	failAuthNConfig := bigcache.Config{
		Shards:           shards,
		LifeWindow:       ttlFailureCache,
		HardMaxCacheSize: 1024,
		CleanWindow:      1 * time.Minute,
		OnRemove: func(key string, element []byte) {
			log.Debugf("REMOVE %s from '%s' cache.", key, authNFailureCacheName)
		},
	}
	failAuthRConfig := bigcache.Config{
		Shards:           shards,
		LifeWindow:       ttlFailureCache,
		HardMaxCacheSize: 1024,
		CleanWindow:      1 * time.Minute,
		OnRemove: func(key string, element []byte) {
			log.Debugf("REMOVE %s from '%s' cache.", key, authRFailureCacheName)
		},
	}
	authNTokenBigCache, initAuthNSuccErr := bigcache.NewBigCache(succAuthNConfig)
	if initAuthNSuccErr != nil {
		log.Fatalf("Cannot initialize authn token cache: %v", initAuthNSuccErr)
	}
	authRTokenBigCache, initAuthRSuccErr := bigcache.NewBigCache(succAuthRConfig)
	if initAuthRSuccErr != nil {
		log.Fatalf("Cannot initialize authr token cache: %v", initAuthRSuccErr)
	}
	authNFailureBigCache, initAuthNFailcErr := bigcache.NewBigCache(failAuthNConfig)
	if initAuthNFailcErr != nil {
		log.Fatalf("Cannot initialize authn token cache: %v", initAuthNFailcErr)
	}
	authRFailureBigCache, initAuthRFailErr := bigcache.NewBigCache(failAuthRConfig)
	if initAuthRFailErr != nil {
		log.Fatalf("Cannot initialize authr token cache: %v", initAuthRFailErr)
	}
	authNTokenCache := CdpTokenCache{
		identifier: authNTokenCacheName,
		cacheTTL:   ttlForSuccessCache,
		cache:      authNTokenBigCache,
	}
	ret.authNtokenCache = &authNTokenCache
	authRTokenCache := CdpTokenCache{
		identifier: authRTokenCacheName,
		cacheTTL:   ttlForSuccessCache,
		cache:      authRTokenBigCache,
	}
	ret.authRtokenCache = &authRTokenCache
	authNFailureCache := CdpTokenCache{
		identifier: authNFailureCacheName,
		cacheTTL:   ttlFailureCache,
		cache:      authNFailureBigCache,
	}
	ret.authNFailureCache = &authNFailureCache
	authRFailureCache := CdpTokenCache{
		identifier: authRFailureCacheName,
		cacheTTL:   ttlFailureCache,
		cache:      authRFailureBigCache,
	}
	ret.authRFailureCache = &authRFailureCache
	if config.AuthorizationConfig != nil && config.AuthorizationConfig.Enabled {
		ret.checkRights = config.AuthorizationConfig.Rights
		ret.checkAccountID = config.AuthorizationConfig.VerifyAccountID
		ret.originalUriHeaderName = config.AuthorizationConfig.OriginalUriHeaderName
	}
	log.Info("UMSClient creation has been finished.")
	return &ret, nil
}

func slowLogInterceptorFromConf(config *UMSAuthProxyConfig) grpc.DialOption {
	reqTimeThreshold := DefaultRequestTimeThreshold
	thresholdNum := config.UMSInfo.RequestTimeoutThresholdSeconds
	if thresholdNum > 0 {
		reqTimeThreshold = time.Duration(thresholdNum) * time.Second
	}
	return WithLoggingInterceptor(reqTimeThreshold)
}

func cdpMetadataInterceptor(
	ctx context.Context,
	method string,
	req interface{},
	reply interface{},
	cc *grpc.ClientConn,
	invoker grpc.UnaryInvoker,
	opts ...grpc.CallOption,
) error {
	callingServiceName := GetCallingServiceName(ctx)
	ctx = metadata.AppendToOutgoingContext(ctx, "callingServiceName", callingServiceName)

	if actorCrn := GetActor(ctx); actorCrn != "" {
		ctx = metadata.AppendToOutgoingContext(ctx, "actorCrn", actorCrn)
	}
	if requestId := GetRequestId(ctx); requestId != "" {
		ctx = metadata.AppendToOutgoingContext(ctx, "requestID", requestId)
	} else {
		ctx = metadata.AppendToOutgoingContext(ctx, "requestID", uuid.New().String())
	}
	return invoker(ctx, method, req, reply, cc, opts...)
}

func WithCdpMetadataInterceptor() grpc.DialOption {
	return grpc.WithChainUnaryInterceptor(cdpMetadataInterceptor)
}

func WithLoggingInterceptor(slowThreshold time.Duration) grpc.DialOption {
	interceptor := func(
		ctx context.Context,
		method string,
		req interface{},
		reply interface{},
		cc *grpc.ClientConn,
		invoker grpc.UnaryInvoker,
		opts ...grpc.CallOption,
	) error {
		startTime := time.Now()
		ret := invoker(ctx, method, req, reply, cc, opts...)
		duration := time.Since(startTime)

		if duration > slowThreshold {
			log.Warn(fmt.Sprintf("%s: Slow RPC detected! %dus exceeds threshold of %dus",
				method,
				int64(duration/time.Microsecond),
				int64(slowThreshold/time.Microsecond)))
		}
		return ret
	}

	return grpc.WithChainUnaryInterceptor(interceptor)
}

func (client *UMSClient) Close() {
	if client.clientConn != nil {
		client.clientConn.Close()
		client.clientConn = nil
		client.authorizationClient = nil
		client.userManagementClient = nil
	}
}

func GetCallingServiceName(ctx context.Context) string {
	val := ctx.Value("callingServiceName")
	ret, ok := val.(string)
	if !ok {
		return "ums-auth-proxy"
	}
	return ret
}

func GetActor(ctx context.Context) string {
	val := ctx.Value("actorCrn")
	ret, ok := val.(string)
	if !ok {
		return ""
	}
	return ret
}

func GetRequestId(ctx context.Context) string {
	val := ctx.Value("requestID")
	ret, ok := val.(string)
	if !ok {
		return ""
	}
	return ret
}

func ContextWithInternalActor(ctx context.Context) context.Context {
	return ContextWithActorCrn(ctx, internalActorCrn)
}

func ContextWithActorCrn(ctx context.Context, crn string) context.Context {
	return context.WithValue(ctx, "actorCrn", crn)
}

func ContextWithRequestId(ctx context.Context, requestId string) context.Context {
	return context.WithValue(ctx, "requestID", requestId)
}
