# CDP Request Signer Proxy

## Description

CDP Request Signer Proxy is a simple proxy that can be used to sign (UMS) any incoming requests and forward them to a user defined endpoint. Works as a reverse proxy not as a forward proxy, so make sure tools are targeting this proxy for the requests and do not use it as a forwarder only.

## Usage

Pull the docker container and run it with a provided configuration.

```bash
docker pull docker-private.infra.cloudera.com/cloudera/cdp-request-signer:1.2.3-b1
docker start -v /my/path/cdp-request-signer.yaml:/etc/cdp-request-signer/cdp-request-signer.yaml docker-private.infra.cloudera.com/cloudera/cdp-request-signer:1.2.3-b1
```

## Configuration sections

Configuration file (YAML) can be set by `--config` argument for the application. The containerized format of the server uses `/etc/cdp-request-signer/cdp-request-signer.yaml` by default.

Example: 

```yaml
port: 61095
endpoint: "http://nginx:61080/remotewrite/account/altus"
base_path: ""
credential:
  file: "/etc/cdp-request-signer/example_credentials"
  profile: default
  access_key_name: cdp_access_key_id
  secret_key_name: cdp_private_key
sign:
  auth_header_name: "x-altus-auth"
  date_header_name: "x-altus-date"
  msg_header_name: "x-altus-msg"
  token_header_name: "x-altus-token"
  token_date_file: "token_exp"
  resource_crn_header_name: "x-cdp-resource-crn"
  instance_header_name: "x-cdp-instance"
  token_validity_min: 5
  use_token_header: true
  use_msg_header: true
  resource_crn_header: "sample-crn"
  instance_header: "sample-instance"
client:
  dump: true
tls:
  key: "/my/generated.key"
  cert: "/my/generated.crt"
auth:
  username: "admin"
  password: "admin"
logger:
  level: debug
  file: false
proxy:
  https_proxy: "http://localhost:3128"
  http_proxy: "http://localhost:3128"
  no_proxy: ".cloudera.com"
```

#### `endpoint`

The URL where the incoming requests will be processed.

#### `port`

Port that is allocated for the application.

### `base_path`

Path that is part of the endpoint but won't be used in the signing. (so the signing will contain the requested path after this base path)

#### `credential`

Altus/Cdp credential (that is used for signing) related configurations.

- `file`: Location of the file (ini) that contains the altus/cdp credential.
- `profile`: Profile that will be picked from the configuration. (default: `default`)
- `access_key_name`: Key name of the access key property. (default: `cdp_access_key_id`)
- `secret_key_name`: Key name of the secret key property. (default: `cdp_private_key`)
- `sign_algorithm_name`: Key name of the digital sign algorithm. (default: `cdp_sign_algorithm`)

Example:

```ini
# file is located at /my/path/config.ini
[default]
cdp_access_key_id=mykey
cdp_private_key=mysecret
cdp_sign_algorithm=ECDSA
```

The configuration for the above example will look like this:

```yaml
credential:
  file: /my/path/config.ini
  profile: default 
  access_key_name: cdp_access_key_id
  secret_key_name: cdp_private_key
  cdp_sign_algorithm: ECDSA
```

Currently supported signing algorithms: `Ed25519` and `ECDSA` (if not set or the value is invalid, `Ed25519` will be used)

You can avoid to use a file for credentials by providing access/secret keys by the following 2 environment variables: `CDP_ACCESS_KEY_ID` and `CDP_PRIVATE_KEY` (If these 2 are set, you need to use the `CDP_SIGN_ALGORITHM` environment variable for the digital signing algorithm).

#### `sign`

Request signing (Altus-way) related configurations.

- `auth_header_name`: Name of the signed auth string header. (default: `x-altus-auth`)
- `date_header_name`: Name of the auth date header. (default: `x-altus-date`)
- `use_msg_header`: If it is set, a message (base64 encoded) will be sent as a header that can help to know how the signed message looked like before signing. At the moment it is recommended to be set. (default:`true`)
- `msg_header_name`: Name of the auth message header. (default: `x-altus-msg`)
- `use_token_header`: If it is set, an additional timestamp will be signed during request processing. By enabling this option, UMS auth proxy will be able to cache incoming requests until the token lives. (default: `true`)
- `token_header_name`: Name of the auth token header. (default: `x-altus-token`)
- `token_date_file`: Local file that will store the generated token timestamp. (default: `token_exp`)
- `token_validity_min`: Validity of the auth tokens in minutes. (default:`60`)
- `use_request_id_header`: If it is set, a requestId will be generated for every processed request. (default: `true`)
- `request_id_header_name`: Name of the requestId header. (default:`x-cdp-request-id`)
- `resource_crn_header_name`: Name of the resource CRN header. (default: `x-cdp-resource-crn`)
- `resource_crn_header`: Value of the resource CRN header. (Header won't be sent if the value is empty)
- `instance_header_name`: Name of the instance header. (default: `x-cdp-instance`)
- `instance_header`: Value of the instance header. (Header won't be sent if the value is empty)

#### `client`

Http client (that process data into user defined endpoints) related configurations.

- `dump`: Print request/response headers for all processed requests. (default: `false`)
- `timeout`: Timeout for processing requests in seconds. (default: `60`)

#### `logger`

Logger related configurations.

- `level`: Level of the logging (default: `debug`)
- `file`: If it is set, logs will directed into `/var/log/cdp-ums-auth-proxy/cdp-ums-auth-proxy.log` (default: `false`)

#### `tls`

TLS/HTTPS related configurations for the request signer server.

- `key`: Key file location.
- `cert`: Certificate file location.

#### `auth`

Basic authentication related configurations.

- `username`: Username that can be used to authenticate against request signer server.
- `password`: Password for the request signer server user.

#### `proxy`

HTTP proxy related configurations. (Values can be defined by default environment variables as well like: `HTTPS_PROXY` not only `PROXY_HTTPS_PROXY`)

- `https_proxy`: Forward proxy that is used for `https` targets.
- `http_proxy`: Forward proxy that is used for `http` targets.
- `no_proxy`: Skip proxy use if target address matches for this comma separated addresses.

### Environment variables

All the configurations can be set by environment variables as well. (but with uppercase format with `_` separator instead of `.`, e.g.: `sign.resource_crn_header` -> `SIGN_RESOURCE_CRN_HEADER`). For container based deployements, use environment variables in order to override only the non-default parameters.

```yaml
environment:
  - ENDPOINT=http://nginx:61080/remotewrite/account/altus
  - SIGN_RESOURCE_CRN_HEADER=sample-crn
  - SIGN_INSTANCE_HEADER=sample-instance
  - CREDENTIAL_FILE=/etc/cdp-request-signer/example_credentials
  - CREDENTIAL_PROFILE=default
  - CREDENTIAL_ACCESS_KEY_NAME=cdp_access_key_id
  - CREDENTIAL_SECRET_KEY_NAME=cdp_private_key
```

## Debug (VM)

![](images/request-signer.gif)

Asciinema cast file: [here](cast/request-signer.cast)

## Increase versions (VM)

Increasing RPM versions needs to be done manually by updating the `cdp-request-signer`version in the root `pom.xml`.

## Integration with Kubernetes-based clusters

For every build / release, also a `cdp-request-signer` image is published to the ECRs. By using proper config maps or environment variables for the container (configurations can be found above in the configuration section) integrating the service should be simple. (`integration-test` module in cdp-infra-tools also using docker container based solutions)

### What is needed for request-signer configs?

- Remote write URL endpoint (can be gathered from internal `environment/telemetry/monitoring` CB endpoint)
- Machine user access/secret key (generate by UMS)
- Assign `ComputeMetricsPublisher` role to the generated machine user (important for PaaS !!!)

### How to "force" an account to use compute monitoring?

By default, accounts with `CDP_SAAS` entitlement will use compute monitoring. For PaaS clusters, it can be enabled by `CDP_CENTRAL_COMPUTE_MONITORING` entitlement. (this entitlement won't be needed if `telemetry.monitoring.paasSupport` config is set globally by CB environment service)

## Integration testing

In `integration-test` module, you can run tests by using:

```bash
make test
```

Or if you do not want to validate the results of the tests, but you would like to setup the docker-based environment, and observing the container manually you can also just do this:

```bash
make run
```

## Available releases

[RelEng portal link](https://release.infra.cloudera.com/stackspage?stack=CDP-INFRA-TOOLS&filter=SHIPPED)
