package tests

import (
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"os"
	"strconv"
	"strings"
	"testing"
	"time"
)

func TestProxyMetrics(t *testing.T) {
	requestSignerMetricsUrl, ok := os.LookupEnv("REQUEST_SINGER_METRICS_URL")
	if !ok {
		fmt.Println("Setting REQUEST_SINGER_METRICS_URL is required")
		os.Exit(1)
	}
	fmt.Println(fmt.Sprintf("Using %s as request signer metrics url", requestSignerMetricsUrl))
	umsAuthMetricsUrl, ok := os.LookupEnv("UMS_SERVER_AUTH_METRICS_URL")
	if !ok {
		fmt.Println("Setting UMS_SERVER_AUTH_METRICS_URL is required")
		os.Exit(1)
	}
	fmt.Println(fmt.Sprintf("Using %s as ums server auth proxy metrics url", umsAuthMetricsUrl))
	itWorkspaceDir, ok := os.LookupEnv("IT_WORKSAPCE_DIR")
	if !ok {
		fmt.Println("Setting IT_WORKSAPCE_DIR is required")
		os.Exit(1)
	}
	testResultJsonFileName, ok := os.LookupEnv("OUTPUT_TEST_FILE")
	if !ok {
		fmt.Println("Setting OUTPUT_TEST_FILE is required")
		os.Exit(1)
	}
	fmt.Println(fmt.Sprintf("Using %s as workspace directory", itWorkspaceDir))
	fmt.Println("Waiting 20 seconds ... (waiting for metrics)")
	time.Sleep(20 * time.Second)
	resultMap := make(map[string][]string)
	resultMap["passes"] = make([]string, 0)
	resultMap["failures"] = make([]string, 0)
	signerMetricsResp, err := getMetrics(requestSignerMetricsUrl)
	if err != nil {
		errorMsg := fmt.Sprintf("Error during processing request signer metrics: %v", err)
		fmt.Println(errorMsg)
		resultMap["failures"] = append(resultMap["failures"], fmt.Sprintf("Request signer metrics cannot be processed on %v.", requestSignerMetricsUrl))
		writeTestJson(resultMap, itWorkspaceDir, testResultJsonFileName)
		t.Error(errorMsg)
		return
	}
	signerMetrics := string(signerMetricsResp)
	signerSuccNum := 0
	signerErrNum := 0
	if len(signerMetrics) > 0 {
		resultMap["passes"] = append(resultMap["passes"], "Request signer proxy has metrics")
		splitByLine := strings.Split(signerMetrics, "\n")
		for _, val := range splitByLine {
			if strings.HasPrefix(val, "request_signer_remote_write_http_requests_total{status_code=\"2") {
				num, _ := strconv.Atoi(strings.Split(val, " ")[1])
				signerSuccNum += num
			} else if strings.HasPrefix(val, "request_signer_remote_write_http_requests_total{status_code=\"3") {
				num, _ := strconv.Atoi(strings.Split(val, " ")[1])
				signerSuccNum += num
			} else if strings.HasPrefix(val, "request_signer_remote_write_http_requests_total{status_code=\"4") {
				num, _ := strconv.Atoi(strings.Split(val, " ")[1])
				signerErrNum += num
			} else if strings.HasPrefix(val, "request_signer_remote_write_http_requests_total{status_code=\"5") {
				num, _ := strconv.Atoi(strings.Split(val, " ")[1])
				signerErrNum += num
			}
		}
	}
	if signerSuccNum < 1 {
		errorMsg := fmt.Sprintf("Successful request signer calls are lower than 1. (success: %v, errors: %v)", signerSuccNum, signerErrNum)
		resultMap["failures"] = append(resultMap["failures"], errorMsg)
		t.Error(errorMsg)
	} else {
		resultMap["passes"] = append(resultMap["passes"], fmt.Sprintf("Request signer has successful requests. (successful: %v, failure: %v)", signerSuccNum, signerErrNum))
	}

	usmServerProxyMetricsResp, err := getMetrics(umsAuthMetricsUrl)
	if err != nil {
		errorMsg := fmt.Sprintf("Error during processing UMS server auth metrics: %v", err)
		fmt.Println(errorMsg)
		resultMap["failures"] = append(resultMap["failures"], fmt.Sprintf("UMS server auth metrics cannot be processed on %v.", umsAuthMetricsUrl))
		writeTestJson(resultMap, itWorkspaceDir, testResultJsonFileName)
		t.Error(errorMsg)
		return
	}
	usmServerProxyMetric := string(usmServerProxyMetricsResp)
	authNSuccCacheSize := 0
	authNFailCacheSize := 0
	authRSuccCacheSize := 0
	authRFailCacheSize := 0
	if len(usmServerProxyMetric) > 0 {
		resultMap["passes"] = append(resultMap["passes"], "UMS server proxy has metrics")
		splitByLine := strings.Split(usmServerProxyMetric, "\n")
		for _, val := range splitByLine {
			if strings.HasPrefix(val, "ums_auth_proxy_cache_item_count{cache_type=\"authn_success\"}") {
				num, _ := strconv.Atoi(strings.Split(val, " ")[1])
				authNSuccCacheSize += num
			} else if strings.HasPrefix(val, "ums_auth_proxy_cache_item_count{cache_type=\"authn_failure\"}") {
				num, _ := strconv.Atoi(strings.Split(val, " ")[1])
				authNFailCacheSize += num
			} else if strings.HasPrefix(val, "ums_auth_proxy_cache_item_count{cache_type=\"authr_success\"}") {
				num, _ := strconv.Atoi(strings.Split(val, " ")[1])
				authRSuccCacheSize += num
			} else if strings.HasPrefix(val, "ums_auth_proxy_cache_item_count{cache_type=\"authr_failure\"}") {
				num, _ := strconv.Atoi(strings.Split(val, " ")[1])
				authRFailCacheSize += num
			}
		}
		if authNSuccCacheSize < 1 {
			errorMsg := fmt.Sprintf("UMS Authentication cache (successful) is empty on UMS server proxy. (success cache size: %v, failure cache size: %v)", authNSuccCacheSize, authNFailCacheSize)
			resultMap["failures"] = append(resultMap["failures"], errorMsg)
			t.Error(errorMsg)
		} else {
			resultMap["passes"] = append(resultMap["passes"], fmt.Sprintf("UMS Authentication cache sizes are the expected values (success cache size: %v, failure cache size: %v)", authNSuccCacheSize, authNFailCacheSize))
		}
		if authRSuccCacheSize < 1 {
			errorMsg := fmt.Sprintf("UMS Authorization cache (successful) is empty on UMS server proxy. (success cache size: %v, failure cache size: %v)", authRSuccCacheSize, authRFailCacheSize)
			resultMap["failures"] = append(resultMap["failures"], errorMsg)
			t.Error(errorMsg)
		} else {
			resultMap["passes"] = append(resultMap["passes"], fmt.Sprintf("UMS Authorization cache sizes are the expected values (success cache size: %v, failure cache size: %v)", authRSuccCacheSize, authRFailCacheSize))
		}
		writeTestJson(resultMap, itWorkspaceDir, testResultJsonFileName)
	}
}

func getMetrics(url string) ([]byte, error) {
	c := http.Client{Timeout: time.Duration(5) * time.Second}
	resp, err := c.Get(url)
	if err != nil {
		return nil, err
	}
	defer resp.Body.Close()
	body, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		return nil, err
	}
	return body, nil
}

func writeTestJson(data map[string][]string, worskapce string, fileName string) {
	file, _ := json.MarshalIndent(data, "", " ")
	_ = ioutil.WriteFile(fmt.Sprintf("%s/%s", worskapce, fileName), file, 0644)
}
